/* $Id: vircam_science_process.c,v 1.10 2013-10-21 14:43:04 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-21 14:43:04 $
 * $Revision: 1.10 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <cpl.h>
#include <math.h>

#include <casu_utils.h>
#include <casu_mask.h>
#include <casu_mods.h>
#include <casu_wcsutils.h>
#include <casu_stats.h>
#include <casu_filt.h>
#include <casu_sky.h>

#include "vircam_utils.h"
#include "vircam_pfits.h"
#include "vircam_dfs.h"
#include "vircam_mods.h"
#include "vircam_paf.h"
#include "vircam_sky.h"

typedef struct {
    cpl_frameset   *contrib;
    cpl_frame      *objmask;
    int            skyalgo;
    cpl_frame      *skyframe;
    cpl_frame      *template;
    char           fname[BUFSIZ];
} skystruct;

typedef struct {
    cpl_frameset *current;
    cpl_frameset *orig;
    int          *whichsky;
    casu_fits    *stack[VIRCAM_NEXTN];
    casu_fits    *stackc[VIRCAM_NEXTN];
    casu_tfits   *cat[VIRCAM_NEXTN];
    casu_tfits   *mstds[VIRCAM_NEXTN];
    cpl_frame    *product_frame_im;
    cpl_frame    *product_frame_conf;
    cpl_frame    *product_frame_cat;
    int          tilenum;
    int          jitternum;
    char         projname[16];
    double       mjd_start;
    double       mjd_end;
    int          ispj;
} pawprint;

/* User supplied command line options */

struct {

    /* Input */

    int         destripe;
    int         str_filt;
    int         skyalgo;
    int         preview_only;
    int         minphotom;
    int         prettynames;
    int         cdssearch;
    char        *cacheloc;
    float       magerrcut;

} vircam_sci_config;


struct {
    int    niter;
    int    ipix;
    float  thresh;
    int    nbsize;
    float  smkern;
} psm;

struct {
    float   lthr;
    float   hthr;
    int     method;
    int     seeing;
    int     fast;
    int     nfst;
} stk;

struct {
    int    ipix;
    float  thresh;
    int    icrowd;
    float  rcore;
    int    nbsize;
} stk_cat;

struct {

    /* Level 0 stuff */

    cpl_size         *labels;
    cpl_frame        *master_dark;
    cpl_frame        *master_twilight_flat;
    cpl_frame        *master_conf;
    cpl_frame        *master_sky;
    cpl_frame        *master_objmask;
    casu_mask         *mask;
    cpl_frame        *chantab;
    cpl_frame        *phottab;
    cpl_table        *tphottab;
    cpl_frame        *readgain_file;
    cpl_frameset     *science_frames;
    cpl_frameset     *offset_skies;
    cpl_frame        **product_frames_simple;
    cpl_frame        **product_frames_simple_off;
    cpl_propertylist *phupaf;
    float            *gaincors;
    char             *catpath;
    char             *catname;
    int              nscipaws;
    pawprint         *science_paws;
    int              noffpaws;
    pawprint         *offsky_paws;
    int              nskys;
    skystruct        *skys;
    cpl_frame        *schlf_n;
    cpl_frame        *schlf_s;

    /* Level 1 stuff */

    casu_fits         *fdark;
    casu_fits         *fflat;
    casu_fits         *fconf;
    casu_fits         *fsky;
    casu_tfits        *fchantab;

    int              nscience;
    casu_fits         **sci_fits;
    int              noffsets;
    casu_fits         **offsky_fits;

} ps;

/* List of data products types */

enum {SIMPLE_FILE,
      SIMPLE_CONF,
      SIMPLE_CAT,
      STACK_FILE,
      STACK_CONF,
      STACK_CAT,
      SKY_FILE
};

#define SKYNONE           0
#define SKYMASTER        -1
#define PAWSKY_MASK       1
#define PAWSKY_MASK_PRE   2
#define TILESKY           3
#define TILESKY_MINUS     4
#define AUTO              5
#define PAWSKY_MINUS      6

#define VMCPROJ "179.B-2003"
#define VIDEOPROJ "179.A-2006"

#define freepaws(_p,_n) if (_p != NULL) {vircam_sci_paws_delete(_n,&(_p)); _p = NULL; _n = 0;}

char current_cat[VIRCAM_PATHSZ];
char current_catpath[VIRCAM_PATHSZ];
char vircam_recipename[VIRCAM_PATHSZ];
char vircam_recipepaf[VIRCAM_PATHSZ];
cpl_propertylist *dummyqc = NULL;
int offsky = 0;

/* Main stuff for CPL/esorex */

static int vircam_science_process_create(cpl_plugin *plugin);
static int vircam_science_process_exec(cpl_plugin *plugin);
static int vircam_science_process_destroy(cpl_plugin *plugin);
static int vircam_science_process(cpl_parameterlist *parlist,
                                  cpl_frameset *framelist);

/* Pawprint manipulation routines */

static void vircam_sci_paw_init(pawprint *paw, cpl_frameset *frms, int ispj, 
                                int *nfrms_tot);
static void vircam_sci_paw_delete(pawprint *paws);
static void vircam_sci_paws_delete(int npaws, pawprint **paws);
static void vircam_sci_paws_create(cpl_frameset *infrms, int *npaws, 
                                   pawprint **paws);

/* Processing routines */

static int vircam_sci_sky_stripe_wcs(int npaw, pawprint *paw, const char *tag,
                                     cpl_frameset *framelist, 
                                     cpl_parameterlist *parlist, int destripe);
static void vircam_sci_cat(pawprint *paw, int extn);
static void vircam_sci_stack(pawprint *paw, int extn);
static void vircam_sci_wcsfit(casu_fits **in, casu_fits **conf, 
                              casu_tfits **incat, int nf, int level);
static void vircam_sci_init(void);
static void vircam_sci_tidy(int level);
static int vircam_destripe_four(casu_fits **curfits, int stripefilt, 
                                int *status);
static void stripe_profile(casu_fits *in, int stripefilt, float *profile,
                           unsigned char *pbpm);
static void vircam_sci_photcal(pawprint *paw);

/* Sky related routines */

static void vircam_sci_choose_skyalgo(int *nskys, skystruct **skys);
static int vircam_sci_makesky(cpl_frameset *framelist, 
                              cpl_parameterlist *parlist, skystruct *sky,
                              const char *tag);
static void vircam_sci_assign_sky_all(pawprint paw, int whichone);
static skystruct vircam_sci_crsky(int algorithm, cpl_frameset *frms,
                                  cpl_frame *template, int snum);
static void vircam_sci_skydefine(void);
static int vircam_sci_tilesky(cpl_frameset *framelist,
                              cpl_parameterlist *parlist, cpl_frameset *in,
                              cpl_frame *template, 
                              char *fname, cpl_frame **product_frame);
static int vircam_sci_pawsky_mask(cpl_frameset *framelist, 
                                  cpl_parameterlist *parlist, int algo,
                                  cpl_frameset *contrib, cpl_frame *template,
                                  char *fname, cpl_frame **product_frame);
static int vircam_sci_tilesky_minus(cpl_frameset *framelist, 
                                    cpl_parameterlist *parlist,
                                    cpl_frameset *in, cpl_frame *template,
                                    const char *tag, char *fname,
                                    cpl_frame **skyframe);
static int vircam_sci_pawsky_minus(cpl_frameset *framelist, 
                                   cpl_parameterlist *parlist,
                                   cpl_frameset *in, cpl_frame *template,
                                   const char *tag, char *skyname,
                                   cpl_frame **skyframe);

/* Utility routines */

static cpl_frame *vircam_sci_findtemplate(cpl_frame *in);
static int vircam_sci_istile(cpl_frame *frm);
static int vircam_sci_ispj(cpl_frameset *frms);
static int vircam_sci_cmp_jit(const cpl_frame *frame1, 
                              const cpl_frame *frame2);
static int vircam_sci_cmp_tstart(const cpl_frame *frame1, 
                                 const cpl_frame *frame2);
static int vircam_sci_cmp_property(const cpl_property *p1, 
                                   const cpl_property *p2);
static cpl_frameset *vircam_sci_merge_framesets(int n, cpl_frameset **in);
static cpl_frameset *vircam_sci_update_frameset(cpl_frameset *frms, int *nfrms_tot);
static int vircam_sci_testfrms(cpl_frameset *frms, int nextn_expected, 
                               int isimg);
static int vircam_sci_testfrm_1(cpl_frame *fr, int nextn_expected, int isimg);
static void vircam_sci_get_readnoise_gain(int jext, float *readnoise, 
                                          float *gain);
static void vircam_sci_product_name(char *template, int producttype, 
                                    int nametype, int fnumber, char *outfname);

/* Save routines */

static int vircam_sci_save_simple(casu_fits *obj, cpl_frameset *framelist,
                                  cpl_parameterlist *parlist, int isprod,
                                  cpl_frame *template, int isfirst,
                                  const char *tag, char *fname,
                                  cpl_frame **product_frame);
static int vircam_sci_save_sky(casu_fits *outsky, cpl_frameset *framelist,
                               cpl_parameterlist *parlist, 
                               char *fname, cpl_frame *template, int isfirst,
                               cpl_frame **product_frame);
static int vircam_sci_save_stack(casu_fits *stack, cpl_frameset *framelist,
                                 cpl_parameterlist *parlist,
                                 cpl_frame *template, int fnametype,
                                 int stack_num, cpl_frame **product_frame);
static int vircam_sci_save_stack_conf(casu_fits *stack, cpl_frameset *framelist,
                                      cpl_parameterlist *parlist,
                                      cpl_frame *template, int fnametype,
                                      int stack_num, cpl_frame **product_frame);
static int vircam_sci_save_stack_cat(casu_tfits *stack, cpl_frameset *framelist,
                                     cpl_parameterlist *parlist, 
                                     cpl_frame *template, int fnametype,
                                     int stack_num, cpl_frame **product_frame);


static char vircam_science_process_description[] =
"vircam_science_process -- VIRCAM science product recipe.\n\n"
"Process a complete pawprint for VIRCAM data. Remove instrumental\n"
"signature, background correct, combine jitters, photometrically\n"
"and astrometrically calibrate the pawprint image\n\n"
"The program accepts the following files in the SOF:\n\n"
"    Tag                   Description\n"
"    -----------------------------------------------------------------------\n"
"    %-22s A list of raw science images or\n"
"    %-22s A list of raw science images (extended)\n"
"    %-22s A list of offset sky exposures (optional)\n"
"    %-22s A master dark frame\n"
"    %-22s A master object mask (optional)\n"
"    %-22s A master twilight flat frame\n"
"    %-22s A master sky frame (optional)\n"
"    %-22s A channel table\n"
"    %-22s A photometric calibration table\n"
"    %-22s A readnoise/gain file\n"
"    %-22s A master confidence map\n"
"    %-22s A master standard star index\n"
"    %-22s Northern Schlegel Map\n"
"    %-22s Southern Schlegel Map\n"
"All of the above are required unless otherwise specified\n"
"\n";

/**@{*/

/**
    \ingroup recipelist
    \defgroup vircam_science_process vircam_science_process
    \brief Science reduction recipe for a full pawprint of VIRCAM data

    \par Name: 
        vircam_science_process
    \par Purpose: 
        Do a complete science reduction for a full pawprint of VIRCAM data
    \par Description: 
        A single pawprint of VIRCAM data in a single filter is corrected 
        for instrumental signature. The sky is removed using a defined
        algorithm and the corrected simple images are dithered into the final 
        stacked image. This stacked image is astrometrically and 
        photometrically calibrated using either locally held catalogues or
        those queried from the CDS.
    \par Language:
        C
    \par Parameters:
        General reduction parameters:
        - \b destripe (bool): If set then the images will be destriped.
        - \b skyalgo (string): The sky algorithm to be used
            - none: Don't do sky correction
            - master: Use a master sky. This must be included in the sof
            - pawsky_mask: pawsky_mask algorithm
            - pawsky_mask_pre: pawsky_mask with pre-existing object mask
            - tilesky: Used if a full tile is given
            - tilesky_minus: Used if a full tile is given and we excluded the current frame
            - auto: let the routine decide which routine to use
            - pawsky_minus: pawsky_minus algorithm
        - \b preview_only (bool): If set we only get a preview of the reduction
        - \b minphotom (int): The minimum number of standards for photometry
        - \b prettynames (bool): True if we're using nice file names
        - \b cdssearch (bool): True if we get 2MASS standards from CDS.
        - \b cacheloc (string): The location of the standards cache
        - \b magerrcut (float): A cut in the magnitude error for photometric
             standard stars
        
        Parameters for source detection on stacks:
        - \b stk_cat_ipix (int): The minimum allowable size of an object
        - \b stk_cat_thresh (float): The detection threshold in sigma above sky
        - \b stk_cat_icrowd (int): If set then the deblending software will be 
             used
        - \b stk_cat_rcore (float): The core radius in pixels
        - \b stk_cat_nbsize (int): The smoothing box size for background map 
             estimation
        
        Parameters for stacking algorithm:
        - \b stk_lthr (float): Low rejection threshold
        - \b stk_hthr (float): Upper rejection threshold
        - \b stk_method (string): The stacking method
            - nearest: Nearest neighbour
            - linear: Bi-linear interpolation
        - \b stk_seeing (bool): TRUE if we want to weight images by seeing
        - \b stk_fast (string): flag for fast or slow stacking algorithm.
             The fast algorithm is much greedier in memory.
            - auto: Let the recipe decide which we use
            - slow: Use the slow algorithm
            - fast: Use the fast algorithm
        - \b stk_nfst (int): The number of frames that can be stacked before
             we revert to the slow stacking algorithm when in auto mode

        Parameters for the destripe algorithm:
        - \b str_filt (bool): True if we filter the stripe pattern before applying

        Parameters for pawsky_mask algorithm:
        - \b psm_ipix (int): The minimum allowable size of an object
        - \b psm_niter (int): The number of iterations in the algorithm
        - \b psm_thresh (float): The detection threshold in sigma above sky
        - \b psm_nbsize (int): The smoothing box size for background map estimation
        - \b psm_smkern (float): The smoothing kernel in the detection
        
    \par Input File Types:
        The following list gives the file types that can appear in the SOF
        file. The word in bold is the DO category value.
        - \b OBJECT (required): The list of input science images.
        - \b OBJECT_EXTENDED (required): The list of input science images with extended
             sources.
        - \b SKY_OFFSET (optional): Offset sky observations if available
        - \b MASTER_DARK (required): A master dark frame.
        - \b MASTER_TWILIGHT_FLAT (required): A master twilight flat frame.
        - \b MASTER_SKY (optional): A master sky frame to be used for
             background correction. If this isn't available, then we
             look for offset skies. If these aren't available then we
             use the object images themselves.
        - \b MASTER_CHANNEL_TABLE (required): The relevant channel table
        - \b PHOTCAL_TAB (required): A photometric calibration table
        - \b MASTER_READGAIN_TABLE (required): A master read/gain table
        - \b MASTER_CONF (required): A master confidence map 
             for the filter used in the science images.
        - \b MASTER_2MASS_CATLAOGUE (required): A master 2MASS catalogue
             location (unless using cdssearch). 
        - \b MASTER_OBJMASK (optional): An object mask to be included if the
             pawsky_mask_pre or pawsky_minus algorithms are requested.
        - \b SCHLEGEL_MAP_NORTH (required): The Northern Schlegel dust map
        - \b SCHLEGEL_MAP_SOUTH (required): The Southern Schlegel dust map
    \par Output Products:
        The following list gives the output data products that are generated
        by this recipe. The word in bold gives the DPR CATG keyword value for
        each product:
        - Copies of the input science frames that have been corrected for
          instrumental signature and sky (\b SIMPLE_IMAGE_SCI)
        - Copies of the input offset sky frames that have been corrected for
          instrumental signature sky (\b SIMPLE_IMAGE_SKY)
        - A dithered stacked image and its associated confidence
          map (\b JITTERED_IMAGE_SCI and \b CONFIDENCE_MAP_SCI)
        - A catalogue of objects extracted from the stacked image 
          (\b OBJECT_CATALOGUE_SCI)
        - A mean offset sky frame, if input raw offset skies are given (\b OFFSET_SKY)
    \par Output QC Parameters:
        - \b SATURATION
             The saturation level in ADUs.
        - \b MEAN_SKY
             The mean level of the background sky over the image in ADUs.
        - \b SKY_NOISE
             The RMS of the background sky over the image in ADUs
        - \b IMAGE_SIZE
             The average size of stellar objects on the image in pixels
        - \b ELLIPTICITY
             The average ellipticity of stellar objects on the image
        - \b POSANG
             The average position angle in degrees from North towards East.
             NB: this value only makes sense if the ellipticity is significant
        - \b APERTURE_CORR
             The aperture correction for an aperture of radius rcore.
        - \b NOISE_OBJ
             The number of noise objects found in the image
        - \b MAGZPT
             The photometric zero point
        - \b MAGZERR
             The internal error in the photometric zero point
        - \b MAGNZPT
             The number of stars used to determine the magnitude zero point
        - \b MAGNCUT
             The number of stars cut from magnitude zero point calculation
        - \b SKYBRIGHT
             The sky brightness in mag/arcsec**2
        - \b LIMITING_MAG
             The limiting magnitude for this image for a 5 sigma detection.
        - \b WCS_DCRVAL1
             The offset of the equatorial coordinate represented by CRVAL1 
             from the raw frame to the reduced one (degrees).
        - \b WCS_DCRVAL2
             The offset of the equatorial coordinate represented by CRVAL2
             from the raw frame to the reduced one (degrees).
        - \b WCS_DTHETA
             The change in the WCS coordinate system rotation from the raw
             to the reduced frame (degrees)
        - \b WCS_SCALE
             The scale of the pixels in the reduced frames in arcseconds per
             pixel
        - \b WCS_SHEAR
             The shear of the astrometric solution in the form of the 
             difference between the rotation of the x solution and the rotation
             of the y solution (abs(xrot) - abs(yrot) in degrees)
        - \b WCS_RMS
             The average error in the WCS fit (arcsec)
    \par Notes
        None.
    \par Fatal Error Conditions:
        - NULL input frameset
        - Input frameset headers incorrect meaning that RAW and CALIB frame
          can't be distinguished
        - No science frames in the input frameset
        - Required master calibration images and tables are missing
        - Unable to save data products
    \par Non-Fatal Error Conditions:
        - None
    \par Conditions Leading To Dummy Products:
        - Master calibration images either won't load or are flagged as dummy
        - The detector for the current image extension is flagged dead in
          all science frames
        - Various processing routines fail.
    \par Author:
        Jim Lewis, CASU
    \par Code Reference: 
        vircam_science_process.c
*/

/* Function code */

/*---------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module.
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*---------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist *list) {
    cpl_recipe  *recipe = cpl_calloc(1,sizeof(*recipe));
    cpl_plugin  *plugin = &recipe->interface;
    char alldesc[SZ_ALLDESC];
    (void)snprintf(alldesc,SZ_ALLDESC,
                   vircam_science_process_description,
                   VIRCAM_SCI_OBJECT_RAW,VIRCAM_SCI_OBJECT_EXT_RAW,
                   VIRCAM_OFFSET_SKY_RAW,VIRCAM_CAL_DARK,VIRCAM_CAL_OBJMASK,
                   VIRCAM_CAL_TWILIGHT_FLAT,VIRCAM_CAL_SKY,VIRCAM_CAL_CHANTAB,
                   VIRCAM_CAL_PHOTTAB,VIRCAM_CAL_READGAINFILE,VIRCAM_CAL_CONF,
                   VIRCAM_CAL_2MASS,VIRCAM_CAL_SCHL_N,VIRCAM_CAL_SCHL_S);

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    VIRCAM_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "vircam_science_process",
                    "VIRCAM science recipe",
                    alldesc,
                    "Jim Lewis",
                    "jrl@ast.cam.ac.uk",
                    vircam_get_license(),
                    vircam_science_process_create,
                    vircam_science_process_exec,
                    vircam_science_process_destroy);

    cpl_pluginlist_append(list,plugin);

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the
  interface.
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_process_create(cpl_plugin *plugin) {
    cpl_recipe      *recipe;
    cpl_parameter   *p;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    /* Create the parameters list in the cpl_recipe object */

    recipe->parameters = cpl_parameterlist_new();

    /* Fill in flag to destripe the images */

    p = cpl_parameter_new_value("vircam.vircam_science_process.destripe",
                                CPL_TYPE_BOOL,"Destripe images?",
                                "vircam.vircam_science_process",TRUE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"destripe");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to correct sky background */

    p = cpl_parameter_new_enum("vircam.vircam_science_process.skyalgo",
                               CPL_TYPE_STRING,"Sky subtraction algorithm",
                               "vircam.vircam_science_process","auto",8,
                               "master","none","pawsky_mask","pawsky_mask_pre",
                               "tilesky","tilesky_minus","auto","pawsky_minus");
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"skyalgo");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to just print out how the data will be processed and 
       then exit */

    p = cpl_parameter_new_value("vircam.vircam_science_process.preview_only",
                                CPL_TYPE_BOOL,"Preview only?",
                                "vircam.vircam_science_process",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"preview_only");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to just print out how the data will be processed and 
       then exit */

    p = cpl_parameter_new_range("vircam.vircam_science_process.minphotom",
                                CPL_TYPE_INT,
                                "Minimum stars for photometry solution",
                                "vircam.vircam_science_process",1,1,100000);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"minphotom");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to decide whether to use predictable names or nice
       names based on input file names */

    p = cpl_parameter_new_value("vircam.vircam_science_process.prettynames",
                                CPL_TYPE_BOOL,"Use pretty product names?",
                                "vircam.vircam_science_process",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"prettynames");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to decide whether to use CDS to get 2MASS standards */

    p = cpl_parameter_new_value("vircam.vircam_science_process.cdssearch",
                                CPL_TYPE_BOOL,"Use CDS for 2MASS standards?",
                                "vircam.vircam_science_process",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cdssearch");
    cpl_parameterlist_append(recipe->parameters,p);

    /* ***Special parameters for object detection on stacks*** */

    /* Fill in the minimum object size */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_cat_ipix",
                                CPL_TYPE_INT,
                                "Minimum pixel area for each detected object",
                                "vircam.vircam_science_process",5,1,
                                1024);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_cat_ipix");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in the detection threshold parameter */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_cat_thresh",
                                CPL_TYPE_DOUBLE,
                                "Detection threshold in sigma above sky",
                                "vircam.vircam_science_process",2.0,1.0e-6,
                                1.0e10);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_cat_thresh");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in flag to use deblending software or not */

    p = cpl_parameter_new_value("vircam.vircam_science_process.stk_cat_icrowd",
                                CPL_TYPE_BOOL,"Use deblending?",
                                "vircam.vircam_science_process",TRUE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_cat_icrowd");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in core radius */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_cat_rcore",
                                CPL_TYPE_DOUBLE,"Value of Rcore in pixels",
                                "vircam.vircam_science_process",3.0,1.0e-6,
                                1024.0);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_cat_rcore");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Fill in background smoothing box size */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_cat_nbsize",
                                CPL_TYPE_INT,"Background smoothing box size",
                                "vircam.vircam_science_process",64,1,2048);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_cat_nbsize");
    cpl_parameterlist_append(recipe->parameters,p);

    /* ***Special parameters for stacking*** */

    /* The lower rejection threshold to be used during stacking */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_lthr",
                                CPL_TYPE_DOUBLE,"Low rejection threshold",
                                "vircam.vircam_science_process",5.0,1.0e-6,
                                1.0e10);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_lthr");
    cpl_parameterlist_append(recipe->parameters,p);

    /* The upper rejection threshold to be used during stacking */

    p = cpl_parameter_new_range("vircam.vircam_science_process.stk_hthr",
                                CPL_TYPE_DOUBLE,"Upper rejection threshold",
                                "vircam.vircam_science_process",5.0,1.0e-6,
                                1.0e10);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_hthr");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Stacking method */

    p = cpl_parameter_new_enum("vircam.vircam_science_process.stk_method",
                               CPL_TYPE_STRING,"Stacking method",
                               "vircam.vircam_science_process","linear",2,
                               "nearest","linear");
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_method");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Flag to use seeing weighting */

    p = cpl_parameter_new_value("vircam.vircam_science_process.stk_seeing",
                                CPL_TYPE_BOOL,"Weight by seeing?",
                                "vircam.vircam_science_process",FALSE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_seeing");
    cpl_parameterlist_append(recipe->parameters,p);
    
    /* Stacking speed method */

    p = cpl_parameter_new_enum("vircam.vircam_science_process.stk_fast",
                               CPL_TYPE_STRING,
                               "Use fast stacking?",
                               "vircam.vircam_science_process",
                               "auto",3,"fast","slow","auto");
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_fast");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Number of frames to stack before we revert to the slow algorithm
       when in auto mode */

    p = cpl_parameter_new_value("vircam.vircam_science_process.stk_nfst",
                                CPL_TYPE_INT,
                                "Nframes to stack in fast mode",
                                "vircam.vircam_science_process",30);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"stk_nfst");
    cpl_parameterlist_append(recipe->parameters,p);
    
    /* ***Special parameters for destripe algorithm*** */

    /* Flag to filter stripe pattern */

    p = cpl_parameter_new_value("vircam.vircam_science_process.str_filt",
                                CPL_TYPE_BOOL,"Filter stripe pattern?",
                                "vircam.vircam_science_process",TRUE);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"str_filt");
    cpl_parameterlist_append(recipe->parameters,p);

    /* ***Special parameters for pawsky_mask algorithm*** */

    /* The pixel area for detected objects */

    p = cpl_parameter_new_range("vircam.vircam_science_process.psm_ipix",
                                CPL_TYPE_INT,
                                "Minimum pixel area for each detected object",
                                "vircam.vircam_science_process",5,1,100000);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"psm_ipix");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Number of iterations */

    p = cpl_parameter_new_range("vircam.vircam_science_process.psm_niter",
                                CPL_TYPE_INT,
                                "Number of iterations in pawsky mask",
                                "vircam.vircam_science_process",5,1,100);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"psm_niter");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Detection threshold */

    p = cpl_parameter_new_range("vircam.vircam_science_process.psm_thresh",
                                CPL_TYPE_DOUBLE,
                                "Detection threshold in sigma above sky",
                                "vircam.vircam_science_process",2.0,1.0e-6,
                                1.0e10);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"psm_thresh");
    cpl_parameterlist_append(recipe->parameters,p);
        
    /* Background smoothing box size */

    p = cpl_parameter_new_range("vircam.vircam_science_process.psm_nbsize",
                                CPL_TYPE_INT,"Background smoothing box size",
                                "vircam.vircam_science_process",50,1,2048);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"psm_nbsize");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Smoothing kernel size */

    p = cpl_parameter_new_range("vircam.vircam_science_process.psm_smkern",
                                CPL_TYPE_DOUBLE,
                                "Smoothing kernel size (pixels)",
                                "vircam.vircam_science_process",2.0,1.0e-6,5.0);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"psm_smkern");
    cpl_parameterlist_append(recipe->parameters,p);
        
    /* The standard star cache location */
    
    p = cpl_parameter_new_value("vircam.vircam_science_process.cacheloc",
                                CPL_TYPE_STRING,
                                "Location of standard star cache?",
                                "vircam.vircam_science_process",".");
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"cacheloc");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Magnitude error cut */
    
    p = cpl_parameter_new_value("vircam.vircam_science_process.magerrcut",
                                CPL_TYPE_DOUBLE,
                                "Magnitude error cut",
                                "vircam.vircam_science_process",100.0);
    cpl_parameter_set_alias(p,CPL_PARAMETER_MODE_CLI,"magerrcut");
    cpl_parameterlist_append(recipe->parameters,p);

    /* Get out of here */

    return(0);
}
    
/*---------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_process_exec(cpl_plugin *plugin) {
    cpl_recipe  *recipe;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    return(vircam_science_process(recipe->parameters,recipe->frames));
}
                                
/*---------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_process_destroy(cpl_plugin *plugin) {
    cpl_recipe *recipe ;

    /* Get the recipe out of the plugin */

    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else 
        return(-1);

    cpl_parameterlist_delete(recipe->parameters);
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*---------------------------------------------------------------------------*/

static int vircam_science_process(cpl_parameterlist *parlist,
                                  cpl_frameset *framelist) {
    const char *fctid="vircam_science_process";
    cpl_parameter *p;
    int nfail,status,i,ndit,j,live,isfirst,nsci,njsteps,n;
    cpl_size nlab;
    float gaincor_fac,readnoise,gain,dit;
    cpl_frame *catindex,*template,*frm;
    cpl_propertylist *pp;
    cpl_table *stdscat,*tab;
    casu_fits *ff;
    char filt[16],projid[16],*tname,fname[BUFSIZ];

    /* Check validity of the input frameset */

    if (framelist == NULL || cpl_frameset_get_size(framelist) <= 0) {
        cpl_msg_error(fctid,"Input framelist NULL or has no input data");
        return(-1);
    }

    /* Initialise some things */

    vircam_sci_init();
    (void)strncpy(vircam_recipename,fctid,VIRCAM_PATHSZ);
    (void)snprintf(vircam_recipepaf,VIRCAM_PATHSZ,"VIRCAM/%s",fctid);

    /* Get the parameters */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.destripe");
    vircam_sci_config.destripe = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.str_filt");
    vircam_sci_config.str_filt = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.skyalgo");
    if (! strcmp(cpl_parameter_get_string(p),"none"))
        vircam_sci_config.skyalgo = SKYNONE;
    else if (! strcmp(cpl_parameter_get_string(p),"master"))
        vircam_sci_config.skyalgo = SKYMASTER;
    else if (! strcmp(cpl_parameter_get_string(p),"pawsky_mask"))
        vircam_sci_config.skyalgo = PAWSKY_MASK;
    else if (! strcmp(cpl_parameter_get_string(p),"pawsky_mask_pre"))
        vircam_sci_config.skyalgo = PAWSKY_MASK_PRE;
    else if (! strcmp(cpl_parameter_get_string(p),"tilesky"))
        vircam_sci_config.skyalgo = TILESKY;
    else if (! strcmp(cpl_parameter_get_string(p),"tilesky_minus"))
        vircam_sci_config.skyalgo = TILESKY_MINUS;
    else if (! strcmp(cpl_parameter_get_string(p),"auto"))
        vircam_sci_config.skyalgo = AUTO;
    else if (! strcmp(cpl_parameter_get_string(p),"pawsky_minus"))
        vircam_sci_config.skyalgo = PAWSKY_MINUS;    
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.preview_only");
    vircam_sci_config.preview_only = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.minphotom");
    vircam_sci_config.minphotom = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.prettynames");
    vircam_sci_config.prettynames = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.cdssearch");
    vircam_sci_config.cdssearch = (cpl_parameter_get_bool(p) ? 1 : 0);

    /* Stacking special parameters */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_lthr");
    stk.lthr = cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_hthr");
    stk.hthr = cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_method");
    stk.method = (! strcmp(cpl_parameter_get_string(p),"nearest") ? 0 : 1);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_seeing");
    stk.seeing = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_fast");
    if (! strcmp(cpl_parameter_get_string(p),"auto"))
        stk.fast = -1;
    else if (! strcmp(cpl_parameter_get_string(p),"slow"))
        stk.fast = 0;
    else if (! strcmp(cpl_parameter_get_string(p),"fast"))
        stk.fast = 1;
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_nfst");
    stk.nfst = cpl_parameter_get_int(p);

    /* Object detection on stacks special parameters */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_cat_ipix");
    stk_cat.ipix = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_cat_thresh");
    stk_cat.thresh = (float)cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_cat_icrowd");
    stk_cat.icrowd = cpl_parameter_get_bool(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_cat_rcore");
    stk_cat.rcore = (float)cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.stk_cat_nbsize");
    stk_cat.nbsize = cpl_parameter_get_int(p);

    /* Pawsky_mask special parameters */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.psm_ipix");
    psm.ipix = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.psm_niter");
    psm.niter = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.psm_thresh");
    psm.thresh = cpl_parameter_get_double(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.psm_nbsize");
    psm.nbsize = cpl_parameter_get_int(p);
    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.psm_smkern");
    psm.smkern = cpl_parameter_get_double(p);

    /* Cache location */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.cacheloc");
    vircam_sci_config.cacheloc = (char *)cpl_parameter_get_string(p);

    /* Magnitude error cut */

    p = cpl_parameterlist_find(parlist,
                               "vircam.vircam_science_process.magerrcut");
    vircam_sci_config.magerrcut = (float)cpl_parameter_get_double(p);

    /* Sort out raw from calib frames */

    if (vircam_dfs_set_groups(framelist) != CASU_OK) {
        cpl_msg_error(fctid,"Cannot identify RAW and CALIB frames");
        vircam_sci_tidy(0);
        return(-1);
    }

    /* Label the input frames */

    if ((ps.labels = cpl_frameset_labelise(framelist,casu_compare_tags,
                                           &nlab)) == NULL) {
        cpl_msg_error(fctid,"Cannot labelise the input frames");
        vircam_sci_tidy(0);
        return(-1);
    }

    /* Get the input science frames */

    nfail = 0;
    if ((ps.science_frames = 
         casu_frameset_subgroup(framelist,ps.labels,nlab,
                                VIRCAM_SCI_OBJECT_RAW)) == NULL) {
        if ((ps.science_frames =
             casu_frameset_subgroup(framelist,ps.labels,nlab,
                                    VIRCAM_SCI_OBJECT_EXT_RAW)) == NULL) {
            cpl_msg_error(fctid,"No science images to process!");
            vircam_sci_tidy(0);
            return(-1);
        }
    }
    /* nfail += vircam_sci_testfrms(ps.science_frames,VIRCAM_NEXTN,1); */

    /* Get the offset sky frames */
    
    if ((ps.master_sky = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                                  VIRCAM_CAL_SKY)) == NULL) {   
        if ((ps.offset_skies = 
             casu_frameset_subgroup(framelist,ps.labels,nlab,
                                    VIRCAM_OFFSET_SKY_RAW)) == NULL) {
            offsky = 0;
        } else {
            offsky = 1;
            nfail += vircam_sci_testfrms(ps.offset_skies,VIRCAM_NEXTN,1);
        }
    } else {
        offsky = -1;
        vircam_sci_config.skyalgo = -1;
        nfail += vircam_sci_testfrm_1(ps.master_sky,VIRCAM_NEXTN,1);
    }

    /* Check to see if there is a master dark frame */

    if ((ps.master_dark = 
         casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                  VIRCAM_CAL_DARK)) == NULL) {
        cpl_msg_error(fctid,"No master dark found");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.master_dark,VIRCAM_NEXTN,1);
        
    /* Check to see if there is a master twilight flat frame */

    if ((ps.master_twilight_flat = 
         casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                  VIRCAM_CAL_TWILIGHT_FLAT)) == NULL) {
        cpl_msg_error(fctid,"No master twilight flat found");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.master_twilight_flat,VIRCAM_NEXTN,1);

    /* Get the master object mask. Does it make sense to have one in
       the context of the sky correction algorithm? */
    
    ps.master_objmask = casu_frameset_subgroup_1(framelist,ps.labels,
                                                 nlab,VIRCAM_CAL_OBJMASK);
    if ((ps.master_objmask == NULL) && 
        (vircam_sci_config.skyalgo == PAWSKY_MASK_PRE)) {
        cpl_msg_error(fctid,
                      "No object mask found. Unable to use requested sky algorithm");
        vircam_sci_tidy(0);
        return(-1);
    } else if (ps.master_objmask != NULL && 
               (vircam_sci_config.skyalgo == PAWSKY_MASK ||
                vircam_sci_config.skyalgo == TILESKY)) {
        cpl_msg_info(fctid,
                     "Requested sky subt algorithm requires no object mask");
    }
    nfail += vircam_sci_testfrm_1(ps.master_objmask,0,1);

    /* Get the gain corrections */

    status = CASU_OK;
    if (casu_gaincor_calc(ps.master_twilight_flat,&i,&(ps.gaincors),
                          &status) != CASU_OK) {
        cpl_msg_error(fctid,"Error calculating gain corrections");
        vircam_sci_tidy(0);
        return(-1);
    }

    /* Check to see if there is a readgain file */

    if ((ps.readgain_file = 
         casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                  VIRCAM_CAL_READGAINFILE)) == NULL) {
        cpl_msg_error(fctid,"No master readnoise/gain file found");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.readgain_file,VIRCAM_NEXTN,0);
        
    /* Check to see if there is a master confidence map. */

    if ((ps.master_conf = 
         casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                  VIRCAM_CAL_CONF)) == NULL) {
        cpl_msg_error(fctid,"No master confidence map file found");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.master_conf,VIRCAM_NEXTN,1);
    ps.mask = casu_mask_define(framelist,ps.labels,nlab,VIRCAM_CAL_CONF,
                               VIRCAM_CAL_BPM);
        
    /* Check to see if there is a channel table */

    if ((ps.chantab = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                               VIRCAM_CAL_CHANTAB)) == NULL) {
        cpl_msg_error(fctid,"No channel table found");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.chantab,VIRCAM_NEXTN,0);

    /* Check to see if there is a photometric table */

    if ((ps.phottab = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                               VIRCAM_CAL_PHOTTAB)) == NULL) {
        cpl_msg_error(fctid,"No photometric table found");
        vircam_sci_tidy(0);
        return(-1);
    }
    n = cpl_frame_get_nextensions(ps.phottab);
    nfail += vircam_sci_testfrm_1(ps.phottab,n,0);

    /* Is the 2mass index file specified? */

    catindex = NULL;
    if (vircam_sci_config.cdssearch == 0) {
        if ((catindex = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                                 VIRCAM_CAL_2MASS)) == NULL) {
            cpl_msg_error(fctid,"No 2MASS index found -- cannot continue");
            vircam_sci_tidy(0);
            return(-1);
        }
        nfail += vircam_sci_testfrm_1(catindex,1,0);
    }

    /* Is the Northern Schlegel file specified? */

    if ((ps.schlf_n = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                               VIRCAM_CAL_SCHL_N)) == NULL) {
        cpl_msg_error(fctid,"Schlegel North map not found -- cannot continue");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.schlf_n,0,1);


    /* Is the Southern Schlegel file specified? */

    if ((ps.schlf_s = casu_frameset_subgroup_1(framelist,ps.labels,nlab,
                                               VIRCAM_CAL_SCHL_S)) == NULL) {
        cpl_msg_error(fctid,"Schlegel South map not found -- cannot continue");
        vircam_sci_tidy(0);
        return(-1);
    }
    nfail += vircam_sci_testfrm_1(ps.schlf_s,0,1);

    /* Check the cache location is writable */

    if (access(vircam_sci_config.cacheloc,R_OK+W_OK+X_OK) != 0) {
        cpl_msg_error(fctid,"Cache location %s inacessible",
                      vircam_sci_config.cacheloc);
        nfail++;
    }

    /* Ok if any of this failed, then get out of here. This makes it a bit 
       simpler later on since we now know that all of the specified files 
       have the correct number of extensions and will load properly */

    if (nfail > 0) {
        cpl_msg_error(fctid,
                      "There are %" CPL_SIZE_FORMAT " input file errors -- cannot continue",
                      (cpl_size)nfail);
        vircam_sci_tidy(0);
        return(-1);
    }
    
    /* Get catalogue parameters */

    if (catindex != NULL) {
        if (casu_catpars(catindex,&(ps.catpath),&(ps.catname)) == CASU_FATAL) {
            vircam_sci_tidy(0);
            cpl_frame_delete(catindex);
            return(-1);
        }
        cpl_frame_delete(catindex);

        /* Set up the photometric calibration catalogue data */

        (void)strcpy(current_catpath,ps.catpath);
        (void)strcpy(current_cat,ps.catname);
    } else {
        ps.catpath = NULL;
        ps.catname = NULL;
    }

    /* Get the number of DITs. This assumes it is the same for all the science
       images (and it damn well better be!) */

    pp = cpl_propertylist_load(cpl_frame_get_filename(cpl_frameset_get_position(ps.science_frames,0)),0);
    if (vircam_pfits_get_ndit(pp,&ndit) != CASU_OK) {
        cpl_msg_error(fctid,"No value for NDIT available");
        freepropertylist(pp);
        vircam_sci_tidy(0);
        return(-1);
    }
    cpl_propertylist_delete(pp);

    /* Load the photometric table */

    ps.tphottab = cpl_table_load(cpl_frame_get_filename(ps.phottab),1,0);

    /* Break these into pawprints now */

    vircam_sci_paws_create(ps.science_frames,&(ps.nscipaws),&(ps.science_paws));
    vircam_sci_paws_create(ps.offset_skies,&(ps.noffpaws),&(ps.offsky_paws));

    /* Check what we have and see if it makes sense. If this isn't a tile
       and there are multiple pawprints, then signal an error */

    if (ps.science_paws[0].tilenum == -1 && ps.nscipaws > 1) {
        cpl_msg_error(fctid,
                      "List has %" CPL_SIZE_FORMAT " pawprints. Only 1 is allowed for non tiles",
                      (cpl_size)ps.nscipaws);
        vircam_sci_tidy(0);
        return(-1);
    }

    /* Define the skys that will be made */

    vircam_sci_skydefine();

    /* Print out some stuff if you want a preview */

    if (vircam_sci_config.preview_only) {
        if (ps.science_paws[0].tilenum == -1) 
            fprintf(stdout,"This is a paw group with %" CPL_SIZE_FORMAT " files\n",
                    cpl_frameset_get_size(ps.science_frames));
        else
            fprintf(stdout,"This is a tile group with %" CPL_SIZE_FORMAT " files\n",
                    cpl_frameset_get_size(ps.science_frames));
        frm = cpl_frameset_get_position(ps.science_frames,0);
        pp = cpl_propertylist_load(cpl_frame_get_filename(frm),0);
        (void)vircam_pfits_get_filter(pp,filt);
        (void)vircam_pfits_get_ndit(pp,&ndit);
        (void)vircam_pfits_get_dit(pp,&dit);
        (void)vircam_pfits_get_projid(pp,projid);
        (void)vircam_pfits_get_njsteps(pp,&njsteps);
        cpl_propertylist_delete(pp);

        fprintf(stdout,"Filter %8s, DIT: %8.1f, NDIT: %" CPL_SIZE_FORMAT ", Project: %s\n",filt,
                dit,(cpl_size)ndit,projid);
        fprintf(stdout,"Dark: %s\n",cpl_frame_get_filename(ps.master_dark));
        fprintf(stdout,"Flat: %s\n",
                cpl_frame_get_filename(ps.master_twilight_flat));
        fprintf(stdout,"CPM: %s\n",cpl_frame_get_filename(ps.master_conf));
        fprintf(stdout,"Channel table: %s\n",
                cpl_frame_get_filename(ps.chantab));
        fprintf(stdout,"A %" CPL_SIZE_FORMAT " point dither will be done\n",
                (cpl_size)njsteps);
        if (ps.science_paws[0].tilenum != -1) 
            fprintf(stdout,"The tile has %" CPL_SIZE_FORMAT " pawprints\n",
                    (cpl_size)ps.nscipaws);
        for (i = 0; i < ps.nscipaws; i++) {
            n = cpl_frameset_get_size(ps.science_paws[i].orig);
            for (j = 0; j < n; j++) {
                frm = cpl_frameset_get_position(ps.science_paws[i].orig,j);
                fprintf(stdout,"%s     sky%" CPL_SIZE_FORMAT "\n",
                        cpl_frame_get_filename(frm),
                        (cpl_size)(ps.science_paws[i].whichsky[j]+1));
            }
        }
        if (ps.noffpaws != 0) {
            fprintf(stdout,"There are %" CPL_SIZE_FORMAT " offset sky frames\n",
                    cpl_frameset_get_size(ps.offset_skies));
            for (i = 0; i < ps.noffpaws; i++) {
                n = cpl_frameset_get_size(ps.offsky_paws[i].orig);
                for (j = 0; j < n; j++) {
                    frm = cpl_frameset_get_position(ps.offsky_paws[i].orig,j);
                    fprintf(stdout,"%s     sky%" CPL_SIZE_FORMAT "\n",cpl_frame_get_filename(frm),
                            (cpl_size)(ps.offsky_paws[i].whichsky[j]+1));
                }
            }
        }
        fprintf(stdout,"\nRequested sky algorithm: %" CPL_SIZE_FORMAT "\n",
                (cpl_size)vircam_sci_config.skyalgo);
        for (i = 0; i < ps.nskys; i++) {
            fprintf(stdout,"Sky%" CPL_SIZE_FORMAT ": %" CPL_SIZE_FORMAT " ",
                    (cpl_size)(i+1),(cpl_size)ps.skys[i].skyalgo);
            if (ps.skys[i].objmask != NULL)
                fprintf(stdout,"mask: %s\nFormed from:\n",
                        cpl_frame_get_filename(ps.skys[i].objmask));
            else
                fprintf(stdout,"\nFormed from:\n");
            for (j = 0; j < cpl_frameset_get_size(ps.skys[i].contrib); j++) {
                frm = cpl_frameset_get_position(ps.skys[i].contrib,j);
                fprintf(stdout,"    %s\n",cpl_frame_get_filename(frm));
            }
        }
        vircam_sci_tidy(0);
        return(0);
    }

    /* Get some workspace for the product frames */

    ps.nscience = cpl_frameset_get_size(ps.science_frames);
    nsci = ps.nscience;
    ps.product_frames_simple = cpl_malloc(ps.nscience*sizeof(cpl_frame *));
    for (i = 0; i < ps.nscience; i++)
        ps.product_frames_simple[i] = NULL;
    if (ps.offset_skies != NULL) {
        ps.noffsets = cpl_frameset_get_size(ps.offset_skies);
        ps.product_frames_simple_off = 
            cpl_malloc(ps.noffsets*sizeof(cpl_frame *));
        for (i = 0; i < ps.noffsets; i++)
            ps.product_frames_simple_off[i] = NULL;
    } else {
        ps.noffsets = 0;
    }

    /* Do the stage1 processing. Loop for all extensions */

    cpl_msg_info(fctid,"Beginning stage1 reduction");
    cpl_msg_indent_more();
    for (j = 1; j <= VIRCAM_NEXTN; j++) {
        isfirst = (j == 1);
        gaincor_fac = (ps.gaincors)[j-1];
        cpl_msg_info(fctid,"Extension [%" CPL_SIZE_FORMAT "]",(cpl_size)j);
        
        /* Load up the calibration frames. We know these won't fail now
           so we can skip all the boring testing... */
        
        ps.fdark = casu_fits_load(ps.master_dark,CPL_TYPE_FLOAT,j);
        ps.fflat = casu_fits_load(ps.master_twilight_flat,CPL_TYPE_FLOAT,j);
        ps.fconf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,j);
        casu_mask_load(ps.mask,j,
                         (int)cpl_image_get_size_x(casu_fits_get_image(ps.fconf)),
                         (int)cpl_image_get_size_y(casu_fits_get_image(ps.fconf)));
        if (offsky == -1) 
            ps.fsky = casu_fits_load(ps.master_sky,CPL_TYPE_FLOAT,j);
        ps.fchantab = casu_tfits_load(ps.chantab,j);

        /* Now load up the science frames and offset skies, if needed */
        
        ps.nscience = cpl_frameset_get_size(ps.science_frames);
        ps.sci_fits = casu_fits_load_list(ps.science_frames,CPL_TYPE_FLOAT,j);
        if (j == 1) 
            ps.phupaf = vircam_paf_phu_items(casu_fits_get_phu(ps.sci_fits[0]));
        if (offsky == 1) {
            ps.noffsets = cpl_frameset_get_size(ps.offset_skies);
            ps.offsky_fits = casu_fits_load_list(ps.offset_skies,
                                                   CPL_TYPE_FLOAT,j);
        } else {
            ps.noffsets = 0;
            ps.offsky_fits = NULL;
        }
           
        /* Loop through and mark the frames where the header says the detector
           wasn't live. Also check to see if the CRVAL == 0 problem is 
           present. Check for live detectors in the offset skies too */

        for (i = 0; i < ps.nscience; i++) {
            ff = ps.sci_fits[i];
            vircam_pfits_get_detlive(casu_fits_get_ehu(ff),&live);
            if (! live) 
                casu_fits_set_error(ff,CASU_FATAL);
            if (vircam_check_crval(casu_fits_get_phu(ff),
                                   casu_fits_get_ehu(ff)) != CASU_OK) {
                cpl_msg_error(fctid,"Unable to correct CRVAL in %s",
                              casu_fits_get_fullname(ff));
                vircam_sci_tidy(0);
                return(-1);
            }
        }
        for (i = 0; i < ps.noffsets; i++) {
            ff = ps.offsky_fits[i];
            vircam_pfits_get_detlive(casu_fits_get_ehu(ff),&live);
            if (! live) 
                casu_fits_set_error(ff,CASU_FATAL);
        }

        /* Do the corrections now. First for the object frames. Write
           out intermediate versions as we go */
       
        vircam_sci_get_readnoise_gain(j,&readnoise,&gain);
        for (i = 0; i < ps.nscience; i++) {
            ff = ps.sci_fits[i];
            cpl_propertylist_update_float(casu_fits_get_ehu(ff),"READNOIS",
                                          readnoise);
            cpl_propertylist_set_comment(casu_fits_get_ehu(ff),"READNOIS",
                                         "[e-] Readnoise used in processing");
            cpl_propertylist_update_float(casu_fits_get_ehu(ff),"GAIN",gain);
            cpl_propertylist_set_comment(casu_fits_get_ehu(ff),"GAIN",
                                         "[e-/adu] Gain used in processing");
            if (casu_fits_get_status(ff) == CASU_FATAL) {
                cpl_msg_info(fctid,"Detector is flagged dead in %s",
                             casu_fits_get_fullname(ff));
            } else {
                status = CASU_OK;
                (void)casu_darkcor(ff,ps.fdark,1.0,&status);
                (void)vircam_lincor(ff,ps.fchantab,1,ndit,&status);
                (void)casu_nditcor(ff,ndit,"EXPTIME",&status);
                (void)casu_flatcor(ff,ps.fflat,&status);
                (void)casu_gaincor(ff,gaincor_fac,&status);
                casu_fits_set_error(ff,status);
            }
            frm = cpl_frameset_get_position(ps.science_frames,i);
            tname = (char *)cpl_frame_get_filename(frm);
            vircam_sci_product_name(tname,SIMPLE_FILE,
                                    vircam_sci_config.prettynames,i+1,fname);
            vircam_sci_save_simple(ff,framelist,parlist,0,
                                   cpl_frameset_get_position(ps.science_frames,i),
                                   isfirst,VIRCAM_PRO_SIMPLE_SCI,fname,
                                   ps.product_frames_simple+i);
        }
        for (i = 0; i < ps.noffsets; i++) {
            ff = ps.offsky_fits[i];
            if (casu_fits_get_status(ff) == CASU_FATAL) {
                cpl_msg_info(fctid,"Detector is flagged dead in %s",
                             casu_fits_get_fullname(ff));
            } else {
                status = CASU_OK;
                (void)casu_darkcor(ff,ps.fdark,1.0,&status);
                (void)vircam_lincor(ff,ps.fchantab,1,ndit,&status);
                (void)casu_nditcor(ff,ndit,"EXPTIME",&status);
                (void)casu_flatcor(ff,ps.fflat,&status);
                (void)casu_gaincor(ff,gaincor_fac,&status);
                casu_fits_set_error(ff,status);
            }
            frm = cpl_frameset_get_position(ps.offset_skies,i);
            tname = (char *)cpl_frame_get_filename(frm);
            vircam_sci_product_name(tname,SIMPLE_FILE,
                                    vircam_sci_config.prettynames,
                                    ps.nscience+i+1,fname);
            vircam_sci_save_simple(ff,framelist,parlist,0,
                                   cpl_frameset_get_position(ps.offset_skies,i),
                                   isfirst,VIRCAM_PRO_SIMPLE_SKY,fname,
                                   ps.product_frames_simple_off+i);
        }

        /* Intermediate tidy */

        vircam_sci_tidy(1);
    }
    for (i = 0; i < nsci; i++) 
        freeframe(ps.product_frames_simple[i]);
    for (i = 0; i < ps.noffsets; i++) 
        freeframe(ps.product_frames_simple_off[i]);

    cpl_msg_indent_less();

    /* Form the sky frames. These are saved and registered as products */

    cpl_msg_info(fctid,"Making skies");
    for (i = 0; i < ps.nskys; i++) {
        if (vircam_sci_makesky(framelist,parlist,&(ps.skys[i]),
                               VIRCAM_PRO_SIMPLE_SCI) != CASU_OK) {
            cpl_msg_error(fctid,"Error making sky %" CPL_SIZE_FORMAT "",
                          (cpl_size)i);
            vircam_sci_tidy(0);
            return(-1);
        }
    }

    /* Do sky subtraction, destripe and write them out */

    cpl_msg_info(fctid,"Beginning stage2 reduction");
    (void)vircam_sci_sky_stripe_wcs(ps.nscipaws,ps.science_paws,
                                    VIRCAM_PRO_SIMPLE_SCI,framelist,parlist,
                                    vircam_sci_config.destripe);
    (void)vircam_sci_sky_stripe_wcs(ps.noffpaws,ps.offsky_paws,
                                    VIRCAM_PRO_SIMPLE_SKY,framelist,parlist,
                                    vircam_sci_config.destripe);

    /* Loop for all science jitters. Jitter, WCSFIT, photcal and write out */

    cpl_msg_info(fctid,"Beginning group processing");
    cpl_msg_indent_more();
    for (i = 0; i < ps.nscipaws; i++) {
        cpl_msg_info(fctid,"Working on paw%" CPL_SIZE_FORMAT "",
                     (cpl_size)(i+1));
        cpl_msg_indent_more();
        for (j = 1; j <= VIRCAM_NEXTN; j++) {
            cpl_msg_info(fctid,
                         "Stacking paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_stack(&(ps.science_paws[i]),j);
            cpl_msg_info(fctid,
                         "Source extraction paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_cat(&(ps.science_paws[i]),j);
            cpl_msg_info(fctid,
                         "WCS fit paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_wcsfit(&(ps.science_paws[i].stack[j-1]),
                              &(ps.science_paws[i].stackc[j-1]),
                              &(ps.science_paws[i].cat[j-1]),1,2);
            vircam_copywcs(casu_fits_get_ehu(ps.science_paws[i].stack[j-1]),
                           casu_fits_get_ehu(ps.science_paws[i].stackc[j-1]));
            (void)casu_getstds(casu_fits_get_ehu(ps.science_paws[i].stack[j-1]),
                               1,current_catpath,current_cat,
                               vircam_sci_config.cdssearch,
                               vircam_sci_config.cacheloc,&stdscat,
                               &status);
            (void)casu_matchstds(casu_tfits_get_table(ps.science_paws[i].cat[j-1]),
                                 stdscat,300.0,
                                 &tab,&status);
            ps.science_paws[i].mstds[j-1] = casu_tfits_wrap(tab,
                                                            ps.science_paws[i].cat[j-1],NULL,NULL);
            freetable(stdscat);
        }
        cpl_msg_indent_less();
        cpl_msg_info(fctid,
                     "Doing photometric calibriation on paw%" CPL_SIZE_FORMAT "",
                     (cpl_size)(i+1));
        vircam_sci_photcal(&(ps.science_paws[i]));
        template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[i].orig,0));
        cpl_msg_info(fctid,"Saving paw%" CPL_SIZE_FORMAT "",(cpl_size)(i+1));
        cpl_msg_indent_more();
        for (j = 1; j <= VIRCAM_NEXTN; j++) {
            cpl_msg_info(fctid,
                         "Saving image paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_save_stack(ps.science_paws[i].stack[j-1],framelist,
                                  parlist,template,
                                  vircam_sci_config.prettynames,i+1,
                                  &(ps.science_paws[i].product_frame_im));
            cpl_msg_info(fctid,"Saving confidence map paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_save_stack_conf(ps.science_paws[i].stackc[j-1],framelist,
                                       parlist,template,
                                       vircam_sci_config.prettynames,i+1,
                                       &(ps.science_paws[i].product_frame_conf));
            cpl_msg_info(fctid,
                         "Saving catalogue paw%" CPL_SIZE_FORMAT "[%" CPL_SIZE_FORMAT "]",
                         (cpl_size)(i+1),(cpl_size)j);
            vircam_sci_save_stack_cat(ps.science_paws[i].cat[j-1],
                                      framelist,parlist,template,
                                      vircam_sci_config.prettynames,i+1,
                                      &(ps.science_paws[i].product_frame_cat));
        }
        freeframe(template);
        cpl_msg_indent_less();
        vircam_sci_paw_delete(ps.science_paws+i);
    }
    cpl_msg_indent_less();

    /* Final cleanup */

    vircam_sci_tidy(0);
    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_photcal
    \par Purpose:
        Photometrically calibrated the stack
    \par Description:
        A photometric calibration will be done on a stack
    \par Language:
        C
    \param paw
        The pawprint structure for the current stack
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_photcal(pawprint *paw) {
    char filt[16];
    int status;

    /* Get the filter name for the pawprint */

    vircam_pfits_get_filter(casu_fits_get_phu(paw->stack[0]),filt);

    /* Call the routine */
    
    status = CASU_OK;
    casu_photcal_extinct(paw->stack,paw->mstds,paw->cat,VIRCAM_NEXTN,
                         filt,ps.tphottab,vircam_sci_config.minphotom,
                         ps.schlf_n,ps.schlf_s,"EXPTIME","ESO TEL AIRM START",
                         vircam_sci_config.magerrcut,&status);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_skydefine
    \par Purpose:
        Set up the sky correction algorithm
    \par Description:
        The parameters for the sky correction algorithm are set up
        based on the user's request
    \par Language:
        C
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_skydefine(void) {
    int i,j;
    cpl_frameset *newfrmset,**in;
    cpl_frame *template;

    /* Switch for the sky algorithm choices. Start with SKYNONE where
       we don't have to do anything as the sky frame index in the pawprints
       references nothing */

    switch (vircam_sci_config.skyalgo) {
    case SKYNONE:
        break;

    /* AUTO means we use the data to determine what is the best way
       to estimate the sky. The routine also assigns the sky to be subtracted
       to each object image */
        
    case AUTO:
        vircam_sci_choose_skyalgo(&(ps.nskys),&(ps.skys));
        break;

    /* A master sky exists. Therefore all we have to do here is to 
       create a sky structure and then assign it to all the input objects */

    case SKYMASTER:
        ps.nskys = 1;
        ps.skys = cpl_malloc(sizeof(skystruct));
        ps.skys[0].skyframe = cpl_frame_duplicate(ps.master_sky);
        ps.skys[0].skyalgo = SKYMASTER;
        ps.skys[0].contrib = NULL;
        ps.skys[0].objmask = NULL;
        ps.skys[0].template = NULL;
        for (i = 0; i < ps.nscipaws; i++) 
            vircam_sci_assign_sky_all(ps.science_paws[i],0);
        break;
    
    /* PAWSKY_MASK is requested. NB: there may be more than one sky because
       the user may have requested this algorithm for a tile. We need a 
       section for the case where offset skies have been provided or not */

    case PAWSKY_MASK:
        if (ps.noffpaws == 0) {
            ps.nskys = ps.nscipaws;
            ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
            for (i = 0; i < ps.nscipaws; i++) {
                newfrmset = cpl_frameset_duplicate(ps.science_paws[i].current);
                template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[i].orig,0));
                ps.skys[i] = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,
                                              i+1);
                vircam_sci_assign_sky_all(ps.science_paws[i],i);        
            }
        } else {
            ps.nskys = ps.noffpaws;
            ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
            for (i = 0; i < ps.noffpaws; i++) {
                newfrmset = cpl_frameset_duplicate(ps.offsky_paws[i].current);
                template = cpl_frame_duplicate(cpl_frameset_get_position(ps.offsky_paws[i].orig,0));
                ps.skys[i] = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,
                                              i+1);
                vircam_sci_assign_sky_all(ps.offsky_paws[i],i); 
            }
            for (i = 0; i < ps.nscipaws; i++) {
                if (i > ps.nskys-1) 
                    j = ps.nskys - 1;
                else
                    j = i;
                vircam_sci_assign_sky_all(ps.science_paws[i],j);        
            }
        }           
        break;

    /* PAWSKY_MASK_PRE is requested. NB: there may be more than one sky because
       the user may have requested this algorithm for a tile. Offset skies
       are ignored because the mask won't cover this. */

    case PAWSKY_MASK_PRE:
        ps.nskys = ps.nscipaws;
        ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
        for (i = 0; i < ps.nscipaws; i++) {
            newfrmset = cpl_frameset_duplicate(ps.science_paws[i].current);
            template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[i].orig,0));
            ps.skys[i] = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,i+1);
            vircam_sci_assign_sky_all(ps.science_paws[i],i);    
        }
        break;

    /* TILESKY is requested. This results in one sky */

    case TILESKY:
        ps.nskys = 1;
        ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
        if (ps.noffpaws == 0) {
            in = cpl_malloc(ps.nscipaws*sizeof(cpl_frameset *));
            for (i = 0; i < ps.nscipaws; i++) 
                in[i] = ps.science_paws[i].current;
            newfrmset = vircam_sci_merge_framesets(ps.nscipaws,in);
            freespace(in);
            template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,0));
            ps.skys[0] = vircam_sci_crsky(TILESKY,newfrmset,template,1);
            for (i = 0; i < ps.nscipaws; i++) 
                vircam_sci_assign_sky_all(ps.science_paws[i],0);        
        } else {
            in = cpl_malloc(ps.noffpaws*sizeof(cpl_frameset *));
            for (i = 0; i < ps.noffpaws; i++) 
                in[i] = ps.offsky_paws[i].current;
            newfrmset = vircam_sci_merge_framesets(ps.noffpaws,in);
            freespace(in);
            template = cpl_frame_duplicate(cpl_frameset_get_position(ps.offsky_paws[0].orig,0));
            ps.skys[0] = vircam_sci_crsky(TILESKY,newfrmset,template,1);
            for (i = 0; i < ps.noffpaws; i++) 
                vircam_sci_assign_sky_all(ps.offsky_paws[i],0); 
            for (i = 0; i < ps.nscipaws; i++) 
                vircam_sci_assign_sky_all(ps.science_paws[i],0);        
        }
        break;

    /* TILESKY_MINUS is requested. This has to be done separately, so 
       the information in the sky structure isn't quite what is going
       to happen. Offset skies are ignored */

    case TILESKY_MINUS:
        ps.nskys = 1;
        ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
        in = cpl_malloc(ps.nscipaws*sizeof(cpl_frameset *));
        for (i = 0; i < ps.nscipaws; i++) 
            in[i] = ps.science_paws[i].current;
        newfrmset = vircam_sci_merge_framesets(ps.nscipaws,in);
        freespace(in);
        template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,0));
        ps.skys[0] = vircam_sci_crsky(TILESKY_MINUS,newfrmset,template,1);
        for (i = 0; i < ps.nscipaws; i++) 
            vircam_sci_assign_sky_all(ps.science_paws[i],0);    
        break;

    /* PAWSKY_MINUS is requested. This has to be done separately, so 
       the information in the sky structure isn't quite what is going
       to happen. Offset skies are ignored */

    case PAWSKY_MINUS:
        ps.nskys = 1;
        ps.skys = cpl_malloc(ps.nskys*sizeof(skystruct));
        newfrmset = cpl_frameset_duplicate(ps.science_paws[0].current);
        template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,0));
        ps.skys[0] = vircam_sci_crsky(PAWSKY_MINUS,newfrmset,template,1);
        vircam_sci_assign_sky_all(ps.science_paws[0],0);        
        break;
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_cat
    \par Purpose:
        Do the source extraction for an extension of a pawprint
    \par Description:
        Do the source extraction for a single extension in a given
        pawprint.
    \par Language:
        C
    \param paw
        The structure for the current pawprint
    \param extn
        The extension to be done
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_cat(pawprint *paw, int extn) {
    int status;
    casu_tfits *tab;
    float gain;

    /* Do the catalogue for the current extension */

    status = CASU_OK;
    (void)vircam_pfits_get_gain(casu_fits_get_ehu(paw->stack[extn-1]),&gain);
    (void)casu_imcore(paw->stack[extn-1],paw->stackc[extn-1],stk_cat.ipix,
                      stk_cat.thresh,stk_cat.icrowd,stk_cat.rcore,
                      stk_cat.nbsize,6,2.0,&tab,gain,&status);
    paw->cat[extn-1] = tab;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_stack
    \par Purpose:
        Do the source extraction for an extension of a pawprint
    \par Description:
        Do the source extraction for a single extension in a given
        pawprint.
    \par Language:
        C
    \param paw
        The structure for the current pawprint
    \param extn
        The extension to be done
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_stack(pawprint *paw, int extn) {
    casu_fits **inf,*inconf,*out,*outc,*outv;
    int ninf,status,i,fastslow;
    double sum,mjd;
    
    /* Do the stack for the current extension */

    inf = casu_fits_load_list(paw->current,CPL_TYPE_FLOAT,extn);
    inconf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,extn);
    ninf = (int)cpl_frameset_get_size(paw->current);
    status = CASU_OK;
    if (stk.fast == -1) 
        fastslow = (ninf <= stk.nfst);
    else
        fastslow = stk.fast;
    casu_imstack(inf,&inconf,NULL,NULL,ninf,1,stk.lthr,stk.hthr,stk.method,
                 stk.seeing,fastslow,1,"EXPTIME",&out,&outc,&outv,&status);

    /* Get the mjd of each input image. Average it and write that to the
       output stack header */

    sum = 0.0;
    for (i = 0; i < ninf; i++) {
        vircam_pfits_get_mjd(casu_fits_get_phu(inf[i]),&mjd);
        sum += mjd;
    }
    sum /= (double)ninf;
    cpl_propertylist_update_double(casu_fits_get_ehu(out),"MJD_MEAN",sum);
    cpl_propertylist_set_comment(casu_fits_get_ehu(out),"MJD_MEAN",
                                 "Mean MJD of the input images");
    freefitslist(inf,ninf);
    freefits(inconf);
    paw->stack[extn-1] = out;
    paw->stackc[extn-1] = outc;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_sky_stripe_wcs
    \par Purpose:
        Do sky, stripe and WCS correction for the images and a pawprint
    \par Description:
        Each of the images in a pawprint is corrected to remove both the
        sky background and the instrumental striping. Later a WCS is 
        fitted to positions of each object in the image. The images are
        also saved here
    \par Language:
        C
    \param npaw
        The number of pawprints
    \param paw
        The array of structures for the pawprints
    \param tag
        The product tag for the images
    \param framelist
        The input framelist
    \param parlist
        The input parameter list
    \param destripe
        If set, then the destriping will be done
    \retval CASU_OK
        if everything is ok
    \retval CASU_FATAL
        if there are problems with the reduction
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_sky_stripe_wcs(int npaw, pawprint *paw, const char *tag,
                                     cpl_frameset *framelist,
                                     cpl_parameterlist *parlist, int destripe) {
    int i,j,jj,k,nf,n,isfirst,ii,status,whichsky,skyminus;
    long nx,ny,npts;
    cpl_frame *curframe,*template,*product_frame = NULL;
    cpl_frameset *objframes;
    cpl_image *img,*simg;
    float med,sig,*sdata;
    casu_fits **sky_extns,*curfits[4],*tmpfits,*fconf;
    unsigned char *bpm[VIRCAM_NEXTN];
    cpl_mask *cplmasks[VIRCAM_NEXTN];
    const char *fctid = "vircam_sci_sky_stripe_wcs";
    char tmpname[BUFSIZ];

    /* NULL input */

    if (npaw == 0 || paw == NULL)
        return(CASU_OK);

    /* Set up the masking from the confidence maps */

    for (i = 1; i <= VIRCAM_NEXTN; i++) {
        casu_mask_load(ps.mask,i,0,0);
        nx = casu_mask_get_size_x(ps.mask);
        ny = casu_mask_get_size_y(ps.mask);
        bpm[i-1] = cpl_calloc(nx*ny,sizeof(unsigned char));
        memmove(bpm[i-1],casu_mask_get_data(ps.mask),
                nx*ny*sizeof(unsigned char));
        cplmasks[i-1] = cpl_mask_wrap((cpl_size)nx,(cpl_size)ny,
                                      (cpl_binary *)bpm[i-1]);
    }

    /* Loop for all the pawprints... */

    cpl_msg_indent_more();
    sky_extns = cpl_calloc(VIRCAM_NEXTN,sizeof(casu_fits *));
    for (j = 0; j < VIRCAM_NEXTN; j++)
        sky_extns[j] = NULL;
    for (j = 0; j < npaw; j++) {

        /* Do the sky subtraction for each frame in the pawprint. */

        objframes = paw[j].current;
        nf = cpl_frameset_get_size(objframes);
        whichsky = -1;
        skyminus = (ps.skys[paw[j].whichsky[0]].skyalgo == TILESKY_MINUS || 
                    ps.skys[paw[j].whichsky[0]].skyalgo == PAWSKY_MINUS);
        for (jj = 0; jj < nf; jj++) {
            curframe = cpl_frameset_get_position(objframes,jj);
            cpl_msg_info(fctid,"Beginning work on %s",
                         cpl_frame_get_filename(curframe));
            template = cpl_frameset_get_position(paw[j].orig,jj);
            n = 0;
            (void)snprintf(tmpname,BUFSIZ,"tmp_%s",
                           cpl_frame_get_filename(curframe));

            /* First load up the sky extensions and normalise the data arrays
               to a zero median (so long as the sky correction hasn't
               already been done). */

            if (! skyminus && 
                (whichsky == -1 || paw[j].whichsky[jj] != whichsky)) {
                whichsky = paw[j].whichsky[jj];
                for (i = 1; i <= VIRCAM_NEXTN; i++) {
                    freefits(sky_extns[i-1]);
                    sky_extns[i-1] = casu_fits_load(ps.skys[whichsky].skyframe,
                                                    CPL_TYPE_FLOAT,i);
                    img = casu_fits_get_image(sky_extns[i-1]);
                    sdata = cpl_image_get_data_float(img);
                    nx = (long)cpl_image_get_size_x(img);
                    ny = (long)cpl_image_get_size_y(img);
                    npts = nx*ny;
                    casu_qmedsig(sdata,bpm[i-1],npts,5.0,3,-1000.0,65535.0,
                                 &med,&sig);
                    for (k = 0; k < npts; k++)
                        sdata[k] -= med;
                }
            }

            /* Loop for all the extensions. First do the sky subtraction
               (if we need to do it) */

            isfirst = 1;
            for (i = 1; i <= VIRCAM_NEXTN; i++) {
                curfits[n] = casu_fits_load(curframe,CPL_TYPE_FLOAT,i);
                img = casu_fits_get_image(curfits[n]);
                if (sky_extns[i-1] != NULL) {
                    simg = casu_fits_get_image(sky_extns[i-1]);
                    cpl_image_subtract(img,(const cpl_image *)simg);
                    cpl_propertylist_update_string(casu_fits_get_ehu(curfits[n]),
                                                   "ESO DRS SKYCOR",
                                                   casu_fits_get_fullname(sky_extns[i-1]));
                }

                /* Reject the bad pixels from the image */

                cpl_image_reject_from_mask(img,cplmasks[i-1]);

                /* When we've done four, destripe them and then write them
                   out in a temporary file */

                n++;
                if (n == 4) {
                    n = 0;
                    status = CASU_OK;
                    if (destripe) 
                        vircam_destripe_four(curfits,vircam_sci_config.str_filt,
                                             &status);
                    for (ii = 0; ii < 4; ii++) {
                        tmpfits = casu_fits_duplicate(curfits[ii]);
                        casu_fits_set_filename(tmpfits,tmpname);
                        fconf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,
                                                 i-3+ii);
                        vircam_sci_wcsfit(&tmpfits,&fconf,NULL,1,1);
                        vircam_sci_save_simple(tmpfits,framelist,parlist,
                                               1,template,isfirst,tag,
                                               tmpname,&product_frame);
                        isfirst = 0;
                        freefits(curfits[ii]);
                        freefits(tmpfits);
                        freefits(fconf);
                    }
                }
            }

            /* When we've finished writing out the new file. Delete the
               old one and replace it with the new one */

            remove(cpl_frame_get_filename(curframe));
            cpl_frame_set_filename(product_frame,
                                   cpl_frame_get_filename(curframe));
            rename(tmpname,cpl_frame_get_filename(curframe));
        }

        /* Clear up some memory */

        for (i = 0; i < VIRCAM_NEXTN; i++)
            freefits(sky_extns[i]);
    }
    for (i = 0; i < VIRCAM_NEXTN; i++) {
        cpl_mask_unwrap(cplmasks[i]);
        freespace(bpm[i]);
    }
    freespace(sky_extns);
    cpl_msg_indent_less();
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_findtemplate
    \par Purpose:
        Look through the input list to find a template for a given frame
    \par Description:
        For a given frame, look at the file names of all the input images
        for each of the pawprints and find the one that matches the
        file name for the input frame
    \par Language:
        C
    \param in
        The input frame that we'd like to match
    \returns
        A reference to frame from the pawprint lists that matches
        the current one.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static cpl_frame *vircam_sci_findtemplate(cpl_frame *in) {
    int i,j,imatch,jmatch,n;
    cpl_frame *frm;

    imatch = -1;
    jmatch = -1;
    for (i = 0; i < ps.nscipaws; i++) {
        n = cpl_frameset_get_size(ps.science_paws[i].current);
        for (j = 0; j < n; j++) {
            frm = cpl_frameset_get_position(ps.science_paws[i].current,j);
            if (! strcmp(cpl_frame_get_filename(in),
                         cpl_frame_get_filename(frm))) {
                imatch = i;
                jmatch = j;
                break;
            }
        }
        if (imatch != -1) 
            break;
    }
    if (imatch == -1) 
        return(NULL);
    else
        return(cpl_frameset_get_position(ps.science_paws[imatch].orig,jmatch));
}
    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_wcsfit
    \par Purpose:
        Fit a WCS to each image in a list
    \par Description:
        Fit a WCS to each image in a list. For a level 1 WCS a temporary
        source catalogue is generated and used for the fit. For a level 2
        WCS, this catalogue must be provided by the calling routine.
    \par Language:
        C
    \param in
        The list of input images
    \param conf
        The list of input confidence maps
    \param incat
        The list of input source catalogues (for level 2 fit)
    \param nf
        The number of entries in the list
    \param level
        The level of the fit. Level 1 is done with a rough source catalogue.
        Level 2 is done with the provided source catalogue and should be 
        considered definitive.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_wcsfit(casu_fits **in, casu_fits **conf, 
                              casu_tfits **incat, int nf, int level) {
    int status,nstd,ncat,slevel,n,i,j;
    float *x,*y,*ra,*dec,gain;
    double r,d;
    casu_tfits *tcat;
    cpl_table *stdscat,*newstds,*cat,*tmp,*tmp2,*matchstds;
    cpl_propertylist *p;
    cpl_wcs *wcs;
    const char *fctid = "vircam_sci_wcsfit";

    /* Loop for all the images */

    for (j = 0; j < nf; j++) {

        /* If this is a level one wcsfit, then we generate a temporary
           catalogue */

        status = CASU_OK;
        (void)vircam_pfits_get_gain(casu_fits_get_ehu(in[j]),&gain);
        if (level == 1) 
            (void)casu_imcore(in[j],conf[j],10,1.5,0,3.0,64,3,2.0,&tcat,
                              gain,&status);
        else 
            tcat = incat[j];

        /* Get some standard stars */

        (void)casu_getstds(casu_fits_get_ehu(in[j]),1,
                           current_catpath,current_cat,
                           vircam_sci_config.cdssearch,
                           vircam_sci_config.cacheloc,&stdscat,&status);
        if (status != CASU_OK) {
            freetable(stdscat);
            cpl_msg_error(fctid,
                          "Failed to find any standards for %s[%" CPL_SIZE_FORMAT "]",
                          casu_fits_get_filename(in[j]),
                          (cpl_size)casu_fits_get_nexten(in[j]));
            if (level == 1)
                freetfits(tcat);
            return;
        }

        /* Restrict to stars with good photometry */

        /* (void)cpl_table_or_selected_float(stdscat,"j_msig",CPL_LESS_THAN,0.2); */
        /* (void)cpl_table_and_selected_float(stdscat,"k_msig",CPL_LESS_THAN,0.2); */
        newstds = cpl_table_extract_selected(stdscat);
        nstd = (int)cpl_table_get_nrow(newstds);

        /* If there are too many objects in the catalogue then first restrict
           ourselves by ellipticity. Cut so that there are similar numbers of
           objects in the standards and the object catalogues by retaining the
           brighter objects */

        cat = casu_tfits_get_table(tcat);
        ncat = (int)cpl_table_get_nrow(cat);
        if (ncat > 500 && ncat > 2.0*nstd) {
            tmp = cpl_table_duplicate(cat);
            (void)cpl_table_or_selected_float(tmp,"Ellipticity",CPL_LESS_THAN,0.5);
            tmp2 = cpl_table_extract_selected(tmp);
            ncat = (int)cpl_table_get_nrow(tmp2);
            freetable(tmp);
            p = cpl_propertylist_new();
            cpl_propertylist_append_bool(p,"Isophotal_flux",TRUE);          
            cpl_table_sort(tmp2,(const cpl_propertylist *)p);
            cpl_propertylist_delete(p);
            slevel = min(ncat,max(1,min(5000,max(500,2*nstd))));
            tmp = cpl_table_extract(tmp2,1,(cpl_size)slevel);
            freetable(tmp2);
            ncat = (int)cpl_table_get_nrow(tmp);
            cat = tmp;
        } else {
            tmp = NULL;
        }

        /* Now match this against the catalogue */

        (void)casu_matchstds(cat,newstds,300.0,&matchstds,&status);
        freetable(stdscat);
        freetable(newstds);
        freetable(tmp);
        if (status != CASU_OK) {
            freetable(matchstds);
            cpl_msg_error(fctid,"Failed to match standards to catalogue");
            if (level == 1)
                freetfits(tcat);
            return;
        }

        /* Fit the plate solution */

        (void)casu_platesol(casu_fits_get_ehu(in[j]),
                            casu_tfits_get_ehu(tcat),matchstds,6,1,&status);
        freetable(matchstds);
        if (status != CASU_OK) {
            cpl_msg_error(fctid,"Failed to fit WCS");
            if (level == 1)
                freetfits(tcat);
            return;
        }

        /* Update the RA and DEC of the objects in the object catalogue */

        if (level == 2) {
            cat = casu_tfits_get_table(tcat);
            n = (int)cpl_table_get_nrow(cat);
            wcs = cpl_wcs_new_from_propertylist(casu_fits_get_ehu(in[j]));
            if (wcs == NULL) {
                cpl_msg_error(fctid,"Failed to fill RA and Dec in catalogue");
                return;
            }
            x = cpl_table_get_data_float(cat,"X_coordinate"); 
            y = cpl_table_get_data_float(cat,"Y_coordinate");
            ra = cpl_table_get_data_float(cat,"RA");
            dec = cpl_table_get_data_float(cat,"DEC");
            for (i = 0; i < n; i++) {
                casu_xytoradec(wcs,x[i],y[i],&r,&d);
                ra[i] = (float)r;
                dec[i] = (float)d;
            }
            cpl_wcs_delete(wcs);
        } else {
            freetfits(tcat);
        }
    }
}
    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_destripe_four
    \par Purpose:
        Destripe a row of four extensions in a MEF
    \par Description:
        The vircam stripe pattern is the same for all four detectors in a 
        row. This routine uses this redundancy to work out a stripe pattern
        for all four detectors simultaneously
    \par Language:
        C
    \param curfits
        The list of four input images
    \param stripefilt
        If set, then the stripe profile will be smoothed before being
        applied to the data.
    \param status
        The usual status variable
    \retval CASU_OK
        if everything is ok
    \retval CASU_FATAL
        if there were any problems with the routine was called.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_destripe_four(casu_fits **curfits, int stripefilt, 
                                int *status) {
    int i,nx,ny,n,k,m,ipstart,ipfin,ip;
    float *profiles[4],*meanprof,buf[4],*data,val,mad,rms;
    unsigned char *bpms[4],*bp;
    cpl_propertylist *plist;

    /* Inherited status */

    if (*status != CASU_OK)
        return(*status);

    /* Get the data size of a representative image and get some memory for the
       mean profile and its bad pixel mask */

    nx = cpl_image_get_size_x(casu_fits_get_image(curfits[0]));
    ny = cpl_image_get_size_y(casu_fits_get_image(curfits[0]));
    meanprof = cpl_malloc(ny*sizeof(float));
    bp = cpl_calloc(ny,sizeof(unsigned char));

    /* Loop for the four images and work out the stripe profile for each
       one. Trap for missing images */

    n = 0;
    for (i = 0; i < 4; i++) {
        if (casu_fits_get_status(curfits[i]) == CASU_FATAL) 
            continue;
        profiles[n] = cpl_malloc(ny*sizeof(float));
        bpms[n] = cpl_calloc(ny,sizeof(unsigned char));
        stripe_profile(curfits[i],stripefilt,profiles[n],bpms[n]);
        n++;
    }

    /* Now median the profiles. If there weren't any contributions
       this is presumably because all the detectors were missing
       so, don't bother doing the destripe. While doing the medianing, 
       subtract the average profile from the row in each of the contributing 
       images */

    if (n > 0) {
        for (k = 0; k < ny; k++) {
            m = 0;
            for (i = 0; i < n; i++) {
                if (bpms[i][k])
                    continue;
                buf[m++] = profiles[i][k];
            }
            if (m != 0) {
                meanprof[k] = casu_med(buf,NULL,(long)m);
                bp[k] = 0;
            } else {
                meanprof[k] = 0.0;
                bp[k] = 1;
            }
            ipstart = k*nx;
            ipfin = ipstart + nx;
            for (i = 0; i < 4; i++) {
                if (casu_fits_get_status(curfits[i]) == CASU_FATAL) 
                    continue;
                data = cpl_image_get_data(casu_fits_get_image(curfits[i]));
                for (ip = ipstart; ip < ipfin; ip++) 
                    data[ip] -= meanprof[k];
            }
        }
        casu_medmad(meanprof,bp,ny,&val,&mad);
        rms = 1.48*mad;

        /* Tidy up some memory */

        for (i = 0; i < n; i++) {
            freespace(profiles[i]);
            freespace(bpms[i]);
        }

    } else {
        rms = 0.0;
    }
    
    /* Put some stuff in the header */
    
    for (i = 0; i < 4; i++) {
        if (casu_fits_get_status(curfits[i]) == CASU_FATAL) {
            k = FALSE;
            val = 0.0;
        } else { 
            k = TRUE;
            val = rms;
        }
        plist = casu_fits_get_ehu(curfits[i]);
        cpl_propertylist_update_bool(plist,"ESO DRS STRIPECOR",k);
        cpl_propertylist_set_comment(plist,"ESO DRS STRIPECOR",
                                     "Stripe correction done");
        cpl_propertylist_update_float(plist,"ESO DRS STRIPERMS",val);
        cpl_propertylist_set_comment(plist,"ESO DRS STRIPERMS",
                                     "RMS of removed stripe profile");
    }

    /* Tidy up a little more */

    freespace(meanprof);
    freespace(bp);

    /* Get out of here */

    GOOD_STATUS
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        stripe_profile
    \par Purpose:
        Form a stripe profile for a given image
    \par Description:
        Form a stripe profile for a given image using interative clipping
        on each row. Filter the result if requested. Return the profile 
        and a mask that flags where a bad pixel might be in the profile.
    \par Language:
        C
    \param in
        The input image
    \param stripefilt
        If set, then the stripe profile will be smoothed before being
        applied to the data.
    \param profile
        An array for the output profile. This must be allocated and 
        passed in from the calling routine.
    \param pbpm
        An array for the output pixel mask. This must be allocated and 
        passed in from the calling routine.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void stripe_profile(casu_fits *in, int stripefilt, float *profile,
                           unsigned char *pbpm) {
    float *data,*copy,*d,lcut,hcut,skymed,skynoise,rmsold,rms,val;
    unsigned char *bpm,*ob,*b;
    cpl_image *img;
    long nx,ny,npts;
    int i,j;
        
    /* Get the input data array and its bad pixel mask */

    img = casu_fits_get_image(in);
    nx = cpl_image_get_size_x(img);
    ny = cpl_image_get_size_y(img);
    data = cpl_image_get_data(img);
    bpm = (unsigned char *)cpl_mask_get_data(cpl_image_get_bpm(img));

    /* First get an estimate of the background median and sigma */

    npts = nx*ny;
    casu_qmedsig(data,bpm,npts,3.0,3,-65535.0,65535.0,&skymed,&skynoise);

    /* Get some workspace for the profile calculation */

    ob = cpl_calloc(ny,sizeof(unsigned char));
    copy = cpl_malloc(ny*sizeof(float));
    
    /* Loop for each row */

    d = data;
    b = bpm;
    for (i = 0; i < ny; i++) {

        /* Get the median of that row, ignoring any bad pixels. Iterate
           to clip out any objects or funny pixels */

        lcut = -1.0e-10;
        hcut = 1.0e10;
        rmsold = 1.0e10;
        for (j = 0; j < 3; j++) {
            casu_medmadcut(d,b,nx,lcut,hcut,&val,&rms);
            if (val == CX_MAXFLOAT) {
                break;
            } else if (rms == rmsold) {
                break;
            } else {
                lcut = val - 3.0*skynoise;
                hcut = val + 3.0*skynoise;
                rmsold = rms;
            }
        }
        pbpm[i] = (val == CX_MAXFLOAT);
        profile[i] = (pbpm[i] ? 0.0 : val);
        copy[i] = profile[i];
        d += nx;
        b += nx;
    }

    /* Take out large scale variation if need be */

    casu_medmad(profile,pbpm,ny,&val,&rms);
    rms *= 1.48;
    if (stripefilt && rms > 15.0) {
        casu_dostat(copy,pbpm,ob,ny,25,MEDIANCALC);
        casu_dostat(copy,pbpm,ob,ny,5,MEANCALC);
        for (i = 0; i < ny; i++) {
            if (! pbpm[i])
                profile[i] -= copy[i];
            else
                profile[i] = 0.0;
        }
    } else {
        for (i = 0; i < ny; i++)
            profile[i] -= val;
    }

    /* Tidy up and get out of here */
    
    freespace(ob);
    freespace(copy);
}
            
    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_paws_delete
    \par Purpose:
        Delete a list of pawprint structures
    \par Description:
        Delete a list of pawprint structures and free up all of
        the memory associated with them.
    \par Language:
        C
    \param npaws
        The number of input pawprints
    \param paws
        The input list of pawprints
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_paws_delete(int npaws, pawprint **paws) {
    int i;

    for (i = 0; i < npaws; i++) 
        vircam_sci_paw_delete(*paws+i);
    freespace(*paws);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_paw_delete
    \par Purpose:
        Delete a single pawprint structure
    \par Description:
        Delete a pawprint structure and free up all of the memory 
        associated with it.
    \par Language:
        C
    \param paw
        The input pawprint
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_paw_delete(pawprint *paw) {
    int i;

    if (paw->current == NULL)
        return;
    freespace(paw->whichsky);
    freeframeset(paw->orig);
    freeframeset(paw->current);
    for (i = 0; i < VIRCAM_NEXTN; i++) {
        freefits(paw->stack[i]);
        freefits(paw->stackc[i]);
        freetfits(paw->cat[i]);
        freetfits(paw->mstds[i]);
    }
}
                
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_paws_create
    \par Purpose:
        Create a set of pawprint structures
    \par Description:
        Given an input frameset create a number of pawprint structures
        based on information that shows how the observations were grouped
    \par Language:
        C
    \param infrms
        The input pawprint
    \param npaws
        The number of output pawprint structures
    \param paws
        The output pawprint structures
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_paws_create(cpl_frameset *infrms, int *npaws, 
                                   pawprint **paws) {
    cpl_size nlab,*labels,*labels2,nlab2;
    cpl_frameset *copy,*testme,*tf;
    int i,j,nalloc,pj,nfrms_tot;

    /* NULL input...*/

    *paws = NULL;
    *npaws = 0;
    if (infrms == NULL)
        return;

    /* Create the grouping */

    copy = cpl_frameset_duplicate(infrms);
    labels = cpl_frameset_labelise(copy,vircam_sci_cmp_tstart,&nlab);

    /* Get memory for the pawprint groups. We will trim this down or 
       expand it as necessary */

    *npaws = 0;
    nalloc = 8;
    *paws = cpl_malloc(nalloc*sizeof(pawprint));

    /* Now extract the groups */

    nfrms_tot = 0;
    for (i = 0; i < nlab; i++) {

        /* First see if this is a tile.  If it is, then break the tile up
           into its constituent pawprints */

        testme = cpl_frameset_extract(copy,labels,(cpl_size)i);
        if (vircam_sci_istile(cpl_frameset_get_position(testme,0))) {
            pj = vircam_sci_ispj(testme);
            labels2 = cpl_frameset_labelise(testme,vircam_sci_cmp_jit,&nlab2);
            for (j = 0; j < nlab2; j++) {
                if (*npaws+1 > nalloc) {
                    nalloc += 8;
                    *paws = cpl_realloc(*paws,nalloc*sizeof(pawprint));
                }
                tf = cpl_frameset_extract(testme,labels2,(cpl_size)j);
                vircam_sci_paw_init(*paws+*npaws,tf,pj,&nfrms_tot);
                freeframeset(tf);
                *npaws += 1;
            }
            freespace(labels2);
        } else {
            vircam_sci_paw_init(*paws+*npaws,testme,0,&nfrms_tot);
            *npaws += 1;
        }
        freeframeset(testme);
    }

    /* Trim down allocation */

    *paws = cpl_realloc(*paws,*npaws*sizeof(pawprint));

    /* Tidy and exit */

    cpl_frameset_delete(copy);
    freespace(labels);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_istile
    \par Purpose:
        Determine whether a frame is part of a tiled observation
    \par Description:
        Use the header information from a single frame to tell whether
        this is part of a tiled observation or not.
    \par Language:
        C
    \param frm
        The input frame
    \retval
        1 if this is part of a tile and 0 if it isn't
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_istile(cpl_frame *frm) {
    cpl_propertylist *plist;
    int hasoffs,noffsets;
    
    /* If the frame has a keyword OFFSTNUM and NOFFSETS != 1 then this
       is part of a tile */

    plist = cpl_propertylist_load(cpl_frame_get_filename(frm),0);
    hasoffs = cpl_propertylist_has(plist,"OFFSTNUM");
    if (hasoffs) 
        noffsets = cpl_propertylist_get_int(plist,"NOFFSETS");
    else
        noffsets = -1;
    cpl_propertylist_delete(plist);
    return(hasoffs && noffsets != 1);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_paw_init
    \par Purpose:
        Initialise a pawprint structure
    \par Description:
        Given a frameset, initialise the pawprint structure.
    \par Language:
        C
    \param paw
        The input pawprint structure
    \param frms
        The frameset for pawprint
    \param ispj
        Set if this pawprint was observed in PJ order
    \param nfrms_tot
        The total number of frames in the full dataset so far
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_paw_init(pawprint *paw, cpl_frameset *frms, int ispj,
    int *nfrms_tot) {
    cpl_propertylist *plist;
    cpl_frame *fr;
    float exptime;
    int i,n;

    /* Initialise everything. Start with all the frames and framesets */

    paw->orig = cpl_frameset_duplicate(frms);
    paw->current = vircam_sci_update_frameset(paw->orig,nfrms_tot);
    n = cpl_frameset_get_size(frms);
    paw->whichsky = cpl_malloc(n*sizeof(int));
    for (i = 0; i < n; i++)
        paw->whichsky[i] = -1;
    paw->ispj = ispj;
    for (i = 0; i < VIRCAM_NEXTN; i++) {
        paw->stack[i] = NULL;
        paw->stackc[i] = NULL;
        paw->cat[i] = NULL;
        paw->mstds[i] = NULL;
    }
    paw->product_frame_im = NULL;
    paw->product_frame_conf = NULL;
    paw->product_frame_cat = NULL;

    /* Now get some useful information about the pawprint from the headers.
       Start with the jitter and tile numbers */

    fr = cpl_frameset_get_position(frms,0);
    plist = cpl_propertylist_load(cpl_frame_get_filename(fr),0);
    (void)vircam_pfits_get_jitternum(plist,&(paw->jitternum));
    if (vircam_sci_istile(fr)) {
        if (vircam_pfits_get_offsetnum(plist,&(paw->tilenum)) != CASU_OK)
            paw->tilenum = -1;
    } else {
        paw->tilenum = -1;
    }

    /* Now the project ID */

    (void)vircam_pfits_get_projid(plist,paw->projname);

    /* The MJD of the start of the jitter and of the end */

    (void)vircam_pfits_get_mjd(plist,&(paw->mjd_start));
    cpl_propertylist_delete(plist);
    n = (int)cpl_frameset_get_size(frms);
    fr = cpl_frameset_get_position(frms,n-1);
    plist = cpl_propertylist_load(cpl_frame_get_filename(fr),0);
    (void)vircam_pfits_get_mjd(plist,&(paw->mjd_end));
    (void)vircam_pfits_get_exptime(plist,&exptime);
    paw->mjd_end += exptime/(24.0*3600.0);

    /* Tidy and exit */

    cpl_propertylist_delete(plist);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_makesky
    \par Purpose:
        Create and save a sky frame 
    \par Description:
        Given a sky structure, create a sky frame using the chosen 
        algorithm. Save the file.
    \par Language:
        C
    \param framelist
        The input raw frameset
    \param parlist
        The input command line parameter list
    \param sky
        The input sky structure
    \param tag
        The product tag to be used when saving the file
    \retval 
        CASU_OK if everything went well. CASU_FATAL if there was a problem
        in one of the sky combine routines.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_makesky(cpl_frameset *framelist, 
                              cpl_parameterlist *parlist, skystruct *sky,
                              const char *tag) {
    int retval;

    /* Switch for the various options to sky creation. Options SKYNONE,
       SKYMASTER and TILESKY_MINUS we don't do anything at this point. */

    retval = CASU_OK;
    switch (sky->skyalgo) {
    case SKYNONE:
        break;
    case SKYMASTER:
        break;
    case PAWSKY_MINUS:
        retval = vircam_sci_pawsky_minus(framelist,parlist,sky->contrib,
                                         sky->template,tag,sky->fname,
                                         &(sky->skyframe));
        break;
    case TILESKY_MINUS:
        retval = vircam_sci_tilesky_minus(framelist,parlist,sky->contrib,
                                          sky->template,tag,sky->fname,
                                          &(sky->skyframe));    
        break;
    case PAWSKY_MASK:
        retval = vircam_sci_pawsky_mask(framelist,parlist,sky->skyalgo,
                                        sky->contrib,sky->template,
                                        sky->fname,&(sky->skyframe));
        break;
    case PAWSKY_MASK_PRE:
        retval = vircam_sci_pawsky_mask(framelist,parlist,sky->skyalgo,
                                        sky->contrib,sky->template,
                                        sky->fname,&(sky->skyframe));
        break;
    case TILESKY:
        retval = vircam_sci_tilesky(framelist,parlist,sky->contrib,
                                    sky->template,sky->fname,&(sky->skyframe));
        break;
    }
    return(retval);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_tilesky_minus
    \par Purpose:
        Create and save a sky frame using the tilesky_minus algorithm
    \par Description:
        Given a list of files do a background estimation and background
        subtraction using the tilesky_minus algorithm. The saved file
        will be a representative sky, but as this algorithm generates a
        sky file for each individual image, the actual sky data used
        to correct the images will not be saved. The sky subtracted images
        will also be saved.
    \par Language:
        C
    \param framelist
        The input raw frameset
    \param parlist
        The input command line parameter list
    \param in
        The input images
    \param template
        A template to be used when saving the representative sky
    \param tag
        The product tag to be used when saving the file
    \param skyname
        The name for the output sky file
    \param skyframe
        The product frame for the output sky file
    \retval 
        CASU_OK if everything went well. CASU_FATAL if there was a problem
        in one of the vircam_tilesky_minus routine.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_tilesky_minus(cpl_frameset *framelist, 
                                    cpl_parameterlist *parlist,
                                    cpl_frameset *in, cpl_frame *template,
                                    const char *tag, char *skyname,
                                    cpl_frame **skyframe) {
    int n,i,status,isfirst,j;
    casu_fits *objmaskfits,**infits,*inconf,*tmpfits,*skyout;
    cpl_frame **product_frames,*curframe;
    char tmpname[BUFSIZ], *skyname2 = NULL;
    const char *fctid = "vircam_sci_tilesky_minus";

    n = cpl_frameset_get_size(in);
    objmaskfits = casu_fits_load(ps.master_objmask,CPL_TYPE_INT,0);
    product_frames = cpl_malloc(n*sizeof(cpl_frame *));
    cpl_msg_info(fctid,"Creating sky %s",skyname);
    cpl_msg_indent_more();
    for (i = 1; i <= VIRCAM_NEXTN; i++) {
        cpl_msg_info(fctid,"Beginning on extn %" CPL_SIZE_FORMAT "",
                     (cpl_size)i);
        infits = casu_fits_load_list(in,CPL_TYPE_FLOAT,i);
        inconf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,i);
        status = CASU_OK;
        if (vircam_tilesky_minus(infits,inconf,objmaskfits,n,&skyout,
                                 &status) != CASU_OK) {
            freefitslist(infits,n);
            freefits(inconf);
            freespace(product_frames);
            return(CASU_FATAL);
        }
        isfirst = (i == 1);
        vircam_sci_save_sky(skyout,framelist,parlist,skyname,template,
                            isfirst,skyframe);
        //(void)snprintf(skyname2,BUFSIZ,"%s[%d]",skyname,i);
        skyname2 = cpl_sprintf("%s[%d]",skyname, i);
        for (j = 0; j < n; j++) {
            tmpfits = casu_fits_duplicate(infits[j]);
            (void)snprintf(tmpname,BUFSIZ,"tmp_%s",
                           casu_fits_get_filename(infits[j]));
            casu_fits_set_filename(tmpfits,tmpname);
            curframe = cpl_frameset_get_position(in,j);
            template = vircam_sci_findtemplate(curframe);
            cpl_propertylist_update_string(casu_fits_get_ehu(tmpfits),
                                           "ESO DRS SKYCOR",skyname2);
            vircam_sci_save_simple(tmpfits,framelist,parlist,0,template,
                                   isfirst,tag,tmpname,(product_frames+j));
            freefits(tmpfits);
        }
        cpl_free(skyname2);
        freefitslist(infits,n);
        freefits(inconf);
        freefits(skyout);
    }
    cpl_msg_indent_less();
    for (i = 0; i < n; i++) {
        curframe = cpl_frameset_get_position(in,i);
        template = product_frames[i];
        remove(cpl_frame_get_filename(curframe));
        rename(cpl_frame_get_filename(template),
               cpl_frame_get_filename(curframe));
        freeframe(product_frames[i]);
    }
    freespace(product_frames);
    freefits(objmaskfits);
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_pawsky_minus
    \par Purpose:
        Create and save a sky frame using the pawsky_minus algorithm
    \par Description:
        Given a list of files do a background estimation and background
        subtraction using the pawsky_minus algorithm. The saved file
        will be a representative sky, but as this algorithm generates a
        sky file for each individual image, the actual sky data used
        to correct the images will not be saved. The sky subtracted images
        will also be saved. If a pre-existing mask exists, it will be used.
    \par Language:
        C
    \param framelist
        The input raw frameset
    \param parlist
        The input command line parameter list
    \param in
        The input images
    \param template
        A template to be used when saving the representative sky
    \param tag
        The product tag to be used when saving the file
    \param skyname
        The name for the output sky file
    \param skyframe
        The product frame for the output sky file
    \retval 
        CASU_OK if everything went well. CASU_FATAL if there was a problem
        in one of the casu_pawsky_minus routine.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_pawsky_minus(cpl_frameset *framelist, 
                                   cpl_parameterlist *parlist,
                                   cpl_frameset *in, cpl_frame *template,
                                   const char *tag, char *skyname,
                                   cpl_frame **skyframe) {
    int n,i,status,isfirst,j,k;
    casu_fits *objmaskfits,**infits,*inconf,*tmpfits,*skyout,*skyvar;
    cpl_frame **product_frames,*curframe;
    char tmpname[BUFSIZ], *skyname2 = NULL;
    const char *fctid = "vircam_sci_pawsky_minus";

    n = cpl_frameset_get_size(in);
    if (ps.master_objmask != NULL) 
        objmaskfits = casu_fits_load(ps.master_objmask,CPL_TYPE_INT,0);
    else
        objmaskfits = NULL;
    product_frames = cpl_malloc(n*sizeof(cpl_frame *));
    cpl_msg_info(fctid,"Creating sky %s",skyname);
    cpl_msg_indent_more();
    for (i = 1; i <= VIRCAM_NEXTN; i++) {
        cpl_msg_info(fctid,"Beginning on extn %" CPL_SIZE_FORMAT "",
                     (cpl_size)i);
        infits = casu_fits_load_list(in,CPL_TYPE_FLOAT,i);
        inconf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,i);
        status = CASU_OK;
        if (ps.master_objmask != NULL) {
            for (k = 0; k < n; k++)
                vircam_sci_wcsfit(infits+k,&inconf,NULL,1,1);
        }
        if (casu_pawsky_minus(infits,NULL,inconf,objmaskfits,n,&skyout,
                              &skyvar,&status) != CASU_OK) {
            freefitslist(infits,n);
            freefits(inconf);
            freespace(product_frames);
            return(CASU_FATAL);
        }
        isfirst = (i == 1);
        vircam_sci_save_sky(skyout,framelist,parlist,skyname,template,
                            isfirst,skyframe);
        //(void)snprintf(skyname2,BUFSIZ,"%s[%d]",skyname,i);
        skyname2 = cpl_sprintf("%s[%d]",skyname, i);
        for (j = 0; j < n; j++) {
            tmpfits = casu_fits_duplicate(infits[j]);
            (void)snprintf(tmpname,BUFSIZ,"tmp_%s",
                           casu_fits_get_filename(infits[j]));
            casu_fits_set_filename(tmpfits,tmpname);
            curframe = cpl_frameset_get_position(in,j);
            template = vircam_sci_findtemplate(curframe);
            cpl_propertylist_update_string(casu_fits_get_ehu(tmpfits),
                                           "ESO DRS SKYCOR",skyname2);
            vircam_sci_save_simple(tmpfits,framelist,parlist,0,template,
                                   isfirst,tag,tmpname,(product_frames+j));
            freefits(tmpfits);
        }
        cpl_free(skyname2);
        freefitslist(infits,n);
        freefits(inconf);
        freefits(skyout);
    }
    cpl_msg_indent_less();
    for (i = 0; i < n; i++) {
        curframe = cpl_frameset_get_position(in,i);
        template = product_frames[i];
        remove(cpl_frame_get_filename(curframe));
        rename(cpl_frame_get_filename(template),
               cpl_frame_get_filename(curframe));
        freeframe(product_frames[i]);
    }
    freespace(product_frames);
    freefits(objmaskfits);
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_choose_skyalgo
    \par Purpose:
        Choose a sky algorithm if the user hasn't specified one
    \par Description:
        If the user doesn't specify a sky algorithm (auto) then choose one
        that appears to be appropriate for the data.
    \par Language:
        C
    \param nskys
        The number of output sky structures
    \param skys
        The list of output sky structures
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_choose_skyalgo(int *nskys, skystruct **skys) {
    float halfhour;
    skystruct newsky;
    int nalloc,i,n,n2,jst,jfn,j,j_1,j_2;
    float dt;
    cpl_frameset **in,*newfrmset;
    cpl_frame *frm,*template;

    /* Initialise a few things */

    halfhour = 1.0/48.0;

    /* In most cases we're only going to create one sky here, but there
       is the possibility of more */

    *nskys = 0;
    nalloc = 1;
    *skys = cpl_malloc(nalloc*sizeof(skystruct));

    /* Ok start with the easy situation. If there are offset skies, then 
       choose just straightforward pawsky_mask or tilesky depending upon
       how the offset skies were observed */

    if (ps.noffpaws > 0) {
        if (vircam_sci_istile(cpl_frameset_get_position(ps.offsky_paws[0].orig,0))) {

            /* Merge all the current framesets into one */

            in = cpl_malloc(ps.noffpaws*sizeof(cpl_frameset *));
            for (i = 0; i < ps.noffpaws; i++) 
                in[i] = ps.offsky_paws[i].current;
            newfrmset = vircam_sci_merge_framesets(ps.noffpaws,in);
            freespace(in);

            /* Create a sky structure */

            template = cpl_frame_duplicate(cpl_frameset_get_position(ps.offsky_paws[0].orig,0));
            newsky = vircam_sci_crsky(TILESKY,newfrmset,template,1);
            *nskys = 1;
            (*skys)[*nskys-1] = newsky;

            /* Assign this one sky to all of the offset and science pawprint
               images */

            for (i = 0; i < ps.noffpaws; i++) 
                vircam_sci_assign_sky_all(ps.offsky_paws[i],*nskys-1);
            for (i = 0; i < ps.nscipaws; i++) 
                vircam_sci_assign_sky_all(ps.science_paws[i],*nskys-1);
        } else {
            newfrmset = cpl_frameset_duplicate(ps.offsky_paws[0].current);
            template = cpl_frame_duplicate(cpl_frameset_get_position(ps.offsky_paws[0].orig,0));
            newsky = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,*nskys+1);
            *nskys += 1;
            (*skys)[*nskys-1] = newsky;
            vircam_sci_assign_sky_all(ps.offsky_paws[0],*nskys-1);
            vircam_sci_assign_sky_all(ps.science_paws[0],*nskys-1);
        }

    /* OK there aren't any offset skies. There are lots more choices now.
       So first we look to see if this is just a pawprint. */

    } else {        

        /* If it's a pawprint, then look to see if it took less than half
           an hour to observe. If it did then split it into two, so long
           as there are enough images in each half to form a sky 
           (nominally 8) */

        if (ps.science_paws[0].tilenum == -1) {
            dt = ps.science_paws[0].mjd_end - ps.science_paws[0].mjd_start;
            n = cpl_frameset_get_size(ps.science_paws[0].current);

            /* Situation where we have a pawprint lasting less than half an
               hour, or we have one that isn't big enough to split */

            if (dt < halfhour || n < 16) {
                newfrmset = cpl_frameset_duplicate(ps.science_paws[0].current);
                template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,0));
                newsky = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,
                                          *nskys+1);
                *nskys = 1;
                (*skys)[*nskys-1] = newsky;
                vircam_sci_assign_sky_all(ps.science_paws[0],*nskys-1);

            /* Pawprint took longer than half an hour and there are enough
               exposures so that we can split into two */
                   
            } else {
                n2 = n/2;
                for (i = 0; i < 2; i++) {
                    if (i == 0) {
                        jst = 0;
                        jfn = n2;
                    } else {
                        jst = n2; 
                        jfn = n;
                    }
                    newfrmset = cpl_frameset_new();
                    for (j = jst; j < jfn; j++) {
                        frm = cpl_frameset_get_position(ps.science_paws[0].current,j);
                        cpl_frameset_insert(newfrmset,cpl_frame_duplicate(frm));
                    }
                    template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,jst));
                    newsky = vircam_sci_crsky(PAWSKY_MASK,newfrmset,template,
                                              *nskys+1);
                    *nskys += 1;
                    if (*nskys > nalloc) {
                        nalloc += 1;
                        *skys = cpl_realloc(*skys,nalloc*sizeof(skystruct));
                    }
                    (*skys)[*nskys-1] = newsky;
                    for (j = jst; j < jfn; j++) 
                        ps.science_paws[0].whichsky[j] = *nskys-1;
                }
            }

        /* Ok this is a tile. There are a couple of options now. If it's a VIDEO tile, 
           then default to TILESKY_MINUS, but with a split */

        } else {
            n = ps.nscipaws;
            dt = ps.science_paws[n-1].mjd_end - ps.science_paws[0].mjd_start;
            if (! strncmp(ps.science_paws[0].projname,VIDEOPROJ,10)) {
                n2 = ps.nscipaws/2;
                for (j = 0; j < 2; j++) {
                    if (j == 0) {
                        j_1 = 0; 
                        j_2 = n2;
                    } else {
                        j_1 = n2;
                        j_2 = ps.nscipaws;
                    }
                    in = cpl_malloc((j_2-j_1)*sizeof(cpl_frameset *));
                    for (i = j_1; i < j_2; i++) 
                        in[i-j_1] = ps.science_paws[i].current;
                    newfrmset = vircam_sci_merge_framesets((j_2-j_1),in);
                    freespace(in);
                    template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[j_1].orig,0));
                    newsky = vircam_sci_crsky(TILESKY_MINUS,newfrmset,template,
                                              *nskys+1);
                    *nskys += 1;
                    if (*nskys > nalloc) {
                        nalloc += 1;
                        *skys = cpl_realloc(*skys,nalloc*sizeof(skystruct));
                    }
                    (*skys)[*nskys-1] = newsky;
                    for (i = j_1; i < j_2; i++) 
                        vircam_sci_assign_sky_all(ps.science_paws[i],*nskys-1);
                }

           /*  Look at the time difference. If the timing is less than half an
               hour or if this is a VMC tile, then just do a straight-forward
               tilesky. */

            } else if (dt < halfhour || 
                       ! strncmp(ps.science_paws[0].projname,VMCPROJ,10)) {
                in = cpl_malloc(ps.nscipaws*sizeof(cpl_frameset *));
                for (i = 0; i < ps.nscipaws; i++) 
                    in[i] = ps.science_paws[i].current;
                newfrmset = vircam_sci_merge_framesets(ps.nscipaws,in);
                freespace(in);

                /* Create a sky structure */

                template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[0].orig,0));
                newsky = vircam_sci_crsky(TILESKY,newfrmset,template,1);
                *nskys = 1;
                (*skys)[*nskys-1] = newsky;

                /* Assign this one sky to all of the science pawprint images */

                for (i = 0; i < ps.nscipaws; i++) 
                    vircam_sci_assign_sky_all(ps.science_paws[i],*nskys-1);

            /* OK, we have a tile that's not VMC and not VIDEO that lasts for
               more than half an hour. If it's the normal PJ ordering, then just
               do a split TILESKY */

            } else {
                if (ps.science_paws[0].ispj) {
                    n2 = ps.nscipaws/2;
                    for (j = 0; j < 2; j++) {
                        if (j == 0) {
                            j_1 = 0; 
                            j_2 = n2;
                        } else {
                            j_1 = n2;
                            j_2 = ps.nscipaws;
                        }
                        in = cpl_malloc((j_2-j_1)*sizeof(cpl_frameset *));
                        for (i = j_1; i < j_2; i++) 
                            in[i-j_1] = ps.science_paws[i].current;
                        newfrmset = vircam_sci_merge_framesets((j_2-j_1),in);
                        freespace(in);
                        template = cpl_frame_duplicate(cpl_frameset_get_position(ps.science_paws[j_1].orig,0));
                        newsky = vircam_sci_crsky(TILESKY,newfrmset,template,
                                                  *nskys+1);
                        *nskys += 1;
                        if (*nskys > nalloc) {
                            nalloc += 1;
                            *skys = cpl_realloc(*skys,nalloc*sizeof(skystruct));
                        }
                        (*skys)[*nskys-1] = newsky;
                        for (i = j_1; i < j_2; i++) 
                            vircam_sci_assign_sky_all(ps.science_paws[i],
                                                      *nskys-1);
                    }
                } else {
                    /******************* bit for jp  */
                }
            } /* end of different type of tiles */
        } /* end of tiles */
                    
    } /* end of offset sky if */        
}                                     

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_isjp
    \par Purpose:
        Work out if an observation was done in PJ order or not
    \par Description:
        Jittering can be done in PJ order (where all of the observations
        for a given jitter sequence are done before moving to the next
        pawprint (OFFSET) position) or in JP order (where a observations
        are cycled between pawprint positions before moving to the next
        jitter position -- this is a stupid thing to do as it reinforces
        the persistence). This routines determines which was used.
    \par Language:
        C
    \param frms
        The input frameset
    \retval
        1: if this is done in PJ order or 0: if this is in JP order.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_ispj(cpl_frameset *frms) {
    int n,offnum,i,offnum1,offnum2;
    cpl_propertylist *plist,*tmp;
    char dateobs[32];
    cpl_frame *frm;

    /* Create a propertylist to be a hash array */

    plist = cpl_propertylist_new();

    /* Loop through the frameset. Name each property by the dateobs of
       the frame. Then give the value of the property the offset number */

    n = cpl_frameset_get_size(frms);
    for (i = 0; i < n; i++) {
        frm = cpl_frameset_get_position(frms,i);
        tmp = cpl_propertylist_load(cpl_frame_get_filename(frm),0);
        strcpy(dateobs,cpl_propertylist_get_string(tmp,"DATE-OBS"));
        offnum = cpl_propertylist_get_int(tmp,"OFFSET_I");
        cpl_propertylist_append_int(plist,dateobs,offnum);
        cpl_propertylist_delete(tmp);
    }

    /* Sort the propertylist by the property name (dateobs). This should
       give a chronological ordering */

    cpl_propertylist_sort(plist,vircam_sci_cmp_property);

    /* Look at the first two now and see if they have the same value
       of the offset number. If they do, then this is in PJ order */

    offnum1 = cpl_property_get_int(cpl_propertylist_get(plist,0));
    offnum2 = cpl_property_get_int(cpl_propertylist_get(plist,1));
    cpl_propertylist_delete(plist);
    return(offnum1 == offnum2); 
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_cmp_property
    \par Purpose:
        Do a string comparison of two properties
    \par Description:
        Do a string comparison of two properties
    \par Language:
        C
    \param p1
        The first property
    \param p2 
        The second property
    \retval
        The result of the strcmp of the two property names
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_cmp_property(const cpl_property *p1, 
                                   const cpl_property *p2) {
    return(strcmp(cpl_property_get_name(p1),cpl_property_get_name(p2)));
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_assign_sky_all
    \par Purpose:
        Assign all the frames in a pawprint to a single sky frame
    \par Description:
        Assign all the frames in a pawprint to a single sky frame
    \par Language:
        C
    \param paw
        The pawprint structure
    \param whichone
        The index number of the sky frame to be used for the images
        in the current pawprint structure.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_assign_sky_all(pawprint paw, int whichone) {
    int i,n;

    n = cpl_frameset_get_size(paw.current);
    for (i = 0; i < n; i++) 
        paw.whichsky[i] = whichone;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_crsky
    \par Purpose:
        Create and initialise a sky structure
    \par Description:
        Create a sky structure and initialise its elements
    \par Language:
        C
    \param algorithm
        The sky estimation algorith to be used
    \param frms
        The set of frames that will be used to estimate the sky
    \param template
        A template frame that can be used to save the sky
    \param snum
        The index number of the sky frame.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static skystruct vircam_sci_crsky(int algorithm, cpl_frameset *frms, 
                                  cpl_frame *template, int snum) {
    skystruct s;

    s.contrib = frms;
    if (ps.master_objmask != NULL) 
        s.objmask = cpl_frame_duplicate(ps.master_objmask);
    else
        s.objmask = NULL;
    s.skyalgo = algorithm;
    s.skyframe = NULL;
    s.template = template;
    vircam_sci_product_name((char *)cpl_frame_get_filename(cpl_frameset_get_position(frms,0)),
                            SKY_FILE,vircam_sci_config.prettynames,snum,
                            s.fname);
    return(s);
}
    
/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_merge_framesets
    \par Purpose:
        Merge a number of framesets into one 
    \par Description:
        A number of framesets are given. The frames from the input framesets
        are duplicated and merged into a single output frameset.
    \par Language:
        C
    \param n
        The number of input framesets
    \param in
        The input framesets
    \returns
        An output frameset with copies of all of the frames from the 
        input framesets.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static cpl_frameset *vircam_sci_merge_framesets(int n, cpl_frameset **in) {
    int i,j,nf;
    cpl_frame *frm;
    cpl_frameset *new;

    /* Get a new frameset */

    new = cpl_frameset_new();

    /* Loop for each input frameset. Insert each frame into the new one */

    for (i = 0; i < n; i++) {
        nf = cpl_frameset_get_size(in[i]);
        for (j = 0; j < nf; j++) {
            frm = cpl_frame_duplicate(cpl_frameset_get_position(in[i],j));
            cpl_frameset_insert(new,frm);
        }
    }
    return(new);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_tilesky
    \par Purpose:
        Create and save a sky frame using the tilesky algorithm
    \par Description:
        Given a list of files do a background estimation using the 
        tilesky algorithm. The sky frame will also be saved.
    \par Language:
        C
    \param framelist
        The input raw frameset
    \param parlist
        The input command line parameter list
    \param in
        The input images
    \param template
        A template to be used when saving the representative sky
    \param tag
        The product tag to be used when saving the file
    \param fname
        The name for the output sky file
    \param product_frame
        The product frame for the output sky file
    \retval 
        CASU_OK if everything went well. CASU_FATAL if there was a problem
        in one of the vircam_tilesky routine.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_tilesky(cpl_frameset *framelist,
                              cpl_parameterlist *parlist, cpl_frameset *in,
                              cpl_frame *template, char *fname,
                              cpl_frame **product_frame) {
    int j,nfiles,isfirst,status;
    const char *fctid="vircam_sci_tilesky";
    casu_fits **infits,*skyout;

    /* Initialise product frame */

    *product_frame = NULL;

    /* Loop for each extension */

    nfiles = cpl_frameset_get_size(in);
    cpl_msg_info(fctid,"Creating sky %s",fname);
    cpl_msg_indent_more();
    for (j = 1; j <= VIRCAM_NEXTN; j++) {
        cpl_msg_info(fctid,"Extension [%" CPL_SIZE_FORMAT "]",(cpl_size)j);
        isfirst = (j == 1);

        /* Load up input images for this extension and associated bad 
           pixel mask... */

        infits = casu_fits_load_list(in,CPL_TYPE_FLOAT,j);
        casu_mask_load(ps.mask,j,
                         (int)cpl_image_get_size_x(casu_fits_get_image(infits[0])),
                         (int)cpl_image_get_size_y(casu_fits_get_image(infits[0])));    
        /* Form the sky for this extension */

        status = CASU_OK;
        skyout = NULL;
        (void)vircam_tilesky(infits,nfiles,ps.mask,&skyout,&status);

        /* Save the sky frame */

        if (vircam_sci_save_sky(skyout,framelist,parlist,fname,template,
                                isfirst,product_frame) != CASU_OK) {
            freefitslist(infits,nfiles);
            freefits(skyout);
            return(CASU_FATAL);
        }

        /* Tidy up and move on to the next one */

        freefitslist(infits,nfiles);
        freefits(skyout);
    }
    cpl_msg_indent_less();
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name
        vircam_sci_pawsky
    \par Purpose:
        Create and save a sky frame using the pawsky_mask algorithm
    \par Description:
        Given a list of files do a background estimation using the 
        pawsky_mask algorithm. The sky frame will also be saved.
    \par Language:
        C
    \param framelist
        The input raw frameset
    \param parlist
        The input command line parameter list
    \param algo
        The sky algorithm. Can be PAWSKY_MASK or PAWSKY_MASK_PRE
    \param contrib
        The input images
    \param template
        A template to be used when saving the representative sky
    \param fname
        The name for the output sky file
    \param product_frame
        The product frame for the output sky file
    \retval 
        CASU_OK if everything went well. CASU_FATAL if there was a problem
        in one of the vircam_tilesky routine.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_pawsky_mask(cpl_frameset *framelist, 
                                  cpl_parameterlist *parlist, int algo,
                                  cpl_frameset *contrib, cpl_frame *template,
                                  char *fname, cpl_frame **product_frame) {
    int j,nfiles,status,isfirst;
    const char *fctid="vircam_sci_pawsky_mask";
    casu_fits **infits,*conf,*skyout,*opmfits,*skyvar;

    /* Output product frame */

    *product_frame = NULL;

    /* Loop for each extension */

    nfiles = cpl_frameset_get_size(contrib);
    cpl_msg_info(fctid,"Creating sky %s",fname);
    cpl_msg_indent_more();
    for (j = 1; j <= VIRCAM_NEXTN; j++) {
        cpl_msg_info(fctid,"Extension [%" CPL_SIZE_FORMAT "]",(cpl_size)j);
        isfirst = (j == 1);
        
        /* Load up input images for this extension + confidence map
           and associated bad pixel mask... */

        infits = casu_fits_load_list(contrib,CPL_TYPE_FLOAT,j);
        conf = casu_fits_load(ps.master_conf,CPL_TYPE_INT,j);
        casu_mask_load(ps.mask,j,
                       (int)cpl_image_get_size_x(casu_fits_get_image(conf)),
                       (int)cpl_image_get_size_y(casu_fits_get_image(conf)));

        /* Form the sky for this extension */

        status = CASU_OK;
        skyout = NULL;
        if (algo == PAWSKY_MASK) {
            (void)casu_pawsky_mask(infits,NULL,nfiles,conf,ps.mask,&skyout,
                                   &skyvar,psm.niter,psm.ipix,psm.thresh,
                                   psm.nbsize,psm.smkern,&status);
        } else {
            opmfits = casu_fits_load(ps.master_objmask,CPL_TYPE_INT,1);
            (void)casu_pawsky_mask_pre(infits,NULL,nfiles,ps.mask,opmfits,
                                       psm.nbsize,&skyout,&skyvar,&status);
            casu_fits_delete(opmfits);
        }

        /* Save the sky frame */

        if (vircam_sci_save_sky(skyout,framelist,parlist,fname,template,
                                isfirst,product_frame) != CASU_OK) {
            freefitslist(infits,nfiles); 
            freefits(conf);
            freefits(skyout);
            return(CASU_FATAL);
        }

        /* Tidy up and move on to the next one */

        freefitslist(infits,nfiles);
        freefits(conf);
        freefits(skyout);
    }
    cpl_msg_indent_less();
    return(CASU_OK);
}               

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_sky
    \par Purpose:
        Save the mean offset sky image products
    \par Description:
        The offsky frame image products are saved here.
    \par Language:
        C
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \return 
        0 if everything is ok
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_save_sky(casu_fits *outsky, cpl_frameset *framelist,
                               cpl_parameterlist *parlist, 
                               char *fname, cpl_frame *template, int isfirst,
                               cpl_frame **product_frame) {
    cpl_propertylist *p;
    int isdummy;
    cpl_image *fim;
    const char *fctid = "vircam_sci_save_sky";

    /* Work out which frame to base the output on. If this particular
       sequence failed for whatever reason, there will be a dummy sky frame. */

    fim = casu_fits_get_image(outsky);
    isdummy = (casu_fits_get_status(outsky) != CASU_OK);

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list */

    if (isfirst) {

        /* Create a new product frame object and define some tags */

        *product_frame = cpl_frame_new();
        cpl_frame_set_filename(*product_frame,fname);
        cpl_frame_set_tag(*product_frame,VIRCAM_PRO_OFFSET_SKY);
        cpl_frame_set_type(*product_frame,CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(*product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(*product_frame,CPL_FRAME_LEVEL_FINAL);

        /* Set up the PHU header. */

        p = casu_fits_get_phu(outsky);
        vircam_dfs_set_product_primary_header(p,*product_frame,framelist,
                                              parlist,vircam_recipename,
                                              "PRO-1.15",template,0);

        /* 'Save' the PHU image */                       

        if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,p,
                           CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(*product_frame);
            return(-1);
        }
        cpl_frameset_insert(framelist,*product_frame);
    }

    /* Get the extension property list */

    p = cpl_propertylist_duplicate(casu_fits_get_ehu(outsky));
    if (isdummy)
        casu_dummy_property(p);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(p,*product_frame,framelist,parlist,
                                        vircam_recipename,"PRO-1.15",template);
    if (cpl_image_save(fim,fname,CPL_TYPE_FLOAT,p,CPL_IO_EXTEND) != 
        CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product image extension");
        return(-1);
    }

    /* Quick tidy */

    cpl_propertylist_delete(p);

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_cmp_jit
    \par Purpose:
        Compare the jitter numbers of two frames
    \par Description:
        Compare the jitter numbers of two frames
    \par Language:
        C
    \param frame1
        The first frame
    \param frame2
        The second frame
    \retval
        The result of integer comparison of the two jitter numbers
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
        
static int vircam_sci_cmp_jit(const cpl_frame *frame1, 
                              const cpl_frame *frame2) {
    int j_1,j_2;
    cpl_propertylist *pp;

    /* Test entries */

    if (frame1 == NULL || frame2 == NULL)
        return(-1);
    pp = cpl_propertylist_load(cpl_frame_get_filename(frame1),0);
    (void)vircam_pfits_get_jitternum(pp,&j_1);
    cpl_propertylist_delete(pp);
    pp = cpl_propertylist_load(cpl_frame_get_filename(frame2),0);
    (void)vircam_pfits_get_jitternum(pp,&j_2);
    cpl_propertylist_delete(pp);

    return(j_1 == j_2);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_cmp_tstart
    \par Purpose:
        Compare the template start time of two frames
    \par Description:
        Compare the template start time of two frames
    \par Language:
        C
    \param frame1
        The first frame
    \param frame2
        The second frame
    \retval
        The result of string comparison of the two start times
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_cmp_tstart(const cpl_frame *frame1, 
                                 const cpl_frame *frame2) {
    char ts1[80],ts2[80];
    cpl_propertylist *pp;

    /* Test entries */

    if (frame1 == NULL || frame2 == NULL)
        return(-1);

    /* Load the propertylist for each frame and extract the template start
       time */

    pp = cpl_propertylist_load(cpl_frame_get_filename(frame1),0);
    (void)vircam_pfits_get_tplstart(pp,ts1);
    cpl_propertylist_delete(pp);
    pp = cpl_propertylist_load(cpl_frame_get_filename(frame2),0);
    (void)vircam_pfits_get_tplstart(pp,ts2);
    cpl_propertylist_delete(pp);

    /* Compare the start times */

    if (strcmp(ts1,ts2))
        return(0);
    else 
        return(1);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_update_frameset
    \par Purpose:
        Update a frameset by giving the frames new filenames
    \par Description:
        Filenames are given to each of the frames in a frameset depending
        upon whether the prettynames option is being used or not.
    \par Language:
        C
    \param frms
        The input frameset.
    \params nfrms_tot
        The total number of frames in the dataset so far
    \returns
        A new frameset which is a duplicate of the input frameset, but with
        the file names changed
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static cpl_frameset *vircam_sci_update_frameset(cpl_frameset *frms, int *nfrms_tot) {
    cpl_frameset *copy;
    cpl_size i,n;
    cpl_frame *fr;
    char *fname,bname[BUFSIZ];

    /* NULL input... */

    if (frms == NULL)
        return(NULL);

    /* First make a copy of the input frameset */

    copy = cpl_frameset_duplicate(frms);

    /* Now go through and change the file names */

    n = cpl_frameset_get_size(frms);
    for (i = 0; i < n; i++) {
        fr = cpl_frameset_get_position(copy,i);
        fname = cpl_strdup(cpl_frame_get_filename(fr));
        (*nfrms_tot)++;
        vircam_sci_product_name(fname,SIMPLE_FILE,vircam_sci_config.prettynames,
                                *nfrms_tot,bname);
        cpl_frame_set_filename(fr,basename(bname));
        cpl_free(fname);
    }

    /* Get out of here */

    return(copy);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_testfrms
    \par Purpose:
        Test a list of files to be sure it's what we expect
    \par Description:
        Test a list of files to be sure they contain what we expect.
    \par Language:
        C
    \param frms
        The input frameset.
    \param nextn_expected
        The number of expected extensions
    \param isimg
        Set if this is a frameset of images as opposed to a frameset of
        tables.
    \retval
        The number of frames in the frameset that failed the tests
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_testfrms(cpl_frameset *frms, int nextn_expected, 
                               int isimg) {
    int i,nf,nerr;
    cpl_frame *fr;

    /* Return immediately if given nonsense */

    if (frms == NULL) 
        return(0);

    /* Loop for each frame in the frameset */

    nf = cpl_frameset_get_size(frms);
    nerr = 0;
    for (i = 0; i < nf; i++) {
        fr = cpl_frameset_get_position(frms,i);
        nerr += vircam_sci_testfrm_1(fr,nextn_expected,isimg);
    }
    
    /* Return value */

    return(nerr);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_testfrm_1
    \par Purpose:
        Test a single file to be sure it's what we expect
    \par Description:
        Test a single file to be sure it contains what we expect. Check
        that there are the correct number of extensions and that each
        extension can be loaded. 
    \par Language:
        C
    \param fr
        The input frame
    \param nextn_expected
        The number of expected extensions
    \param isimg
        Set if this is a frameset of images as opposed to a frameset of
        tables.
    \retval
        The number of frames in the frameset that failed the tests
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_testfrm_1(cpl_frame *fr, int nextn_expected, int isimg) {
    int nextn,nerr,j;
    casu_fits *test;
    casu_tfits *testt;
    const char *fctid="vircam_sci_testfrm";

    /* Return immediately if given nonsense */

    if (fr == NULL)
        return(0);

    /* Test to see how many extensions there are and compare to see
       if it matches the number expected */

    nextn = cpl_frame_get_nextensions(fr);
    if (nextn != nextn_expected) {
        cpl_msg_error(fctid,"Frame %s has %" CPL_SIZE_FORMAT " extensions, expected %" CPL_SIZE_FORMAT "\n",
                      cpl_frame_get_filename(fr),(cpl_size)nextn,
                      (cpl_size)nextn_expected);
        return(1);
    }

    /* Test to see if you can load each of the extensions */

    nerr = 0;
    for (j = 1; j <= nextn; j++) {
        if (isimg) {
            test = casu_fits_load(fr,CPL_TYPE_FLOAT,j);
            if (test == NULL) {
                cpl_msg_error(fctid,
                              "Frame image %s[%" CPL_SIZE_FORMAT "] won't load\n",
                              cpl_frame_get_filename(fr),(cpl_size)j);
                nerr++;
                continue;
            }
            freefits(test);
        } else {
            testt = casu_tfits_load(fr,j);
            if (testt == NULL) {
                cpl_msg_error(fctid,
                              "Frame table %s[%" CPL_SIZE_FORMAT "] won't load\n",
                              cpl_frame_get_filename(fr),(cpl_size)j);
                nerr++;
                continue;
            }
            freetfits(testt);
        }
    }
    return(nerr);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_get_readnoise_gain
    \par Purpose:
        Get a readnoise and gain estimate
    \par Description:
        Get a readnoise and gain estimate. If one doesn't exist, then fill
        one in.
    \par Language:
        C
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_get_readnoise_gain(int jext, float *readnoise, 
                                          float *gain) {
    cpl_propertylist *p_rg;
    const char *fctid = "vircam_sci_get_readnoise_gain";

    /* Load the propertylist */

    p_rg = cpl_propertylist_load(cpl_frame_get_filename(ps.readgain_file),
                                 (cpl_size)jext);

    /* Check the readnoise property type and read it */

    switch (cpl_propertylist_get_type(p_rg,"ESO QC READNOISE")) {
    case CPL_TYPE_FLOAT:
        *readnoise = cpl_propertylist_get_float(p_rg,"ESO QC READNOISE");
        break;
    case CPL_TYPE_DOUBLE:
        *readnoise = (float)cpl_propertylist_get_double(p_rg,
                                                        "ESO QC READNOISE");
        break;
    default:
        cpl_error_reset();
        *readnoise = 25.0;
        cpl_msg_error(fctid,"Unable to get READNOISE estimate, guessing %g\n",
                      *readnoise);
    }

    /* Now the gain */

    switch (cpl_propertylist_get_type(p_rg,"ESO QC CONAD")) {
    case CPL_TYPE_FLOAT:
        *gain = cpl_propertylist_get_float(p_rg,"ESO QC CONAD");
        break;
    case CPL_TYPE_DOUBLE:
        *gain = (float)cpl_propertylist_get_double(p_rg,"ESO QC CONAD");
        break;
    default:
        cpl_error_reset();
        *gain = 1.0;
        cpl_msg_error(fctid,"Unable to get GAIN estimate, guessing %g\n",
                      *gain);
    }
    cpl_propertylist_delete(p_rg);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_simple
    \par Purpose:
        Save the simple image products
    \par Description:
        The simple image products are saved here.
    \par Language:
        C
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \return 
        0 if everything is ok
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_save_simple(casu_fits *obj, cpl_frameset *framelist,
                                  cpl_parameterlist *parlist, int isprod,
                                  cpl_frame *template, int isfirst,
                                  const char *tag, char *fname,
                                  cpl_frame **product_frame) {
    cpl_propertylist *plist;
    int isdummy;
    const char *fctid = "vircam_sci_save_simple";

    cpl_ensure(product_frame != NULL, CPL_ERROR_NULL_INPUT, -1);

    /* Get some information about this guy */

    isdummy = (casu_fits_get_status(obj) != CASU_OK);

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list. Get rid of the file if it already exists */

    if (isfirst) {
        if (access(fname,F_OK)) 
            remove(fname);

        /* Create a new product frame object and define some tags */

        *product_frame = cpl_frame_new();
        cpl_frame_set_filename(*product_frame,fname);

        /* Tag the product */

        cpl_frame_set_tag(*product_frame,tag);
        cpl_frame_set_type(*product_frame,CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(*product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(*product_frame,CPL_FRAME_LEVEL_FINAL);

        /* Set up the PHU header */

        plist = casu_fits_get_phu(obj);
        vircam_dfs_set_product_primary_header(plist,*product_frame,framelist,
                                              parlist,vircam_recipename,
                                              "PRO-1.15",template,1);

        /* 'Save' the PHU image */                       

        if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,plist,CPL_IO_DEFAULT) != 
            CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(*product_frame);
            return(-1);
        }
        if (isprod) 
            cpl_frameset_insert(framelist,*product_frame);
    }

    /* Get the extension property list */

    plist = casu_fits_get_ehu(obj);
    if (isdummy)
        casu_dummy_property(plist);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(plist,*product_frame,framelist,parlist,
                                        vircam_recipename,"PRO-1.15",template);
    if (cpl_image_save(casu_fits_get_image(obj),fname,CPL_TYPE_FLOAT,plist,
                       CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product image extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_stack
    \par Purpose:
        Save the simple image products
    \par Description:
        The simple image products are saved here.
    \par Language:
        C
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \return 
        0 if everything is ok
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_save_stack(casu_fits *stack, cpl_frameset *framelist,
                                 cpl_parameterlist *parlist,
                                 cpl_frame *template, int fnametype, 
                                 int stack_num, cpl_frame **product_frame) {
    cpl_propertylist *plist;
    int isfirst,isdummy;
    char fname[BUFSIZ],*base,*tname;
    const char *fctid = "vircam_sci_save_stack";

    /* Get some information about this guy */

    isdummy = (casu_fits_get_status(stack) != CASU_OK);
    isfirst = (*product_frame == NULL);
    tname = cpl_strdup(cpl_frame_get_filename(template));
    base = basename(tname);
    vircam_sci_product_name(base,STACK_FILE,fnametype,stack_num,fname);
    freespace(tname);

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list. Get rid of the file if it already exists */

    if (isfirst) {
        if (access(fname,F_OK)) 
            remove(fname);

        /* Create a new product frame object and define some tags */

        *product_frame = cpl_frame_new();
        cpl_frame_set_filename(*product_frame,fname);

        /* Tag the product */

        cpl_frame_set_tag(*product_frame,VIRCAM_PRO_JITTERED_SCI);
        cpl_frame_set_type(*product_frame,CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(*product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(*product_frame,CPL_FRAME_LEVEL_FINAL);

        /* Set up the PHU header */

        plist = casu_fits_get_phu(stack);
        vircam_dfs_set_product_primary_header(plist,*product_frame,
                                              framelist,parlist,
                                              vircam_recipename,
                                              "PRO-1.15",template,1);

        /* 'Save' the PHU image */                       

        if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,plist,
                           CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(*product_frame);
            return(-1);
        }
        cpl_frameset_insert(framelist,*product_frame);
    }

    /* Get the extension property list */

    plist = casu_fits_get_ehu(stack);
    if (isdummy)
        casu_dummy_property(plist);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(plist,*product_frame,framelist,
                                        parlist,vircam_recipename,
                                        "PRO-1.15",template);
    if (cpl_image_save(casu_fits_get_image(stack),fname,CPL_TYPE_FLOAT,
                       plist,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product image extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_stack_conf
    \par Purpose:
        Save the simple image products
    \par Description:
        The simple image products are saved here.
    \par Language:
        C
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \return 
        0 if everything is ok
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_save_stack_conf(casu_fits *stack, cpl_frameset *framelist,
                                      cpl_parameterlist *parlist, 
                                      cpl_frame *template, int fnametype,
                                      int stack_num, 
                                      cpl_frame **product_frame) {
    cpl_propertylist *plist;
    int isfirst,isdummy;
    char fname[BUFSIZ],*tname,*base;
    const char *fctid = "vircam_sci_save_stack_conf";

    /* Get some information about this guy */

    isdummy = (casu_fits_get_status(stack) != CASU_OK);
    isfirst = (*product_frame == NULL);
    tname = cpl_strdup(cpl_frame_get_filename(template));
    base = basename(tname);
    vircam_sci_product_name(base,STACK_CONF,fnametype,stack_num,fname);
    freespace(tname);

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list. Get rid of the file if it already exists */

    if (isfirst) {
        if (access(fname,F_OK)) 
            remove(fname);

        /* Create a new product frame object and define some tags */

        *product_frame = cpl_frame_new();
        cpl_frame_set_filename(*product_frame,fname);

        /* Tag the product */

        cpl_frame_set_tag(*product_frame,VIRCAM_PRO_CONF_SCI);
        cpl_frame_set_type(*product_frame,CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(*product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(*product_frame,CPL_FRAME_LEVEL_FINAL);

        /* Set up the PHU header */

        plist = casu_fits_get_phu(stack);
        vircam_dfs_set_product_primary_header(plist,*product_frame,framelist,
                                              parlist,vircam_recipename,
                                              "PRO-1.15",template,1);

        /* 'Save' the PHU image */                       

        if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,plist,
                           CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(*product_frame);
            return(-1);
        }
        cpl_frameset_insert(framelist,*product_frame);
    }

    /* Get the extension property list */

    plist = casu_fits_get_ehu(stack);
    if (isdummy)
        casu_dummy_property(plist);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(plist,*product_frame,framelist,
                                        parlist,vircam_recipename,
                                        "PRO-1.15",template);
    if (cpl_image_save(casu_fits_get_image(stack),fname,CPL_TYPE_INT,
                       plist,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product image extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_save_stack_cat
    \par Purpose:
        Save the simple image products
    \par Description:
        The simple image products are saved here.
    \par Language:
        C
    \param framelist
        The input recipe frameset
    \param parlist
        The input recipe parameter list
    \return 
        0 if everything is ok
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int vircam_sci_save_stack_cat(casu_tfits *stack, cpl_frameset *framelist,
                                     cpl_parameterlist *parlist, 
                                     cpl_frame *template, int fnametype,
                                     int stack_num, cpl_frame **product_frame) {
    cpl_propertylist *plist;
    int isfirst,isdummy;
    char fname[BUFSIZ],*tname,*base;
    const char *fctid = "vircam_sci_save_stack_cat";

    /* Get some information about this guy */

    isdummy = (casu_tfits_get_status(stack) != CASU_OK);
    isfirst = (*product_frame == NULL);
    tname = cpl_strdup(cpl_frame_get_filename(template));
    base = basename(tname);
    vircam_sci_product_name(base,STACK_CAT,fnametype,stack_num,fname);
    freespace(tname);

    /* If we need to make a PHU then do that now based on the first frame
       in the input frame list. Get rid of the file if it already exists */

    if (isfirst) {
        if (access(fname,F_OK)) 
            remove(fname);

        /* Create a new product frame object and define some tags */

        *product_frame = cpl_frame_new();
        cpl_frame_set_filename(*product_frame,fname);

        /* Tag the product */

        cpl_frame_set_tag(*product_frame,VIRCAM_PRO_OBJCAT_SCI);
        cpl_frame_set_type(*product_frame,CPL_FRAME_TYPE_TABLE);
        cpl_frame_set_group(*product_frame,CPL_FRAME_GROUP_PRODUCT);
        cpl_frame_set_level(*product_frame,CPL_FRAME_LEVEL_FINAL);

        /* Set up the PHU header */

        plist = casu_tfits_get_phu(stack);
        vircam_dfs_set_product_primary_header(plist,*product_frame,framelist,
                                              parlist,vircam_recipename,
                                              "PRO-1.15",NULL,0);

        /* 'Save' the PHU image */                       

        if (cpl_image_save(NULL,fname,CPL_TYPE_UCHAR,plist,
                           CPL_IO_DEFAULT) != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Cannot save product PHU");
            cpl_frame_delete(*product_frame);
            return(-1);
        }
        cpl_frameset_insert(framelist,*product_frame);
    }

    /* Get the extension property list */

    plist = casu_tfits_get_ehu(stack);
    if (isdummy)
        casu_dummy_property(plist);

    /* Fiddle with the header now */

    vircam_dfs_set_product_exten_header(plist,*product_frame,framelist,
                                        parlist,vircam_recipename,
                                        "PRO-1.15",NULL);
    if (cpl_table_save(casu_tfits_get_table(stack),NULL,
                       plist,fname,CPL_IO_EXTEND) != CPL_ERROR_NONE) {
        cpl_msg_error(fctid,"Cannot save product table extension -- %s",
                      cpl_error_get_message());
        return(-1);
    }

    /* Get out of here */

    return(0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_product_name
    \par Purpose:
        Set up an output product file name
    \par Description:
        An output file name is defined based on a template name,  a
        'predictable' name or a temporary file name.
    \par Language:
        C
    \param template
        If the pretty file name option is used or if this is a temporary
        file name, then the this will be used as a reference name.
    \param producttype
        The type of product you are writing out. These are enumerated above.
    \param nametype
        This is: 0 -> predictable names, 1 -> pretty names based on input
        file names or 2 -> temporary file names.
    \param fnumber
        If the predictable file names are going to be used, then this is 
        a number that is appended to a file name root.
    \param outfname
        The output file name
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_product_name(char *template, int producttype, 
                                    int nametype, int fnumber, char *outfname) {
    const char *esonames[] = {"exp_","exp_conf_","exp_cat_","stack_",
                              "stack_conf_","stack_cat_","sky_"};
    const char *suffix[] = {"_ex","_ex_conf","_ex_cat","_st","_st_conf",
                            "_st_cat","sky_"};
    char *fname,*bname,*dot;

    /* If the name type is the ESO predictable sort, then it's easy. Just
       use the esonames defined above and append the correct number */

    switch (nametype) {
    case 0:
        (void)sprintf(outfname,"%s%d.fits",esonames[producttype],fnumber);
        break;

    /* If this is a temporary file, then just append tmp_ to the template
       name and return that */

    case 2:
        fname = cpl_strdup(template);
        bname = basename(fname);
        (void)sprintf(outfname,"tmp_%s",bname);
        freespace(fname);
        break;

    /* Ok, we want a pretty name... */

    case 1:
        fname = cpl_strdup(template);
        bname = basename(fname);
        if (producttype != SKY_FILE) {
            (void)sprintf(outfname,"%s",bname);
            dot = strrchr(outfname,'.');
            (void)sprintf(dot,"%s.fits",suffix[producttype]);
        } else {
            sprintf(outfname,"%s%s",suffix[producttype],bname);
        }
        freespace(fname);
        break;

    /* something else ?? */

    default:
        (void)strcpy(outfname,"");
        break;
    }
    return;
}   

/**@}*/

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_init
    \par Purpose:
        Initialise pointers in the global memory structure.
    \par Description:
        Initialise pointers in the global memory structure.
    \par Language:
        C
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_init(void) {

    /* Level 0 stuff */

    ps.labels = NULL;
    ps.master_dark = NULL;
    ps.master_twilight_flat = NULL;
    ps.master_conf = NULL;
    ps.master_sky = NULL;
    ps.master_objmask = NULL;
    ps.mask = NULL;
    ps.chantab = NULL;
    ps.phottab = NULL;
    ps.tphottab = NULL;
    ps.readgain_file = NULL;
    ps.science_frames = NULL;
    ps.offset_skies = NULL;
    ps.product_frames_simple = NULL;
    ps.product_frames_simple_off = NULL;
    ps.phupaf = NULL;
    ps.gaincors = NULL;
    ps.catpath = NULL;
    ps.catname = NULL;
    ps.nscipaws = 0;
    ps.science_paws = NULL;
    ps.noffpaws = 0;
    ps.offsky_paws = NULL;
    ps.nskys = 0;
    ps.skys = NULL;

    /* Level 1 stuff */

    ps.fdark = NULL;
    ps.fflat = NULL;
    ps.fconf = NULL;
    ps.fsky = NULL;
    ps.fchantab = NULL;
    ps.nscience = 0;
    ps.sci_fits = NULL;
    ps.noffsets = 0;
    ps.offsky_fits = NULL;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_sci_tidy
    \par Purpose:
        Free allocated workspace in the global memory structure
    \par Description:
        Free allocated workspace in the global memory structure. The tidy works
        on two levels. Level 1 is for items that are usually cleared up after
        each extension is processed. Level 2 is for cleaning up the whole
        recipe
    \par Language:
        C
    \param level
        The level of the tidy to be done. 1: Tidy up after finishing an 
        extension, 2: Tidy up after finishing the recipe.
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void vircam_sci_tidy(int level) {
    int i;

    /* Level 1 stuff */

    freefits(ps.fdark); 
    freefits(ps.fflat); 
    freefits(ps.fconf);
    freefits(ps.fsky);
    freetfits(ps.fchantab);
    freefitslist(ps.sci_fits,ps.nscience);
    freefitslist(ps.offsky_fits,ps.noffsets);
    ps.nscience = 0;

    if (level == 1)
        return;
    
    /* Level 0 stuff */

    freespace(ps.labels);
    freeframe(ps.master_dark);
    freeframe(ps.master_twilight_flat);
    freeframe(ps.master_conf);
    freeframe(ps.master_sky);
    freeframe(ps.master_objmask);
    freemask(ps.mask);
    freeframe(ps.chantab);
    freeframe(ps.phottab);
    freeframe(ps.readgain_file);
    freetable(ps.tphottab);
    freeframeset(ps.science_frames);
    freeframeset(ps.offset_skies);
    freepropertylist(ps.phupaf);
    freespace(ps.product_frames_simple);     /* NB: We only have to delete */
                                             /* the arrays and not the frames */
                                             /* as these get passed back */
    freespace(ps.product_frames_simple_off); /* to esorex */
    freespace(ps.gaincors);
    freespace(ps.catpath);
    freespace(ps.catname);
    for (i = 0; i < ps.nskys; i++) {
        freeframeset(ps.skys[i].contrib);
        freeframe(ps.skys[i].objmask);
        freeframe(ps.skys[i].template);
    }
    freespace(ps.skys);
    freepaws(ps.science_paws,ps.nscipaws);
    freepaws(ps.offsky_paws,ps.noffpaws);
    freeframe(ps.schlf_n);
    freeframe(ps.schlf_s);
}


/*

$Log: not supported by cvs2svn $
Revision 1.9  2013/10/15 17:01:45  jim
new entry


*/
