/* $Id: vircam_grout.c,v 1.1 2013-10-15 16:30:07 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:30:07 $
 * $Revision: 1.1 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cpl.h>
#include <math.h>
#include <string.h>

#include <casu_utils.h>
#include <casu_wcsutils.h>
#include <casu_stats.h>

#include "vircam_mods.h"

/* This is specifically for VIRCAM files. Only correct the first 7 apertures */

#define NPAWS 6
#define NEXTN 16
#define NAPCOR 7

typedef struct {
    cpl_wcs       *wcs;
    int           nxout;
    int           nyout;
    float         apcors[NAPCOR];
    float         magzpt;
    float         seeing;
    float         ellipt;
    float         mjdobs;
} tables;

const char *apcorkeys[NAPCOR] = {"APCOR1","APCOR2","APCOR3","APCOR4","APCOR5",
                                 "APCOR6","APCOR7"};

static int gettabinfo(cpl_propertylist *plist, tables *tab, double *mjdobs);
static void tidytabs(tables maintab, tables *tabs);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Correct input tile catalogues 

    \par Name:
        vircam_grout
    \par Purpose:
        Correct input tile catalogues for the effects of variable seeing and
        magnitude zeropoints of the images that went into that tile.
    \par Description:
        Variable conditions during the observations of the 6 pawprints of
        a VIRCAM tile can lead to spatial systematics in the fluxes measured
        from a tiled image. This routine takes that effect out.
    \par Language:
        C
    \param intab 
        The input tile catalogue
    \param input_cats
        The catalogues for the input pawprints that formed the tile
    \param input_confs
        The confidence maps for the input pawprints that formed the tile
    \param outtab
        The output tile catalogue
    \param status 
        An input/output status that is the same as the returned values below.
    \retval CASU_OK 
        if everything is ok
    \retval CASU_FATAL 
        if image data fails to load
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_grout(casu_tfits *intab, cpl_frameset *input_cats, 
                        cpl_frameset *input_confs, casu_tfits **outtab,
                        int *status) {
    int retval,i,j,m,nrows,*conf_data,k,ix,iy,ind,k2,nc,mjd_day = 0;
    tables maintab,pawtabs[NPAWS*NEXTN];
    double *x,*y,*x2,*y2,expon,*dspace,mjdobs;
    cpl_array *colnames_array;
    cpl_frame *fr;
    cpl_propertylist *plist;
    casu_fits *conf_fits;
    float delta_mag,xmax,ymax,*mjds,*cnumb,*fspace,*fluxcor,*fluxes[NAPCOR+1];
    float *fluxerrs[NAPCOR+1],work[NEXTN],confwt,scale,*xptr,*yptr,*data;
    char **colnames,msg[81],sval[81];
    const char *fluxcols[NAPCOR+1] =  {"Aper_flux_1","Aper_flux_2",
                                       "Aper_flux_3","Aper_flux_4",
                                       "Aper_flux_5","Aper_flux_6",
                                       "Aper_flux_7","Peak_height"};
    const char *errcols[NAPCOR+1] =  {"Aper_flux_1_err","Aper_flux_2_err",
                                      "Aper_flux_3_err","Aper_flux_4_err",
                                      "Aper_flux_5_err","Aper_flux_6_err",
                                      "Aper_flux_7_err","Peak_height_err"};

    /* Inherited status */

    *outtab = NULL;
    if (*status != CASU_OK)
        return(*status);

    /* Make sure that this table hasn't already been grouted. If it has
       then just return with a copy of the input table */

    if (cpl_propertylist_has(casu_tfits_get_ehu(intab),"ESO DRS GROUTED")) {
        *outtab = casu_tfits_duplicate(intab);
        return(*status);
    }

    /* Make a copy of the input table which will then be updated */

    *outtab = casu_tfits_duplicate(intab);

    /* Initialise the WCS structures */
    
    maintab.wcs = NULL;
    for (i = 0; i < NPAWS*NEXTN; i++) 
        pawtabs[i].wcs = NULL;

    /* Read the WCS, apcors, etc from the input table */

    retval = gettabinfo(casu_tfits_get_ehu(*outtab),&maintab,&mjdobs);
    if (retval != CASU_OK) {
        *status = retval;
        return(*status);
    }

    /* Get some space to use for coordinate transformations */

    nrows = (int)cpl_table_get_nrow(casu_tfits_get_table(*outtab));
    dspace = cpl_malloc(4*nrows*sizeof(double));
    x = dspace;
    y = dspace + nrows;
    x2 = dspace + 2*nrows;
    y2 = dspace + 3*nrows;

    /* Get some space for an MJD column and for summations */

    fspace = cpl_calloc((2+NAPCOR)*nrows,sizeof(float));
    mjds = fspace;
    cnumb = fspace + nrows;
    fluxcor = fspace + 2*nrows;

    /* Get the coordinate information from the input table and convert to
       double */

    xptr = cpl_table_get_data_float(casu_tfits_get_table(*outtab),
                                    "X_coordinate");
    yptr = cpl_table_get_data_float(casu_tfits_get_table(*outtab),
                                    "Y_coordinate");
    for (i = 0; i < nrows; i++) {
        x[i] = (double)xptr[i];
        y[i] = (double)yptr[i];
    }

    /* Now read all the necessary information about the input tables */

    m = 0;
    for (i = 0; i < NPAWS; i++) {
        fr = cpl_frameset_get_position(input_cats,i);
        for (j = 0; j < NEXTN; j++) {
            plist = cpl_propertylist_load(cpl_frame_get_filename(fr),j+1);
            retval = gettabinfo(plist,&(pawtabs[m]),&mjdobs);
            if (retval != CASU_OK) {
                freespace(dspace);
                freespace(mjds);
                freetfits(*outtab);
                freepropertylist(plist);
                return(CASU_FATAL);
            }
            if (i == 0 && j == 0) 
                mjd_day = (int)mjdobs;
            pawtabs[m].mjdobs = (float)(1440.0*(mjdobs - (double)mjd_day));
            m++;
            freepropertylist(plist);
        }
    }

    /* Loop for each stack image extension */

    m = -1;
    for (i = 0; i < NPAWS; i++) {
        fr = cpl_frameset_get_position(input_confs,i);
        for (j = 0; j < NEXTN; j++) {
            conf_fits = casu_fits_load(fr,CPL_TYPE_INT,j+1);
            conf_data = cpl_image_get_data_int(casu_fits_get_image(conf_fits));
            m++;
            delta_mag = maintab.magzpt - pawtabs[m].magzpt;

            /* Shift the tile object coordinates to the reference frame of
               the current pawprint */

            casu_xytoxy_list(maintab.wcs,pawtabs[m].wcs,nrows,x,y,x2,y2);

            /* Loop through each object now and see if it's on this image.
               If it is, then work out the contribution to the flux correction
               that you get from the current image */

            xmax = (float)pawtabs[m].nxout - 0.5;
            ymax = (float)pawtabs[m].nyout;
            for (k = 0; k < nrows; k++) {
                if (x2[k] >= 0.5 && x2[k] < xmax && 
                    y2[k] >= 0.5 && y2[k] < ymax) {
                    ix = (int)(x2[k] + 0.5) - 1;
                    iy = (int)(y2[k] + 0.5) - 1;
                    ind = iy*pawtabs[m].nxout + ix;
                    confwt = (float)max(1.0,conf_data[ind]);
                    mjds[k] += confwt*pawtabs[m].mjdobs;
                    cnumb[k] += confwt;
                    for (k2 = 0; k2 < NAPCOR; k2++) {
                        expon = 0.4*(double)(pawtabs[m].apcors[k2] - 
                                     maintab.apcors[k2] + delta_mag);
                        fluxcor[k2*nrows+k] += confwt*pow(10.0,expon);
                    }
                }
            }
            freefits(conf_fits);
        }
    }

    /* Get pointers to each of the fluxes and their errors */

    for (i = 0; i < NAPCOR+1; i++) {
        fluxes[i] = cpl_table_get_data_float(casu_tfits_get_table(*outtab),
                                             fluxcols[i]);
        fluxerrs[i] = cpl_table_get_data_float(casu_tfits_get_table(*outtab),
                                               errcols[i]);
    }

    /* Now normalise the flux corrections and apply them */

    for (k = 0; k < nrows; k++) {
        for (k2 = 0; k2 < NAPCOR; k2++) {
            if (cnumb[k] != 0.0)
                scale = fluxcor[k2*nrows+k]/cnumb[k];
            else
                scale = 1.0;
            fluxes[k2][k] *= scale;
            fluxerrs[k2][k] *= scale;
            if (k2 == 0) {
                fluxes[NAPCOR][k] *= scale;
                fluxerrs[NAPCOR][k] *= scale;
            }
        }
        if (cnumb[k] != 0.0)
            mjds[k] /= cnumb[k];
    }

    /* Ok, we now have altered the fluxes, create a column for the MJD. Find
       the first column with the 'Blank' in its name. Then map the data array
       to it and copy the mjd array. */

    colnames_array = cpl_table_get_column_names(casu_tfits_get_table(*outtab));
    colnames = cpl_array_get_data_string(colnames_array);
    nc = cpl_array_get_size(colnames_array);
    i = 0;
    while (i < nc && strncmp(colnames[i],"Blank",5))
        i++;
    if (i < nc) {
        cpl_table_name_column(casu_tfits_get_table(*outtab),colnames[i],
                                                     "MJDoff");
        cpl_table_set_column_unit(casu_tfits_get_table(*outtab),"MJDoff",
                                  "Minutes");
        data = cpl_table_get_data_float(casu_tfits_get_table(*outtab),
                                        "MJDoff");
        memmove(data,mjds,nrows*sizeof(float));
    }
    cpl_array_delete(colnames_array);
        
    /* Set up some header items */

    plist = casu_tfits_get_ehu(*outtab);
    cpl_propertylist_append_bool(plist,"ESO DRS GROUTED",1);
    cpl_propertylist_set_comment(plist,"ESO DRS GROUTED",
                                 "Table has been grouted");
    cpl_propertylist_append_float(plist,"MJD_DAY",mjd_day);
    cpl_propertylist_set_comment(plist,"MJD_DAY","MJD day");

    /* Add a card with the mean magnitude zeropoint for each input pawprint */

    msg[0] = '\0';
    m = 0;
    for (i = 0; i < NPAWS; i++) {
        for (j = 0; j < NEXTN; j++) 
            work[j] = pawtabs[m++].magzpt;
        delta_mag = casu_med(work,NULL,NEXTN);
        (void)sprintf(sval,"%0.3f ",delta_mag);
        (void)strcat(msg,sval);
    }
    cpl_propertylist_append_string(plist,"ESO DRS PAWMAGZP",msg);
    cpl_propertylist_set_comment(plist,"ESO DRS PAWMAGZP",
                                 "Mag Zpt of input pawprints");

    /* Add a card with the mean ellipticity for each input pawprint */

    msg[0] = '\0';
    m = 0;
    for (i = 0; i < NPAWS; i++) {
        for (j = 0; j < NEXTN; j++) 
            work[j] = pawtabs[m++].ellipt;
        delta_mag = casu_med(work,NULL,NEXTN);
        (void)sprintf(sval,"%0.3f ",delta_mag);
        (void)strcat(msg,sval);
    }
    cpl_propertylist_append_string(plist,"ESO DRS PAWELLPT",msg);
    cpl_propertylist_set_comment(plist,"ESO DRS PAWELLPT",
                                 "Median paw ELLIPTIC");
    
    /* Add a card with the median seeing for each of the pawprints */

    m = 0;
    msg[0] = '\0';
    for (i = 0; i < NPAWS; i++) {
        for (j = 0; j < NEXTN; j++) 
            work[j] = pawtabs[m++].seeing;
        delta_mag = casu_med(work,NULL,NEXTN);
        (void)sprintf(sval,"%0.3f ",delta_mag);
        (void)strcat(msg,sval);
    }
    cpl_propertylist_append_string(plist,"ESO DRS PAWSEENG",msg);
    cpl_propertylist_set_comment(plist,"ESO DRS PAWSEENG",
                                 "[arcsec] Median paw SEEING");

    /* Free up some memory */

    freespace(fspace);
    freespace(dspace);
    tidytabs(maintab,pawtabs);

    /* Get out of here */

    *status = CASU_OK;
    return(*status);
}


static int gettabinfo(cpl_propertylist *plist, tables *tab, double *mjdobs) {
    int i;

    /* Get the WCS from the propertylist */

    tab->wcs = cpl_wcs_new_from_propertylist(plist);

    /* Get the aperture corrections */

    for (i = 0; i < NAPCOR; i++) 
        tab->apcors[i] = cpl_propertylist_get_float(plist,apcorkeys[i]);

    /* Get the magnitude zero point */

    tab->magzpt = cpl_propertylist_get_float(plist,"ESO QC MAGZPT");

    /* And the seeing */

    tab->seeing = cpl_propertylist_get_float(plist,"ESO DRS SEEING");

    /* And the ellipticity */

    tab->ellipt = cpl_propertylist_get_float(plist,"ESO QC ELLIPTICITY");

    /* And the mean MJD of the pawprint */

    *mjdobs = cpl_propertylist_get_float(plist,"MJD_MEAN");

    /* Finally the image sizes */

    tab->nxout = cpl_propertylist_get_int(plist,"ESO DRS NXOUT");
    tab->nyout = cpl_propertylist_get_int(plist,"ESO DRS NYOUT");
    return(CASU_OK);
}
    
static void tidytabs(tables maintab, tables *tabs) {
    int i;

    freewcs(maintab.wcs);
    for (i = 0; i < NEXTN*NPAWS; i++) 
        freewcs(tabs[i].wcs);
}

/**@}*/

/*

$Log: not supported by cvs2svn $

*/
