/* $Id: vircam_mosaic.c,v 1.1 2013-10-15 16:30:07 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:30:07 $
 * $Revision: 1.1 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <string.h>
#include <libgen.h>
#include <cpl.h>
#include <casu_utils.h>
#include <casu_stats.h>
#include <casu_wcsutils.h>

#include "vircam_mods.h"

typedef struct {
    double     crval[2];
    double     crpix[2];
    double     cd[4];
    double     secd;
    double     tand;
    double     pv21;
    double     pv23;
    double     pv25;
    int        nx;
    int        ny;
    int        istan;
} mywcs;

typedef struct {
    cpl_frame       *infile;
    cpl_frame       *conf;
    float           global_sky;
    float           global_noise;
    double          tpa;
    double          tpd;
    float           exptime;

    int             nhdus;
    mywcs           **wcs_extn;
    int             *nx_extn;
    int             *ny_extn;
    float           *magzpt_extn;

    int             cur_extn;
    float           *sky_extn;
    float           *scalefac;
    float           *skydiff;
} dstrct;

typedef struct {
    cpl_image      *outim;
    cpl_image      *outcim;
    int            nxo;
    int            nyo;
    mywcs          *outwcs;
    cpl_image      *work;
} outstrct;

static dstrct *fileptrs = NULL;

static void prov(int interp, casu_fits *file, cpl_frameset *fset);
static void writewcs(casu_fits *fitsfile, mywcs *o);
static void average(dstrct *dd, int nextn, float *data, int *cdata, 
                    outstrct *outstr, int interp, int conflim);
static void output_images(int nimages, outstrct *outstr);
static mywcs *dupwcs(mywcs *in);
static float distort_corr(double x, double y, mywcs *w);
static void outloc(mywcs *win, double xin, double yin, mywcs *wout, 
                   double *xout, double *yout);
static void xytord(mywcs *win, double xin, double yin, double *ra, 
                   double *dec);
/* static void rdtoxy(mywcs *win, double ra, double dec, double *x, double *y); */
static void dstrct_init(int nfiles);
static void dstrct_extn_init(dstrct *dd, int nhdu);
static int dstrct_extn_fill_from_plist(dstrct *dd, int extnum, 
                                       const char *expkey,
                                       cpl_propertylist *plist, char *msg);
static void dstrct_close(int nfiles);
static void normal(outstrct *o);

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Create a vircam mosaic

    \par Name:
        vircam_mosaic
    \par Purpose:
        Create a mosaic from a list of vircam pawprints
    \par Description:
        A mosaic is generated from (nominally six) pawprints of a vircam
        observation. The backgrounds are matched by scaling to exposure time
        (which should in theory be all the same, but maybe not...) and by
        photometric zeropoint. This requires the header item ESO QC MAGZPT
        to be present. If this is missing then this counts as a fatal error.
        If they are present, but if any one of them is set to zero, then
        a warning message is issued but the stacking continues without
        the photometric scaling.
    \par Language:
        C
    \param infiles
        A frameset with the input stacks
    \param inconf
        A frameset with the input confidence maps
    \param interp
        The interpolation algorithm: 0 => nearest neighbour, 
        1 => bilinear interpolation
    \param skyflag
        A flag for how you want the sky. 0 => don't adjust the input skies
        at all, 1 => set the background sky to the value of skywish_in,
        2 => set the background sky to the average background for each 
        input image
    \param expkey
        The FITS keyword for the exposure time
    \param conflim
        A lower limit to be placed on the input confidence. Pixels with 
        confidence below this limit will be ignored.
    \param out
        The output mosaic
    \param outc
        The output confidence map
    \param status
        An input/output status that is the same as the returned values below.
    \retval CASU_OK
        if everything is ok
    \retval CASU_FATAL
        if there was some error detected in the input images
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_mosaic(cpl_frameset *infiles, cpl_frameset *inconf,
                         int interp, int skyflag, float skywish_in, 
                         const char *expkey, int conflim, casu_fits **out,
                         casu_fits **outc, int *status) {
    cpl_size i,nimages,nconfs;
    int nerrs,nhdu,j,nstart,nalloc,npts,*cdata,jj,ind1,ind,ii,sczpt;
    cpl_frame *fr,*frc;
    outstrct outstr;
    dstrct *dd;
    cpl_propertylist *plist1,*plist2;
    char msg[BUFSIZ];
    float *alldata,back,noise,skywish,zpt1,exp1,sky1,scalefac,ds,*data,distort;
    double x,y;
    cpl_image *im,*cim;
    casu_fits *ref;
    const char *fctid="vircam_mosaic";

    /* Inherited status */

    *out = NULL;
    *outc = NULL;
    if (*status != CASU_OK)
        return(*status);

    /* Quick sanity check */

    nimages = cpl_frameset_get_size(infiles);
    nconfs = cpl_frameset_get_size(inconf);
    if (nimages <= 0) {
        cpl_msg_error(fctid,"No images to combine");
        FATAL_ERROR
    }
    if (nconfs != 1 && nconfs != nimages) {
        cpl_msg_error(fctid,"Number of conf maps must be 1 or nimages");
        FATAL_ERROR
    }

    /* Get some space for the file descriptors */

    dstrct_init(nimages);

    /* Loop through the input files and make sure we can open them all. 
       Compare the data array sizes in each of the extensions to those in
       the relevant confidence maps. */

    nerrs = 0;
    for (i = 0; i < nimages; i++) {
        dd = fileptrs + i;
        fr = cpl_frameset_get_position(infiles,i);
        dd->infile = fr;
        if (nconfs == 1) 
            frc = cpl_frameset_get_position(inconf,0);
        else
            frc = cpl_frameset_get_position(inconf,i);
        dd->conf = frc;
        
        /* Start by checking the number of extensions in each */

        if (cpl_frame_get_nextensions(fr) != cpl_frame_get_nextensions(frc)) {
            cpl_msg_error(fctid,"Number of extensions in %s and %s must match",
                          cpl_frame_get_filename(fr),
                          cpl_frame_get_filename(frc));
            nerrs++;
            continue;
        }
        
        /* Now loop through each extension. Load both the image and confidence
           map headers and make sure the image dimensions match */

        nhdu = (int)cpl_frame_get_nextensions(fr);
        dstrct_extn_init(dd,nhdu);
        for (j = 1; j <= nhdu; j++) {
            plist1 = cpl_propertylist_load(cpl_frame_get_filename(fr),j);
            if (cpl_error_get_code() != 0) {
                nerrs++;
                cpl_msg_error(fctid,"Unable to get header from %s[%d]",
                              cpl_frame_get_filename(fr),j);
                continue;
            }
            plist2 = cpl_propertylist_load(cpl_frame_get_filename(frc),j);
            if (cpl_error_get_code() != 0) {
                nerrs++;
                cpl_msg_error(fctid,"Unable to get header from %s[%d]",
                              cpl_frame_get_filename(frc),j);
                cpl_propertylist_delete(plist1);
                continue;
            }
            if (cpl_propertylist_get_int(plist1,"NAXIS1") !=
                cpl_propertylist_get_int(plist2,"NAXIS1") || 
                cpl_propertylist_get_int(plist1,"NAXIS2") !=
                cpl_propertylist_get_int(plist2,"NAXIS2")) {
                nerrs++;
                cpl_msg_error(fctid,"%s[%d] and %s[%d] image dimensions don't match",
                              cpl_frame_get_filename(fr),j,
                              cpl_frame_get_filename(frc),j);
                cpl_propertylist_delete(plist1);
                cpl_propertylist_delete(plist2);
                continue;
            }
            
            /* Get information from extension headers */

            if (dstrct_extn_fill_from_plist(dd,j,expkey,plist1,msg) != CASU_OK) {
                nerrs++;
                cpl_msg_error(fctid,"Error reading header in %s[%d] -- %s",
                              cpl_frame_get_filename(fr),j,msg);
                cpl_propertylist_delete(plist1);
                cpl_propertylist_delete(plist2);
                continue;
            }
            cpl_propertylist_delete(plist1);
            cpl_propertylist_delete(plist2);
        }
    }
    if (nerrs != 0) {
        cpl_msg_error(fctid,"%d fatal errors",nerrs);
        FATAL_ERROR
    }

    /* Loop through the files again and get some background stats */

    for (i = 0; i < nimages; i++) {
        nstart = 0;
        nalloc = 0;
        dd = fileptrs + i;
        alldata = NULL;
        for (j = 1; j <= dd->nhdus; j++) {
    
            /* Get some space to hold all the data from all extensions */

            nstart = nalloc;
            npts = (dd->nx_extn[j-1])*(dd->ny_extn[j-1]);
            nalloc += npts;
            alldata = cpl_realloc(alldata,nalloc*sizeof(float));

            /* Load the current image */

            im = cpl_image_load(cpl_frame_get_filename(dd->infile),
                                CPL_TYPE_FLOAT,0,j);
            memmove(alldata+nstart,cpl_image_get_data_float(im),
                    npts*sizeof(float));
            
            /* Get the stats for the current image */

            casu_qmedsig(cpl_image_get_data_float(im),NULL,(long)npts,
                         3.0,3,-1000.0,65535.0,&back,&noise);
            dd->sky_extn[j-1] = back;
            cpl_image_delete(im);
        }

        /* Get the global background for the whole file */

        casu_qmedsig(alldata,NULL,(long)nalloc,3.0,3,-1000.0,65535.0,
                     &back,&noise);
        dd->global_sky = back;
        dd->global_noise = noise;
        freespace(alldata);
    }

    /* Now adjust some of the background estimates depending on what's been
       requested */

    if (skyflag == 2) {
        skywish = 0.0;
        for (i = 0; i < nimages; i++) 
            skywish += (fileptrs+i)->global_sky;
        skywish /= (float)nimages;
    } else if (skyflag == 1) {
        skywish = skywish_in;
    } else {
        skywish = 0.0;
    }
    if (skyflag != 0) {
        for (i = 0; i < nimages; i++) 
            (fileptrs+i)->global_sky = skywish;
    }

    /* Check the magnitude zeropoints. If there are any zeros, then cancel 
       the whole idea of scaling by zeropoint */

    sczpt = 1;
    for (i = 0; i < nimages; i++) {
        dd = fileptrs + i;
        for (j = 1; j <= dd->nhdus; j++) {
            if (dd->magzpt_extn[j-1] == 0.0) {
                sczpt = 0;
                break;
            }
        }
        if (! sczpt)
            break;
    }
    if (! sczpt) 
        cpl_msg_warning(fctid,"Missing magnitude zeropoint -- no photometric scaling done");

    /* Finally work out scale factors and sky offsets */

    zpt1 = fileptrs->magzpt_extn[0];
    exp1 = fileptrs->exptime;
    sky1 = fileptrs->global_sky;
    for (i = 0; i < nimages; i++) {
        dd = fileptrs + i;
        for (j = 1; j <= dd->nhdus; j++) {
            scalefac = (exp1/dd->exptime);
            if (sczpt) 
                scalefac *= pow(10.0,0.4*(zpt1-dd->magzpt_extn[j-1]));
            ds = sky1 - scalefac*(dd->global_sky);
            dd->scalefac[j-1] = scalefac;
            dd->skydiff[j-1] = ds;
        }
    }

    /* Now work out the dimensions of the output image and create it */

    output_images((int)nimages,&outstr);

    /* Loop through each extension now and do the summation */

    for (i = 0; i < nimages; i++) {
        dd = fileptrs + i;
        for (j = 1; j <= dd->nhdus; j++) {
            im = cpl_image_load(cpl_frame_get_filename(dd->infile),
                                CPL_TYPE_FLOAT,0,(cpl_size)j);
            data = cpl_image_get_data_float(im);
            cim = cpl_image_load(cpl_frame_get_filename(dd->conf),
                                CPL_TYPE_INT,0,(cpl_size)j);
            cdata = cpl_image_get_data_int(cim);
            
            /* Modify the data array so that they are all on the same
               system as the first image. Also take out astrometric
               distortion */

            if (skyflag != 0) {
                for (jj = 0; jj < dd->ny_extn[j-1]; jj++) {
                    ind1 = jj*(dd->nx_extn[j-1]);
                    y = (double)(jj + 1);
                    for (ii = 0; ii < dd->nx_extn[j-1]; ii++) {
                        ind = ind1 + ii;
                        x = (double)(ii + 1);
                        distort = distort_corr(x,y,dd->wcs_extn[j-1]);
                        data[ind] = (data[ind] - dd->sky_extn[j-1])*distort + 
                            skywish;
                    }
                }
            }

            /* Now average it in */

            average(dd,j,data,cdata,&outstr,interp,conflim);
            cpl_image_delete(im);
            cpl_image_delete(cim);
        }
    }

    /* Normalise the output maps */

    normal(&outstr);

    /* Wrap the result */

    ref = casu_fits_load(fileptrs->infile,CPL_TYPE_FLOAT,1);
    *out = casu_fits_wrap(outstr.outim,ref,NULL,NULL);
    casu_fits_delete(ref);
    ref = casu_fits_load(fileptrs->conf,CPL_TYPE_INT,1);
    *outc = casu_fits_wrap(outstr.outcim,ref,NULL,NULL);
    casu_fits_delete(ref);

    /* Write the WCS in both */

    writewcs(*out,outstr.outwcs);
    writewcs(*outc,outstr.outwcs);

    /* Write the provenance */

    prov(interp,*out,infiles);
    prov(interp,*outc,infiles);

    /* Add a DRIBBLE keyword */

    if (interp != 0) {
        cpl_propertylist_update_string(casu_fits_get_ehu(*out),"DRIBBLE",
                                       "bilinear");
        cpl_propertylist_set_comment(casu_fits_get_ehu(*out),"DRIBBLE",
                                       "Interpolation method");
    }

    /* Tidy and exit */

    freespace(outstr.outwcs);
    cpl_image_delete(outstr.work);
    dstrct_close(nimages);
    return(CASU_OK);
}

/**@}*/

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        prov
    \par Purpose:
        Write the provenance information to the header
    \par Description:
        Provenance information for is written to the header
    \par Language:
        C
    \param interp
        The interpolation scheme used
    \param file
        The output file 
    \param fset
        The frameset for the input data
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void prov(int interp, casu_fits *file, cpl_frameset *fset) {
    int nfset,ncard,i;
    cpl_propertylist *ehu;
    char value[81],key[9],comment[81];
    const char *algorithm[] = {"Nearest-neighbour","Bi-linear interpolation"};

    /* Start by removing any pre-existing PROV keywords */

    ehu = casu_fits_get_ehu(file);
    cpl_propertylist_erase_regexp(ehu,"PROV[0-9]+",0);
    
    /* Add the zeroth card */

    ncard = 0;
    (void)snprintf(key,9,"PROV%04d",ncard);
    (void)snprintf(value,81,"File formed from %s of",algorithm[interp]);
    (void)snprintf(comment,81,"File combination algorithm");
    cpl_propertylist_update_string(ehu,key,value);
    cpl_propertylist_set_comment(ehu,key,comment);

    /* Now add a card for each input file */

    nfset = cpl_frameset_get_size(fset);
    for (i = 0; i < nfset; i++) {
        ncard++;
        (void)snprintf(key,9,"PROV%04d",ncard);
        (void)snprintf(value,81,"%s",
                       basename((char *)cpl_frame_get_filename(cpl_frameset_get_position(fset,i))));
        (void)snprintf(comment,81,"Card # %d",ncard);
        cpl_propertylist_update_string(ehu,key,value);
        cpl_propertylist_set_comment(ehu,key,comment);
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        writewcs
    \par Purpose:
        Write WCS information to a header
    \par Description:
        The world coordinate system information is written to the FITS
        header of the output file.
    \par Language:
        C
    \param fitsfile
        The file to which the info will be written
    \param o
        The WCS descriptor
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void writewcs(casu_fits *fitsfile, mywcs *o) {
    char key[16],sval[16],proj[4];
    int i,j,n;
    float val;
    double dval;
    cpl_propertylist *ehu;

    /* Get a convenient reference */

    ehu = casu_fits_get_ehu(fitsfile);
    cpl_propertylist_update_int(ehu,"NAXIS1",
                                (int)cpl_image_get_size_x(casu_fits_get_image(fitsfile)));
    cpl_propertylist_update_int(ehu,"NAXIS2",
                                (int)cpl_image_get_size_y(casu_fits_get_image(fitsfile)));

    /* Write out CTYPE */

    if (o->istan)
        (void)strcpy(proj,"TAN");
    else 
        (void)strcpy(proj,"ZPN");
    for (i = 1; i <= 2; i++) {
        (void)sprintf(key,"CTYPE%d",i);
        if (i == 1) 
            (void)sprintf(sval,"RA---%s",proj);
        else
            (void)sprintf(sval,"DEC--%s",proj);
        (void)cpl_propertylist_update_string(ehu,key,sval);
    }

    /* CRVAL, CRPIX */

    for (i = 1; i <= 2; i++) {
        (void)sprintf(key,"CRVAL%d",i);
        dval = (o->crval[i-1])*DEGRAD;
        cpl_propertylist_update_double(ehu,key,dval);
    }
    for (i = 1; i <= 2; i++) {
        (void)sprintf(key,"CRPIX%d",i);
        dval = (o->crpix[i-1]);
        cpl_propertylist_update_double(ehu,key,dval);
    }

    /* CD matrix */

    n = 0;
    for (j = 1; j <= 2; j++) {
        for (i = 1; i <= 2; i++) {
            (void)sprintf(key,"CD%d_%d",j,i);
            dval = (o->cd[n])*DEGRAD;
            cpl_propertylist_update_double(ehu,key,dval);
            n++;
        }
    }

    /* PV vector */

    if (! o->istan) {
        val = (float)(o->pv21);
        cpl_propertylist_update_float(ehu,"PV2_1",val);
        val = (float)(o->pv23);
        cpl_propertylist_update_float(ehu,"PV2_3",val);
        val = (float)(o->pv25);
        cpl_propertylist_update_float(ehu,"PV2_5",val);
    } else {
        if (cpl_propertylist_has(ehu,"PV2_1")) 
            cpl_propertylist_erase(ehu,"PV2_1");
        if (cpl_propertylist_has(ehu,"PV2_3")) 
            cpl_propertylist_erase(ehu,"PV2_3");
        if (cpl_propertylist_has(ehu,"PV2_5")) 
            cpl_propertylist_erase(ehu,"PV2_5");
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        average
    \par Purpose:
        Do the averaging of the data into the output image
    \par Description:
        The world coordinate system of the input/output files is used
        to create a mapping of input pixels to the output grid. Then the
        appropriate interplation scheme is used to add the contribution 
        of the input pixels to the output grid.
    \par Language:
        C
    \param dd
        The input file descriptor
    \param nextn
        The extension being used.
    \param data
        The input data array
    \param cdata
        The input confidence map
    \param outstr
        The output data descriptor
    \param interp
        The interpolation scheme
    \param conflim
        The lower limit on confidence below which pixels will be ignored.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void average(dstrct *dd, int nextn, float *data, int *cdata, 
                    outstrct *outstr, int interp, int conflim) {
    int i,j,ind1,ind,outind,j_1,j_2,i_1,i_2,jj,ii,ind1o,indo,ixo,iyo;
    float *odata,*work,wi,wj,scf,fcd,ds;
    double x_in,y_in,x_out,y_out;

    /* Set up some convenience variables */

    odata = cpl_image_get_data_float(outstr->outim);
    scf = dd->scalefac[nextn-1];
    ds = dd->skydiff[nextn-1];

    /* Get some workspace for the sum of the confidence */

    work = cpl_image_get_data_float(outstr->work);

    /* Have separate loops for nearest neighbour and interpolation to
       speed things up */

    if (interp == 0) {
        for (j = 0; j < dd->ny_extn[nextn-1]; j++) {
            ind1 = j*dd->nx_extn[nextn-1];
            y_in = (double)(j+1);
            for (i = 0; i < dd->nx_extn[nextn-1]; i++) {
                ind = ind1 + i;
                if (cdata[ind] <= conflim)
                    continue;
                x_in = (double)(i+1);
                outloc(dd->wcs_extn[nextn-1],x_in,y_in,outstr->outwcs,
                       &x_out,&y_out);
                ixo = casu_nint(x_out);
                iyo = casu_nint(y_out);
                if (ixo < 1 || ixo > outstr->nxo || iyo < 1 || 
                    iyo > outstr->nyo)
                    continue;
                outind = (iyo-1)*outstr->nxo + ixo - 1;
                fcd = (float)cdata[ind];
                odata[outind] += fcd*(data[ind]*scf + ds);
                work[outind] += fcd;
            }
        }
    } else {
        for (j = 0; j < dd->ny_extn[nextn-1]; j++) {
            ind1 = j*dd->nx_extn[nextn-1];
            y_in = (double)(j+1);
            for (i = 0; i < dd->nx_extn[nextn-1]; i++) {
                ind = ind1 + i;
                if (cdata[ind] <= conflim)
                    continue;
                x_in = (double)(i+1);
                outloc(dd->wcs_extn[nextn-1],x_in,y_in,outstr->outwcs,
                       &x_out,&y_out);
                if (x_out < 1.0 || y_out < 1.0 || x_out >= outstr->nxo || y_out >= outstr->nyo) {
/*                  fprintf(stderr,"%d %d %g %g %g %g\n",i,j,x_in,y_in,x_out,y_out); */
                    
/*                  xytord(dd->wcs_extn,x_in,y_in,&x_out,&y_out); */
/*                  fprintf(stderr,"%g %g***\n",x_out,y_out); */
/*                  exit(1); */
                    continue;
                }
                j_1 = (int)y_out;
                j_2 = j_1 + 1;
                i_1 = (int)x_out;
                i_2 = i_1 + 1;
                fcd = (float)cdata[ind];
                for (jj = j_1; jj <= j_2; jj++) {
                    wj = 1.0 - fabs(y_out - (double)jj);
                    ind1o = (jj - 1)*outstr->nxo;
                    for (ii = i_1; ii <= i_2; ii++) {
                        indo = ind1o + ii - 1;
                        wi = 1.0 - fabs(x_out - (double)ii);
                        odata[indo] += wi*wj*fcd*(scf*data[ind] + ds);
                        work[indo] += wi*wj*fcd;
                    }
                }
            }
        }
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        output_images
    \par Purpose:
        Create the output image structures
    \par Description:
        The WCS of the input images are examined and used to create the
        grid for the output image and confidence map.
    \par Language:
        C
    \param nimages
        The number of input images
    \param outstr
        The output descriptor
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void output_images(int nimages, outstrct *outstr) {
    double lowerleft[2],upperleft[2],lowerright[2],upperright[2];
    double xout,yout,xmin,ymin,ra,dec;
    float *data;
    int i,*idata,j,nx,ny,ixo,iyo;
    dstrct *dd;
    mywcs *refwcs,*curwcs,*outwcs;

    /* Initialise the corners */

    lowerleft[0] = 1.0e8;
    lowerleft[1] = 1.0e8;
    upperleft[0] = 1.0e8;
    upperleft[1] = -1.0e8;
    lowerright[0] = -1.0e8;
    lowerright[1] = 1.0e8;
    upperright[0] = -1.0e8;
    upperright[1] = -1.0e8;

    /* Loop for all images relative to the first one */

    refwcs = NULL;
    for (i = 0; i < nimages; i++) {
        dd = fileptrs + i;

        /* Open the image and loop for each hdu */

        for (j = 1; j <= dd->nhdus; j++) {
            curwcs = dd->wcs_extn[j-1];
            if (i == 0 && j == 1) {
                refwcs = dupwcs(curwcs);
                refwcs->istan = 1;
            }
            nx = dd->nx_extn[j-1];
            ny = dd->ny_extn[j-1];
            outloc(curwcs,(double)1.0,(double)1.0,refwcs,&xout,&yout);
            lowerleft[0] = min(lowerleft[0],xout);
            lowerleft[1] = min(lowerleft[1],yout);
            outloc(curwcs,(double)1.0,(double)ny,refwcs,&xout,&yout);
            upperleft[0] = min(upperleft[0],xout);
            upperleft[1] = max(upperleft[1],yout);
            outloc(curwcs,(double)nx,(double)1.0,refwcs,&xout,&yout);
            lowerright[0] = max(lowerright[0],xout);
            lowerright[1] = min(lowerright[1],yout);
            outloc(curwcs,(double)nx,(double)ny,refwcs,&xout,&yout);
            upperright[0] = max(upperright[0],xout);
            upperright[1] = max(upperright[1],yout);
        }
    }

    /* Ok, what are the limits? */

    ixo = casu_nint(max(lowerright[0]-lowerleft[0],
                        upperright[0]-upperleft[0])) + 5;
    iyo = casu_nint(max(upperright[1]-lowerright[1],
                        upperleft[1]-lowerleft[1])) + 5;
    xmin = min(lowerleft[0],upperleft[0]);
    ymin = min(lowerleft[1],lowerright[1]);
    outstr->nxo = ixo;
    outstr->nyo = iyo;

    /* Create the output images and work image. Make sure they are zeroed
       because they're going to be used in summations */

    outstr->outim = cpl_image_new((cpl_size)ixo,(cpl_size)iyo,CPL_TYPE_FLOAT);
    data = cpl_image_get_data_float(outstr->outim);
    memset(data,0,ixo*iyo*sizeof(float));
    outstr->work = cpl_image_new((cpl_size)ixo,(cpl_size)iyo,CPL_TYPE_FLOAT);
    data = cpl_image_get_data_float(outstr->work);
    memset(data,0,ixo*iyo*sizeof(float));
    outstr->outcim = cpl_image_new((cpl_size)ixo,(cpl_size)iyo,CPL_TYPE_INT);
    idata = cpl_image_get_data_int(outstr->outcim);
    memset(idata,0,ixo*iyo*sizeof(int));
        
    /* Update the reference point for the WCS */

    outwcs = cpl_malloc(sizeof(mywcs));
    outwcs->crpix[0] = refwcs->crpix[0] - xmin + 1.0;
    outwcs->crpix[1] = refwcs->crpix[1] - ymin + 1.0;
    outwcs->crval[0] = refwcs->crval[0];
    outwcs->crval[1] = refwcs->crval[1];
    for (i = 0; i < 4; i++)
        outwcs->cd[i] = refwcs->cd[i];
    outwcs->pv21 = 1.0;
    outwcs->pv23 = -0.33333333333;
    outwcs->pv25 = 0.0;
    outwcs->nx = ixo;
    outwcs->ny = iyo;
    outwcs->istan = 1;
    outwcs->tand = tan(outwcs->crval[1]);
    outwcs->secd = 1.0/cos(outwcs->crval[1]);
    outstr->outwcs = outwcs;

    /* Now see what the RA,Dec of the centre of the map is and shift the
       reference point to that place */

    xout = 0.5*(double)ixo;
    yout = 0.5*(double)iyo;
    xytord(outwcs,xout,yout,&ra,&dec);
    outwcs->crpix[0] = xout;
    outwcs->crpix[1] = yout;
    outwcs->crval[0] = ra/DEGRAD;
    outwcs->crval[1] = dec/DEGRAD;
    outwcs->tand = tan(outwcs->crval[1]);
    outwcs->secd = 1.0/cos(outwcs->crval[1]);
    freespace(refwcs);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        distort_corr
    \par Purpose:
        Work out the pixel size distortion based on the camera projection 
        geometry
    \par Description:
        The size of a pixel will vary depending upon its location within
        the camera projection. This routine calculates the ratio of the
        pixel size at a give position over the pixel size at the reference
        point.
    \par Language:
        C
    \param x
        The X position of the pixel
    \param y
        The Y position of the pixel
    \param w
        The WCS descriptor
    \return
        The output ratio
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static float distort_corr(double x, double y, mywcs *w) {
    double xi,xn,r,rprime,drprime_bydr,dc,tanr;

    xi = (x - w->crpix[0])*(w->cd[0]) + (y - w->crpix[1])*(w->cd[1]);
    xn = (x - w->crpix[0])*(w->cd[2]) + (y - w->crpix[1])*(w->cd[3]);
    r = sqrt(xi*xi + xn*xn);
    if (r == 0.0) {
        dc = 1.0;
    } else {
        if (! w->istan) {
            rprime = r*w->pv21 + pow(r,(double)3.0)*w->pv23 + 
                pow(r,(double)5.0)*w->pv25;
            drprime_bydr = w->pv21 + 3.0*pow(r,(double)2.0)*w->pv23 + 
                5.0*pow(r,(double)4.0)*w->pv25;
            tanr = tan(rprime);
            dc = tanr*drprime_bydr*(1.0 + tanr*tanr)/r;
        } else {
            tanr = tan(r);
            dc = (1.0 + tanr*tanr)*tanr/r;
        }
    }
    return(dc);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        outloc
    \par Purpose:
        Work out the position of an input pixel on the output grid
    \par Description:
        Use the input WCS of an image to work out where in the output
        grid a particular pixel will fall.
    \par Language:
        C
    \param win
        The WCS descriptor for the input image
    \param xin
        The input X position of the pixel
    \param yin
        The input Y position of the pixel
    \param wout
        The WCS descriptor for the output grid
    \param xout
        The output X position of the pixel
    \param yout
        The output Y position of the pixel
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void outloc(mywcs *win, double xin, double yin, mywcs *wout, 
                   double *xout, double *yout) {
    double xt,yt,xi,eta,r,rfac,aa,tandec,denom,rp;
    int i;

    /* Do the conversion. First to standard coordinates in the frame of
       the input image */

    xt = xin - win->crpix[0];
    yt = yin - win->crpix[1];
    xi = win->cd[0]*xt + win->cd[1]*yt;
    eta = win->cd[2]*xt + win->cd[3]*yt;
    if (! win->istan) {
        if (fabs(xt) < 1.0e-6 && fabs(yt) < 1.0e-6) {
            rfac = 1.0;
        } else {
            rp = sqrt(xi*xi + eta*eta);
            r = rp;
            for (i = 0; i < 3; i++) {
                rfac = win->pv21 + win->pv23*pow(r,2.0) + win->pv25*pow(r,4.0);
                r = rp/rfac;
            }
            rfac = tan(r)/rp;
        }
    } else {
        rfac = 1.0;
    }
    xi *= rfac;
    eta *= rfac;
    aa = atan(xi*win->secd/(1.0-eta*win->tand));
    if (xi != 0.0) 
        tandec = (eta+win->tand)*sin(aa)/(xi*win->secd);
    else
        tandec = (eta+win->tand)/(1.0 - eta*win->tand);

    /* Now form standard coordinates in the frame of the output image */

    aa += (win->crval[0] - wout->crval[0]);
    denom = wout->tand*tandec + cos(aa);
    xi = wout->secd*sin(aa)/denom;
    eta = (tandec - wout->tand*cos(aa))/denom;
    if (! wout->istan) {
        rp = sqrt(xi*xi + eta*eta);
        r = atan(rp);
        rfac = wout->pv21 + wout->pv23*pow(rp,2.0) + wout->pv25*pow(rp,4.0);
        rfac *= r/rp;
    } else {
        rfac = 1.0;
    }
    xi *= rfac;
    eta *= rfac; 
    denom = wout->cd[0]*wout->cd[3] - wout->cd[1]*wout->cd[2];
    *xout = (xi*wout->cd[3] - eta*wout->cd[1])/denom + wout->crpix[0];
    *yout = (eta*wout->cd[0] - xi*wout->cd[2])/denom + wout->crpix[1];
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        xytord
    \par Purpose:
        Work out the RA and Dec of a given pixel
    \par Description:
        Use a given WCS descriptor to do the transformation between
        x,y and RA,Dec
    \par Language:
        C
    \param win
        The WCS descriptor for the input image
    \param xin
        The input X position of the pixel
    \param yin
        The input Y position of the pixel
    \param ra
        The output RA
    \param dec
        The output Dec
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void xytord(mywcs *win, double xin, double yin, double *ra, 
                   double *dec) {
    double xt,yt,xi,eta,r,rfac,aa;

    xt = xin - win->crpix[0];
    yt = yin - win->crpix[1];
    xi = win->cd[0]*xt + win->cd[1]*yt;
    eta = win->cd[2]*xt + win->cd[3]*yt;
    r = sqrt(xi*xi + eta*eta);
    if (! win->istan) {
        rfac = win->pv21 + win->pv23*pow(r,2.0) + win->pv25*pow(r,4.0);
        r /= rfac;
        rfac = win->pv21 + win->pv23*pow(r,2.0) + win->pv25*pow(r,4.0);
        xi /= rfac;
        eta /= rfac;
    } else {
        if (r == 0.0) 
            rfac = 1.0;
        else
            rfac = tan(r)/r;
        xi *= rfac;
        eta *= rfac;
    }
    aa = atan(xi*win->secd/(1.0-eta*win->tand));
    *ra = aa + win->crval[0];
    if (xi != 0.0) {
        *dec = atan((eta + win->tand)*sin(aa)/(xi*win->secd));
    } else {
        *dec = atan((eta + win->tand)/(1.0 - eta*win->tand));
    }
    if (*ra > 2.0*CPL_MATH_PI)
        *ra -= 2.0*CPL_MATH_PI;
    else if (*ra < 0.0)
        *ra += 2.0*CPL_MATH_PI;
    *ra *= DEGRAD;
    *dec *= DEGRAD;
}

/* static void rdtoxy(mywcs *win, double ra, double dec, double *x, double *y) { */
/*     double xi,eta,rra,ddec,denom,rp,rfac; */

/*     rra = ra/DEGRAD - win->crval[0]; */
/*     ddec = dec/DEGRAD; */
/*     denom = win->tand*tan(ddec) + cos(rra); */
/*     xi = win->secd*sin(rra)/denom; */
/*     eta = (tan(ddec) - win->tand*cos(rra))/denom; */
/*     rp = sqrt(xi*xi + eta*eta); */
/*     if (win->istan) { */
/*      if (rp == 0)  */
/*          rfac = 1.0; */
/*      else  */
/*          rfac = atan(rp)/rp; */
/*     } else { */
/*      rfac = win->pv21 + win->pv23*pow(rp,2.0) + win->pv25*pow(rp,4.0); */
/*     } */
/*     xi *= rfac; */
/*     eta *= rfac; */
/*     denom = win->cd[0]*win->cd[3] - win->cd[1]*win->cd[2]; */
/*     *x = (xi*win->cd[3] - eta*win->cd[1])/denom + win->crpix[0]; */
/*     *y = (eta*win->cd[0] - xi*win->cd[2])/denom + win->crpix[1]; */
/* } */

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        dupwcs
    \par Purpose:
        Duplicate a WCS descriptor
    \par Description:
        Creates an exact copy of a given WCS descriptor
    \par Language:
        C
    \param in
        The input WCS descriptor
    \return
        The output copied WCS descriptor
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static mywcs *dupwcs(mywcs *in) {
    int i;
    mywcs *out;
    
    out = cpl_malloc(sizeof(mywcs));
    for (i = 0; i < 2; i++) {
        out->crval[i] = in->crval[i];
        out->crpix[i] = in->crpix[i];
    }
    for (i = 0; i < 4; i++) 
        out->cd[i] = in->cd[i];
    out->secd = in->secd;
    out->tand = in->tand;
    out->pv21 = in->pv21;
    out->pv23 = in->pv23;
    out->pv25 = in->pv25;
    out->nx = in->nx;
    out->ny = in->ny;
    out->istan = in->istan;
    return(out);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        dstrct_init
    \par Purpose:
        Initialise the global properties of the dstrct descriptors
    \par Description:
        The dstrct array is allocated and the global properties are
        initialised to something sensible. The array fileptrs is global
    \par Language:
        C
    \param nfiles
        The number of input files.
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void dstrct_init(int nfiles) {
    int i;
    dstrct *dd;

    fileptrs = cpl_malloc(nfiles*sizeof(dstrct));
    for (i = 0; i < nfiles; i++) {
        dd = fileptrs + i;
        dd->infile = NULL;
        dd->conf = NULL;
        dd->global_sky = 0.0;
        dd->global_noise = 0.0;
        dd->tpa = 0.0;
        dd->tpd = 0.0;
        dd->exptime = 0.0;
        dd->nhdus = 0;
        dd->wcs_extn = NULL;
        dd->nx_extn = NULL;
        dd->ny_extn = NULL;
        dd->magzpt_extn = NULL;
        dd->sky_extn = NULL;
        dd->cur_extn = 0;
        dd->scalefac = NULL;
        dd->skydiff = NULL;
    }
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        dstrct_extn_init
    \par Purpose:
        Initialise the extension properties of a dstrct descriptor
    \par Description:
        Initialise the extension properties of a single dstrct descriptor
        and give them sensible values if needed
    \par Language:
        C
    \param dd
        The input dstrct structure.
    \param nhdu
        The number of FITS extensions needed
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void dstrct_extn_init(dstrct *dd, int nhdu) {
    int i;
    dd->nhdus = nhdu;
    dd->wcs_extn = cpl_malloc(nhdu*sizeof(mywcs *));
    for (i = 0; i < nhdu; i++)
        dd->wcs_extn[i] = NULL;
    dd->nx_extn = cpl_malloc(nhdu*sizeof(int));
    dd->ny_extn = cpl_malloc(nhdu*sizeof(int));
    dd->magzpt_extn = cpl_malloc(nhdu*sizeof(float));
    dd->sky_extn = cpl_malloc(nhdu*sizeof(float));
    dd->scalefac = cpl_malloc(nhdu*sizeof(float));
    dd->skydiff = cpl_malloc(nhdu*sizeof(float));
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        dstrct_extn_fill_from_plist
    \par Purpose:
        Fill the extension level properties of a dstrct
    \par Description:
        The extension level properties for a given dstrct are
        filled in from the headers for a single extension.
    \par Language:
        C
    \param dd
        The input dstrct structure.
    \param extnum
        The extension number 
    \param expkey
        The name of the FITS keyword for exposure time
    \param plist
        The propertylist for the given image header
    \param msg
        An message passed back in case of an error
    \return 
        CASU_OK => all went well, CASU_FATAL => missing header
        items or WCS didn't parse
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static int dstrct_extn_fill_from_plist(dstrct *dd, int extnum, 
                                       const char *expkey,
                                       cpl_propertylist *plist, char *msg) {
    mywcs *w;
    int i;

    /* Fill in some global parameters when reading the first header */

    if (extnum == 1) {
        if (cpl_propertylist_has(plist,"CRVAL1")) {
            dd->tpa = cpl_propertylist_get_double(plist,"CRVAL1");
        } else {
            (void)sprintf(msg,"Missing CRVAL1 in extension %d",extnum);
            return(CASU_FATAL);
        }
        if (cpl_propertylist_has(plist,"CRVAL2")) {
            dd->tpd = cpl_propertylist_get_double(plist,"CRVAL2");
        } else {
            (void)sprintf(msg,"Missing CRVAL2 in extension %d",extnum);
            return(CASU_FATAL);
        }
        if (cpl_propertylist_has(plist,expkey)) {
            dd->exptime = cpl_propertylist_get_float(plist,expkey);
        } else {
            (void)sprintf(msg,"Missing %s in extension %d",expkey,extnum);
            return(CASU_FATAL);
        }
    }

    /* Now some extension specific stuff. Start with the data array size
       and mag zeropoint */

    if (cpl_propertylist_has(plist,"NAXIS1")) {
        dd->nx_extn[extnum-1] = cpl_propertylist_get_int(plist,"NAXIS1");
    } else {
        (void)sprintf(msg,"Missing NAXIS1 in extension %d",extnum);
        return(CASU_FATAL);
    }
    if (cpl_propertylist_has(plist,"NAXIS2")) {
        dd->ny_extn[extnum-1] = cpl_propertylist_get_int(plist,"NAXIS2");
    } else {
        (void)sprintf(msg,"Missing NAXIS2 in extension %d",extnum);
        return(CASU_FATAL);
    }
    if (cpl_propertylist_has(plist,"ESO QC MAGZPT")) {
        dd->magzpt_extn[extnum-1] = cpl_propertylist_get_float(plist,
                                                               "ESO QC MAGZPT");
    } else {
        (void)sprintf(msg,"Missing ESO QC MAGZPT in extension %d",extnum);
        return(CASU_FATAL);
    }

    /* Now for the wcs */

    w = cpl_malloc(sizeof(mywcs));
    if (strstr(cpl_propertylist_get_string(plist,"CTYPE1"),"TAN") == NULL) 
        w->istan = 0;
    else
        w->istan = 1;
    w->crval[0] = cpl_propertylist_get_double(plist,"CRVAL1");
    w->crval[1] = cpl_propertylist_get_double(plist,"CRVAL2");
    w->crpix[0] = cpl_propertylist_get_double(plist,"CRPIX1");
    w->crpix[1] = cpl_propertylist_get_double(plist,"CRPIX2");
    w->cd[0] = cpl_propertylist_get_double(plist,"CD1_1");
    w->cd[1] = cpl_propertylist_get_double(plist,"CD1_2");
    w->cd[2] = cpl_propertylist_get_double(plist,"CD2_1");
    w->cd[3] = cpl_propertylist_get_double(plist,"CD2_2");
    if (! w->istan) {
        if (cpl_propertylist_has(plist,"PV2_1")) 
            w->pv21 = cpl_propertylist_get_double(plist,"PV2_1");
        else
            w->pv21 = 1.0;
        if (cpl_propertylist_has(plist,"PV2_3")) 
            w->pv23 = cpl_propertylist_get_double(plist,"PV2_3");
        else
            w->pv23 = -0.3333333333;
        if (cpl_propertylist_has(plist,"PV2_5")) 
            w->pv25 = cpl_propertylist_get_double(plist,"PV2_5");
        else
            w->pv25 = 0.0;
    } else {
        w->pv21 = 1.0;
        w->pv23 = -0.33333333;
        w->pv25 = 0.0;
    }
    w->nx = dd->nx_extn[extnum-1];
    w->ny = dd->ny_extn[extnum-1];

    /* Change angles to radians */

    for (i = 0; i < 2; i++) 
        w->crval[i] /= DEGRAD;
    for (i = 0; i < 4; i++)
        w->cd[i] /= DEGRAD;

    /* Add a couple of convenience values */

    w->tand = tan(w->crval[1]);
    w->secd = 1.0/cos(w->crval[1]);

    /* Now the store away WCS... */

    dd->wcs_extn[extnum-1] = w;
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        dstrct_close
    \par Purpose:
        Close the dstrct array and free up all workspace
    \par Description:
        The array of dstrcts is closed and all the allocated memory
        is freed
    \par Language:
        C
    \param nfiles
        The number of input files
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/
        
static void dstrct_close(int nfiles) {
    int i,j;
    dstrct *dd;

    for (i = 0; i < nfiles; i++) {
        dd = fileptrs + i;
        if (dd->wcs_extn != NULL) {
            for (j = 0; j < dd->nhdus; j++) 
                freespace(dd->wcs_extn[j]);
            freespace(dd->wcs_extn);
        }
        freespace(dd->nx_extn);
        freespace(dd->ny_extn);
        freespace(dd->magzpt_extn);
        freespace(dd->sky_extn);
        freespace(dd->scalefac);
        freespace(dd->skydiff);
    }
    freespace(fileptrs);
}
            
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        normal
    \par Purpose:
        Normalise the output array and create the confidence map
    \par Description:
        The values in the output array are normalised according to
        the number of pixels that contribute to each output pixel.
    \par Language:
        C
    \param o
        The structure containing the output data and confidence map
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

static void normal(outstrct *o) {
    int npo,i;
    float *odata,*work,renorm,val,junk;
    int *ocdata;
    unsigned char *bpm;

    /* Get some convenience variables */

    npo = o->nxo*o->nyo;
    odata = cpl_image_get_data_float(o->outim);
    work = cpl_image_get_data_float(o->work);
    ocdata = cpl_image_get_data_int(o->outcim);
    bpm = cpl_calloc(npo,sizeof(unsigned char));

    /* Normalise the output image */

    for (i = 0; i < npo; i++)  {
        if (work[i] != 0.0) {
            odata[i] /= work[i];
        } else {
            bpm[i] = 1;
        }
    }
            
    /* Work out a mean background that you can use to fill in the blank spots */

    (void)casu_qmedsig(odata,bpm,(long)npo,3.0,3,-1000.0,65535.0,&val,&junk);
    for (i = 0; i < npo; i++)
        if (bpm[i])
            odata[i] = val;

    /* Now normalise the confidence map */

    (void)casu_qmedsig(work,bpm,(long)npo,3.0,3,-1000.0,65535.0,&val,&junk);
    renorm = 100.0/val;
    for (i = 0; i < npo; i++)
        ocdata[i] = max(0,min(1000,casu_nint(work[i]*renorm)));
    cpl_free(bpm);
}       


/*

$Log: not supported by cvs2svn $

*/
