/* $Id: vircam_utils.c,v 1.77 2013-10-15 16:54:41 jim Exp $
 *
 * This file is part of the VIRCAM Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2013-10-15 16:54:41 $
 * $Revision: 1.77 $
 * $Name: not supported by cvs2svn $
 */

/* Includes */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _XOPEN_SOURCE
#include <sys/time.h>
#include <time.h>
#include <libgen.h>
#include <string.h>
#include <unistd.h>

#include <cpl.h>
#include <math.h>

#include <casu_utils.h>
#include <casu_stats.h>
#include <casu_fits.h>
#include <catalogue/imcore.h>

#include "vircam_utils.h"
#include "vircam_pfits.h"


#define SZKEY 32
#define SZVAL 64
#define SZBUF 1024
#define INITALLOC 1024
#define NOMPIXSIZE 0.34

typedef struct {
    char *filter_name;
    float atm_extcoef;
    float mag_offset;
    char **coleq_columns;
    float *coleq_coefs;
    float gal_extcoef;
    float default_zp;
    float default_zp_err;
    int ncolumns_coleq;
} photstrct;

/* Define some columns for illumination correction tables */

#define NI_COLS 5
static const char *illcor_cols[NI_COLS] = {"xmin","xmax","ymin","ymax",
                                           "illcor"};

/* Static subroutine prototypes */

static float madfunc(int npts, float *xt, float *yt, float b);
static double pixsize (cpl_propertylist *plist);
static int vircam_phot_open(cpl_table *phottab, char *filt, photstrct *p);
static void vircam_phot_close(photstrct *p);
static int extract_columns(cpl_table *tab, photstrct *p);
static int extract_coleq(cpl_table *tab, photstrct *p);

/**
    \defgroup vircam_utils vircam_utils
    \ingroup supportroutines

    \brief
    These are utility routines of various types.

    \author
    Jim Lewis, CASU
*/

/**@{*/

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_get_license
    \par Purpose:
        Get the pipeline copyright and license
    \par Description:
        The function returns a pointer to the statically allocated license 
        string. This string should not be modified using the returned pointer.
    \par Language:
        C
    \return The copyright and license string
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern const char *vircam_get_license(void) {
    const char  *vircam_license = 
        "This file is part of the VIRCAM Instrument Pipeline\n"
        "Copyright (C) 2015 European Southern Observatory\n"
        "\n"
        "This program is free software; you can redistribute it and/or modify\n"
        "it under the terms of the GNU General Public License as published by\n"
        "the Free Software Foundation; either version 2 of the License, or\n"
        "(at your option) any later version.\n"
        "\n"
        "This program is distributed in the hope that it will be useful,\n"
        "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
        "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
        "GNU General Public License for more details.\n"
        "\n"
        "You should have received a copy of the GNU General Public License\n"
        "along with this program; if not, write to the Free Software\n"
        "Foundation, Inc., 59 Temple Place, Suite 330, Boston, \n"
        "MA  02111-1307  USA";
    return(vircam_license);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_exten_range
    \par Purpose:
        Define the range of image extensions to use
    \par Description:
        If the input extension specified is zero, then all the extensions
        in a vircam MEF should be done. If not, then just the extension
        specified should be done.
    \par Language:
        C
    \param inexten
        Input image extension from the recipe invocation
    \param fr
        An input frame used to test whether the requested frame(s)
        exist or not
    \param out1
        Lower part of extension range (-1 if there is an error)
    \param out2
        Upper part of extension range (-1 if there is an error)
    \return Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_exten_range(int inexten, const cpl_frame *fr, int *out1, 
                               int *out2) {
    int nvircam = 16,n,nmax;
    const char *fctid = "vircam_exten_range";

    /* Right, how many frames actually exist? */

    n = cpl_frame_get_nextensions(fr);

    /* If there are less than the cannonical number, then issue a warning
       message here */

    if (n < nvircam) {
        cpl_msg_warning(fctid,
                        "Only %" CPL_SIZE_FORMAT " extensions out of %" CPL_SIZE_FORMAT " are present",
                        (cpl_size)n,(cpl_size)nvircam);
        nmax = n;
    } else {
        nmax = nvircam;
    }

    /* If the extension number requested is zero, then do all the extensions
       that are available */
    
    if (inexten == 0) {
        *out1 = 1;
        *out2 = nmax;

    /* Otherwise, check that the requested extension is actually present */

    } else {

        if (inexten > nmax) {
            cpl_msg_error(fctid,
                          "Requested extension %" CPL_SIZE_FORMAT " is not present",
                          (cpl_size)inexten);
            *out1 = -1;
            *out2 = -1;
        } else {
            *out1 = inexten;
            *out2 = inexten;
        }
    }
    return;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_madfit
    \par Purpose:
        Fit a line to data by minimising the sum of MADs
    \par Description:
        Standard linear fit routine, with the modification that the 
        minimisation scheme is done on the sum of MADs.
    \par Language:
        C
    \param npts
        The number of data points
    \param xdata
        The X data
    \param ydata
        The Y data
    \param intercept 
        The fitted intercept
    \param slope
        The fitted slope 
    \return   
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_madfit(int npts, float *xdata, float *ydata, 
                          float *intercept, float *slope) {
    int j;
    float sx,sy,sxx,sxy,det,aa,bb,temp,chisq,sigb,b1,f1,b2,f2,f;

    /* Do a linear least squares for a first estimate */

    sx = 0.0;
    sy = 0.0;
    sxx = 0.0;
    sxy = 0.0;
    for (j = 0; j < npts; j++) {
        sx += xdata[j];
        sy += ydata[j];
        sxx += xdata[j]*xdata[j];
        sxy += xdata[j]*ydata[j];
    }
    det = (float)npts*sxx - sx*sx;
    if (det == 0.0) {
        *slope = 0.0;
        *intercept = 0.0;
        return;
    }
    aa = (sxx*sy - sx*sxy)/det;
    bb = ((float)npts*sxy - sx*sy)/det;
    chisq = 0.0;
    for (j = 0; j < npts; j++) {
        temp = ydata[j] - (aa + bb*xdata[j]);
        chisq += temp*temp;
    }
    sigb = sqrt(chisq/det);
    if (sigb == 0.0) {
        *slope = bb;
        *intercept = aa;
        return;
    }

    /* Now bracket the solution */

    b1 = bb;
    f1 = madfunc(npts,xdata,ydata,b1);
    b2 = bb + ((f1 > 0.0) ? fabs(3.0*sigb) : -fabs(3.0*sigb));
    f2 = madfunc(npts,xdata,ydata,b2);
    while (f1*f2 > 0.0) {
        bb = 2.0*b2 - b1;
        b1 = b2;
        f1 = f2;
        b2 = bb;
        f2 = madfunc(npts,xdata,ydata,b2);
    }

    /* Iterate to find the minimum value */

    sigb = 0.01*sigb;
    while (fabs(b2 - b1) > sigb) {
        bb = 0.5*(b1 + b2);
        if (bb == b1 || bb == b2) 
            break;
        f = madfunc(npts,xdata,ydata,bb);
        if (f*f1 >= 0.0) {
            f1 = f;
            b1 = bb;
        } else {
            f2 = f;
            b2 = bb;
        }
    }
    *intercept = aa;
    *slope = bb;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        madfunc
    \par Purpose:
        Service routine for vircam_madfit
    \par Description:
        Service routine for vircam_madfit.
    \par Language:
        C
    \param npts
        The number of data points
    \param xt
        The X data
    \param yt
        The Y data
    \param b
        The fitted slope 
    \return  
        The sum of the MADs
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/


static float madfunc(int npts, float *xt, float *yt, float b) {
    float *arr,aa,d,sum;
    int j;

    arr = cpl_malloc(npts*sizeof(*arr));
    for (j = 0; j < npts; j++) 
        arr[j] = yt[j] - b*xt[j];
    aa = casu_med(arr,NULL,(long)npts);
    sum = 0.0;
    for (j = 0; j < npts; j++) {
        d = yt[j] - (b*xt[j] + aa);
        sum += d > 0.0 ? xt[j] : -xt[j];
    }
    cpl_free(arr);
    return(sum);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_linfit
    \par Purpose:
        Fit a line to data using standard linear least squares
    \par Description:
        Standard linear fit routine, 
    \par Language:
        C
    \param npts
        The number of data points
    \param xdata
        The X data
    \param ydata
        The Y data
    \param intercept 
        The fitted intercept
    \param slope
        The fitted slope 
    \param sig
        The fit error
    \return   
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_linfit(int npts, double *xdata, double *ydata, 
                          double *intercept, double *slope, double *sig) {
    int j;
    double sx,sy,sxx,sxy,det,aa,bb,temp,sum,sumsq;

    /* Do a linear least squares */

    sx = 0.0;
    sy = 0.0;
    sxx = 0.0;
    sxy = 0.0;
    for (j = 0; j < npts; j++) {
        sx += xdata[j];
        sy += ydata[j];
        sxx += xdata[j]*xdata[j];
        sxy += xdata[j]*ydata[j];
    }
    det = (double)npts*sxx - sx*sx;
    if (det == 0.0) {
        *slope = 0.0;
        *intercept = 0.0;
        *sig = 0.0;
        return;
    }

    /* Intercept and slope */

    aa = (sxx*sy - sx*sxy)/det;
    bb = ((double)npts*sxy - sx*sy)/det;

    /* Work out RMS of fit */

    sum = 0.0;
    sumsq = 0.0;
    for (j = 0; j < npts; j++) {
        temp = ydata[j] - (aa + bb*xdata[j]);
        sum += temp;
        sumsq += temp*temp;
    }
    sum /= (double)npts;

    /* Set return values */

    *sig = sqrt(sumsq/(double)npts - sum*sum);
    *slope = bb;
    *intercept = aa;
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_solve_gauss
    \par Purpose:
        Use Gauss-elimination method to solve for fit coefficients
    \par Description:
        This is a standard Gauss-elimination method routine.
    \par Language:
        C
    \param a
        The matrix part of the system (a[m][m])
    \param b
        The vector part of the system (b[m]). The output fit coefficients
        overwrite the input values in this vector.
    \param m
        The size of the vectors
    \retval CASU_OK
        If everything went OK
    \retval CASU_FATAL
        If matrix a has a zero determinant
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_solve_gauss(double **a, double *b, int m) {
    double temp,big,pivot,rmax;
    int i,iu,j,k,jl,ib,ir;
    int l = 0;

    iu = m - 1;
    for (i = 0; i < iu; i++) {
        big = 0.0;

        /* find largest remaining term in ith column for pivot */

        for (k = i; k < m; k++) {
            rmax = fabs(a[i][k]);
            if (rmax > big) {
                big = rmax;
                l = k;
            }
        }

        /* check for non-zero term */

        if (big == 0.0) {
            for (ib = 0; ib < m; ib++)
                b[ib] = 0.0;
            cpl_msg_error("vircam_solve_gauss","Zero Determinant\n");
            return(CASU_FATAL);
        }

        if (i != l) {

            /* switch rows */

            for (j = 0; j < m; j++) {
                temp = a[j][i];
                a[j][i] = a[j][l];
                a[j][l] = temp;
            }
            temp = b[i];
            b[i] = b[l];
            b[l] = temp;
        }


        /* pivotal reduction */

        pivot = a[i][i];
        jl = i+1;

        for (j = jl; j < m; j++) {
            temp = a[i][j]/pivot;
            b[j] -= temp*b[i];
            for (k = i; k < m; k++)
                a[k][j] -= temp*a[k][i];
        }
    }

    /* back substitution for solution */

    for (i = 0; i < m; i++) {
        ir = m - 1 - i;
        if (a[ir][ir] != 0.0) {
            temp = b[ir];
            if (ir != m - 1) {
                for (j = 1; j <= i; j++) {
                    k = m - j;
                    temp -= a[k][ir]*b[k];
                }
            }
            b[ir] = temp/a[ir][ir];
        } else
            b[ir] = 0.0;
    }
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_polyfit
    \par Purpose:
        Fit a polynomial to a set of points
    \par Description:
        Fit a polynomial to a set of points. If requested the fit will be done 
        iteratively. If requested the lower order coefficients will be 
        defined as zero. This is done by adjusting the parameter ilim. 
        The number of coefficients returned will be ncoefs - ilim.
    \par Language:
        C
    \param xarray
        The independent variable array
    \param yarray
        The dependent variable array
    \param ncoefs
        The requested polynomial order plus 1.
    \param ilim
        The number of low order coefficients not to fit. 
    \param niter
        The number of iterations
    \param lclip
        The number of sigma for the lower clipping threshold
    \param hclip
        The number of sigma for the upper clipping threshold
    \param polycf
        The returned array of polynomial coefficients.
    \param sigfit
        The total error of the fit
    \retval CASU_OK
        If everything went OK
    \retval CASU_FATAL
        If there was some sort of error
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_polyfit(const cpl_array *xarray, const cpl_array *yarray, 
                          int ncoefs, int ilim, int niter, float lclip, 
                          float hclip, cpl_array **polycf, double *sigfit) {
    const char *fctid = "vircam_polyfit";
    int npts,iter,i,j,nnew,k,retval,n;
    double *xdata,*ydata,*pdata,*res,**a,*b,temp,sum,sumsq,val;
    double lcut,hcut;
    unsigned char *pm;

    /* Initialise a few things */

    *polycf = NULL;
    *sigfit = -1.0;

    /* How many data points do we have? */

    npts = (int)cpl_array_get_size(xarray);

    /* Do we have enough points for the required order of the fit? */

    if (npts < ncoefs) {
        cpl_msg_warning(fctid,
                        "Not data for fit, Npts = %" CPL_SIZE_FORMAT ", Ncoefs = %" CPL_SIZE_FORMAT,
                        (cpl_size)npts,(cpl_size)ncoefs);
        return(CASU_FATAL);
    }

    /* Create some arrays */

    a = cpl_malloc(ncoefs*sizeof(double *));
    b = cpl_calloc(ncoefs,sizeof(double));
    for (i = 0; i < ncoefs; i++) 
        a[i] = cpl_calloc(ncoefs,sizeof(double));
    pm = cpl_calloc(npts,sizeof(unsigned char));
    res = cpl_malloc(npts*sizeof(double));

    /* Get pointers to the input arrays */

    xdata = (double *)cpl_array_get_data_double_const(xarray);
    ydata = (double *)cpl_array_get_data_double_const(yarray);

    /* Get an output array */

    *polycf = cpl_array_new((cpl_size)ncoefs,CPL_TYPE_DOUBLE);
    pdata = cpl_array_get_data_double(*polycf);

    /* Iteration loop */

    for (iter = 0; iter <= niter; iter++) {

        /* Zero some accumulators */

        for (i = 0; i < ncoefs; i++)  {
            for (j = 0; j < ncoefs; j++) 
                a[i][j] = 0.0;
            b[i] = 0.0;
        }
        nnew = 0;

        /* Cumulate sums */
 
        for (i = 0; i < npts; i++) {
            if (pm[i] == 1)
                continue;
            for (k = 0; k < ncoefs; k++) {
                temp = 1.0;
                if (k + ilim != 0)
                    temp = pow(xdata[i],(double)(k+ilim));
                b[k] += ydata[i]*temp;
                for (j = 0; j <= k; j++) {
                    temp = 1.0;
                    if (k + j + 2*ilim != 0)
                        temp = pow(xdata[i],(double)(k+j+2*ilim));
                    a[j][k] += temp;
                }
            }
        }
        for (k = 1; k < ncoefs; k++) 
            for (j = 0; j < k; j++) 
                a[k][j] = a[j][k];

        /* Solve linear equations */

        retval = vircam_solve_gauss(a,b,ncoefs);
        if (retval != CASU_OK) {
            cpl_msg_warning(fctid,"Fit failed");
            freearray(*polycf);
            freespace2(a,ncoefs);
            freespace(b);
            freespace(pm);
            freespace(res);
            return(CASU_FATAL);
        }

        /* Ok, assuming this is OK, then fill the polynomial coefficients */

        for (i = 0; i < ncoefs; i++)
            pdata[i] = b[i];

        /* Calculate the fit quality */

        sum = 0.0;
        sumsq = 0.0;
        n = 0;
        for (i = 0; i < npts; i++) {
            if (pm[i] == 1)
                continue;
            val = 0.0;
            for (j = 0; j < ncoefs; j++)                
                val += pdata[j]*pow(xdata[i],(double)j+ilim);
            res[i] = val - ydata[i];
            sum += res[i];
            sumsq += pow(res[i],2.0);
            n++;
        }
        sum /= (double)n;
        *sigfit = sqrt(sumsq/(double)n - sum*sum);

        /* If this is not the last iteration, then do some clipping */

        lcut = sum - lclip*(*sigfit);
        hcut = sum + hclip*(*sigfit);
        if (iter < niter) {
            for (i = 0; i < npts; i++) {
                if (pm[i] == 1) 
                    continue;
                if (res[i] > hcut || res[i] < lcut) {
                    nnew++;
                    pm[i] = 1;
                }
            }
        }
        
        /* If no new points have been clipped, then get out of here now... */

        if (nnew == 0) 
            break;
    }

    /* Tidy up and get out of here */

    freespace2(a,ncoefs);
    freespace(b);
    freespace(pm);
    freespace(res);
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_difference_image
    \par Purpose:
        Create a difference/ratio image and difference/ratio image stats table
    \par Description:
        A difference/ratio image is created from an input image and a master 
        image. A global difference and RMS are calculated from the difference 
        image. If a channel table has been included, then a difference image 
        stats table will be created. This breaks up the difference image into 
        cells and calculates some basic stats in each cell.
    \par Language:
        C
    \param master
        The master calibration image
    \param prog
        The new mean image
    \param bpm 
        Input bad pixel mask
    \param chantab
        The channel table for the detector
    \param ncells
        The number of cells per channel
    \param oper
        The operation to be performed:
        - 1. Subtract the images
        - 2. Divide the images
    \param global_diff 
        The median difference over the whole difference image
    \param global_rms  
        The rms difference over the whole difference image
    \param diffim
        The output difference/ratio image
    \param diffimstats 
        The output difference/ratio image statistics table
    \return   
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_difference_image(cpl_image *master, cpl_image *prog, 
                                    unsigned char *bpm, cpl_table *chantab, 
                                    int ncells, int oper, float *global_diff, 
                                    float *global_rms, cpl_image **diffim, 
                                    cpl_table **diffimstats) {
    float *ddata,*work,mean,sig,med,mad;
    long nx,ny,npts;
    int nrows,i,nc1,nc2,nr,ixmin,ixmax,iymin,iymax,cnum,cx,cy,idx,idy;
    int icx,icy,indy1,indy2,indx1,indx2,jp,jcx,jj,jcy,ii,ncx,ncy;
    const char *fctid = "vircam_difference_image";

    /* Initialise the output */

    *diffim = NULL;
    *diffimstats = NULL;
    *global_diff = 0.0;
    *global_rms = 0.0;

    /* Is there a programme frame or a master? */

    if (prog == NULL || master == NULL)
        return;

    /* Start by subtracting the master image from the programme image */

    switch (oper) {
    case 1:
        *diffim = cpl_image_subtract_create(prog,master);
        break;
    case 2:
        *diffim = cpl_image_divide_create(prog,master);
        break;
    default:
        *diffim = NULL;
        cpl_msg_error(fctid,"Invalid operation requested %" CPL_SIZE_FORMAT,
                      (cpl_size)oper);
        break;
    }      
    if (*diffim == NULL)
        return;

    /* Work out a median difference */

    ddata = cpl_image_get_data_float(*diffim);
    nx = (int)cpl_image_get_size_x(*diffim);
    ny = (int)cpl_image_get_size_y(*diffim);
    npts = nx*ny;
    casu_medmad(ddata,bpm,npts,global_diff,global_rms);
    *global_rms *= 1.48;

    /* Is there a channel table? */

    if (chantab == NULL)
        return;

    /* Before going any further, check that the channel table has all of 
       the columns we need */

    if (! cpl_table_has_column(chantab,"ixmin") ||
        ! cpl_table_has_column(chantab,"ixmax") ||
        ! cpl_table_has_column(chantab,"iymin") ||
        ! cpl_table_has_column(chantab,"iymax") ||
        ! cpl_table_has_column(chantab,"channum")) {
            cpl_msg_error(fctid,"Channel table is missing one of the required columns");

            return;
    }

    /* Work out how to divide the channels */

    switch (ncells) {
    case 1:
        nc1 = 1;
        nc2 = 1;
        break;
    case 2:
        nc1 = 2;
        nc2 = 1;
        break;
    case 4:
        nc1 = 4;
        nc2 = 1;
        break;
    case 8:
        nc1 = 8;
        nc2 = 1;
        break;
    case 16:
        nc1 = 16;
        nc2 = 1;
        break;
    case 32:
        nc1 = 16;
        nc2 = 2;
        break;
    case 64:
        nc1 = 32;
        nc2 = 2;
        break;
    default:
        nc1 = 32;
        nc2 = 2;
        break;
    }

    /* Create a difference image stats table */

    nrows = (int)cpl_table_count_selected(chantab);
    *diffimstats = vircam_create_diffimg_stats(nrows*nc1*nc2);

    /* Loop for each data channel now */

    nr = 0;
    for (i = 0; i < nrows; i++) {
        ixmin = cpl_table_get_int(chantab,"ixmin",(cpl_size)i,NULL);
        ixmax = cpl_table_get_int(chantab,"ixmax",(cpl_size)i,NULL);
        iymin = cpl_table_get_int(chantab,"iymin",(cpl_size)i,NULL);
        iymax = cpl_table_get_int(chantab,"iymax",(cpl_size)i,NULL);
        cnum = cpl_table_get_int(chantab,"channum",(cpl_size)i,NULL);

        /* Which is the long axis? If the channels are rectangular then
           divide the long axis by the greater number of cells. If the number
           of cells is less than 8, then just divide the long axis. If the
           channel is square, then divide the X axis more finely */

        cx = ixmax - ixmin + 1;
        cy = iymax - iymin + 1;
        if (cx > cy) {
            ncx = max(nc1,nc2);
            ncy = min(nc1,nc2);
        } else if (cx < cy) {
            ncy = max(nc1,nc2);
            ncx = min(nc1,nc2);
        } else {
            ncx = max(nc1,nc2);
            ncy = min(nc1,nc2);
        }
        
        /* How big is the cell? */

        idy = cy/ncy;
        idx = cx/ncx;
        work = cpl_malloc(idx*idy*sizeof(*work));

        /* Now loop for aach cell */

        for (icy = 0; icy < ncy; icy++) {
            indy1 = idy*icy;
            indy2 = min(iymax,indy1+idy-1);
            for (icx = 0; icx < ncx; icx++) {
                indx1 = idx*icx;
                indx2 = min(ixmax,indx1+idx-1);
                jp = 0;
                for (jcy = indy1; jcy < indy2; jcy++) {
                    jj = jcy*nx;
                    for (jcx = indx1; jcx < indx2; jcx++) {
                        ii = jj + jcx;
                        if (bpm != NULL && bpm[ii] == 0)
                            work[jp++] = ddata[ii];
                    }
                }
                (void)casu_meansig(work,NULL,(long)jp,&mean,&sig);
                (void)casu_medmad(work,NULL,(long)jp,&med,&mad);
                cpl_table_set_int(*diffimstats,"xmin",(cpl_size)nr,indx1+1);
                cpl_table_set_int(*diffimstats,"xmax",(cpl_size)nr,indx2+1);
                cpl_table_set_int(*diffimstats,"ymin",(cpl_size)nr,indy1+1);
                cpl_table_set_int(*diffimstats,"ymax",(cpl_size)nr,indy2+1);
                cpl_table_set_int(*diffimstats,"chan",(cpl_size)nr,cnum);
                cpl_table_set_float(*diffimstats,"mean",(cpl_size)nr,mean);
                cpl_table_set_float(*diffimstats,"median",(cpl_size)nr,med);
                cpl_table_set_float(*diffimstats,"variance",(cpl_size)nr,
                                    (sig*sig));
                cpl_table_set_float(*diffimstats,"mad",(cpl_size)(nr++),mad);
            }
        }
        cpl_free(work);
    }
}               

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_create_diffimg_stats
    \par Purpose:
        Create an empty difference image stats table
    \par Description:
        Create an empty difference image stats table
    \par Language:
        C
    \param nrows
        The number of rows for the table
    \returns
        The cpl_table pointer for the new stats table
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern cpl_table *vircam_create_diffimg_stats(int nrows) {
    cpl_table *diffimstats;

    diffimstats = cpl_table_new((cpl_size)nrows);
    cpl_table_new_column(diffimstats,"xmin",CPL_TYPE_INT);
    cpl_table_set_column_unit(diffimstats,"xmin","pixels");
    cpl_table_new_column(diffimstats,"xmax",CPL_TYPE_INT);
    cpl_table_set_column_unit(diffimstats,"xmax","pixels");
    cpl_table_new_column(diffimstats,"ymin",CPL_TYPE_INT);
    cpl_table_set_column_unit(diffimstats,"ymin","pixels");
    cpl_table_new_column(diffimstats,"ymax",CPL_TYPE_INT);
    cpl_table_set_column_unit(diffimstats,"ymax","pixels");
    cpl_table_new_column(diffimstats,"chan",CPL_TYPE_INT);
    cpl_table_set_column_unit(diffimstats,"chan","pixels");
    cpl_table_new_column(diffimstats,"mean",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(diffimstats,"mean","ADU");
    cpl_table_new_column(diffimstats,"median",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(diffimstats,"median","ADU");
    cpl_table_new_column(diffimstats,"variance",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(diffimstats,"variance","ADU**2");
    cpl_table_new_column(diffimstats,"mad",CPL_TYPE_FLOAT);
    cpl_table_set_column_unit(diffimstats,"mad","ADU");
    return(diffimstats);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_dummy_confidence
    \par Purpose:
        Create a dummy confidence map with all pixels equally good
    \par Description:
        An integer array is returned with all values equal to 100. This
        array will need to be deallocated when you're finished with it.
    \par Language:
        C
    \param n
        The number of pixels
    \return
        The output dummy confidence map. This array will need to be deallocated
        when you're finished with it.
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

extern int *vircam_dummy_confidence(long n) {
    int *cdata,i;
 
    cdata = cpl_malloc(n*sizeof(*cdata));
    for (i = 0; i < n; i++)
        cdata[i] = 100;
    return(cdata);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_is_dummy
    \par Purpose:
        See if an image or table is a dummy output product
    \par Description:
        The input propertylist is searched for the key ESO DRS IMADUMMY.
        If it exists, then this header is part of a dummy image or table.
    \par Language:
        C
    \param p
        The input propertylist
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

extern int vircam_is_dummy(cpl_propertylist *p) {

    /* Check for silly input */

    if (p == NULL)
        return(0);

    /* Check the propertylist and return the result */

    return(cpl_propertylist_has(p,"ESO DRS IMADUMMY"));
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_illcor_newtab
    \par Purpose:
        Create a new illumination correction table
    \par Description:
        Given the number of rows, create an illumination correction table
        with no information in it.
    \par Language:
        C
    \param nrows
        The number of rows in the table
    \return The pointer to the new table
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

extern cpl_table *vircam_illcor_newtab(int nrows) {
    cpl_table *illcor;
    int i;

    /* Create the table structure and fill in the columns */

    illcor = cpl_table_new((cpl_size)nrows);
    for (i = 0; i < NI_COLS; i++) 
        cpl_table_new_column(illcor,illcor_cols[i],CPL_TYPE_FLOAT);
    return(illcor);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_timestamp
    \par Purpose:
        Create a timestamp string
    \par Description:
        Create a timestamp string similar to DATE-OBS. This can be used
        to associate files together.
    \par Language:
        C
    \param out
        The output string. It should be at least 25 characters long. NB:
        if it isn't at least 25 characters long (and provided you haven't
        done something stupid like pass in a value of n which is larger
        than the size of the string) then the output value will truncated
        to n characters
    \param n
        The maximum size of the string.
    \returns
        Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

extern void vircam_timestamp(char *out, int n) {
    struct timeval tv;
    struct tm *tm;
    float sec;
    struct tm result;
    /* Get the Greenwich Mean Time */

    (void)gettimeofday(&tv,NULL);

    tm = gmtime_r(&(tv.tv_sec), &result);
    sec = (float)tm->tm_sec + 1.0e-6*(float)tv.tv_usec;

    /* Now format it */

    (void)snprintf(out,n,"%04d-%02d-%02dT%02d:%02d:%07.4f",1900+tm->tm_year,
                   tm->tm_mon+1,tm->tm_mday,tm->tm_hour,tm->tm_min,sec);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_check_crval
    \par Purpose:
        Check CRVAL == 0 problem
    \par Description:
        The extension header is checked to see if CRVAL values exist. If they
        don't that's an error. If they do and they are exactly zero, then this
        is considered an error. In this case the primary header is searched
        for RA and DEC values and these are silently substituted into the
        extension header as CRVALs
    \par Language:
        C
    \param phu
        The primary header propertylist
    \param ehu
        The extension header propertylist
    \retval CASU_OK 
        if everything is ok
    \retval CASU_FATAL
        if no values of CRVAL are in the extension or no values of RA,DEC
        are available in the primary
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_check_crval(cpl_propertylist *phu, cpl_propertylist *ehu) {
    double crval1,crval2;
    cpl_property *p;

    /* Get the values of CRVAL1,2 from the extension header. If it doesn't
       exist, then signal an error */

    if ((vircam_pfits_get_crval1(ehu,&crval1) != CASU_OK) ||
        (vircam_pfits_get_crval2(ehu,&crval2) != CASU_OK))
        return(CASU_FATAL);

    /* If the values are both zero, then grab the values from the primary
       and silently update the extension header. NB: have to do this silly
       song and dance where we create new properties, rename the old ones,
       insert the new ones in the correct place and then erase the old 
       ones because CPL won't do the update if it thinks the original 
       CRVALs are floats rather than a double (which it inevitably will, 
       since this is where CRVAL has been set identically to zero...) */

    if (fabs(crval1) < 1.0e-6 && fabs(crval2) < 1.0e-6) {
        if ((vircam_pfits_get_ra(phu,&crval1) != CASU_OK) ||
            (vircam_pfits_get_dec(phu,&crval2) != CASU_OK))
            return(CASU_FATAL);
        p = cpl_propertylist_get_property(ehu,"CRVAL1");
        cpl_property_set_name(p,"OLDCR1");
        p = cpl_propertylist_get_property(ehu,"CRVAL2");
        cpl_property_set_name(p,"OLDCR2");
        cpl_propertylist_insert_after_double(ehu,"OLDCR2","CRVAL1",crval1);
        cpl_propertylist_insert_after_double(ehu,"CRVAL1","CRVAL2",crval2);
        cpl_propertylist_erase(ehu,"OLDCR1");
        cpl_propertylist_erase(ehu,"OLDCR2");
    }
    
    /* Get out of here */

    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_copywcs
    \par Purpose:
        Copy a WCS from one propertylist to another
    \par Description:
        Copy a full WCS from the first propertylist into the second one
    \par Language:
        C
    \param p1
        The first input propertylist. This is the source of the WCS
    \param p2
        The second input propertylist. The WCS will be copied to this
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_copywcs(cpl_propertylist *p1, cpl_propertylist *p2) {
    (void)cpl_propertylist_erase_regexp(p2,CPL_WCS_REGEXP,0);
    cpl_propertylist_copy_property_regexp(p2,p1,CPL_WCS_REGEXP,0);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_cull
    \par Purpose:
        Work out which images in a flat field frameset to remove
    \par Description:
        Get stats on images in a frameset. Work out which ones to remove.
        Send out status flags for dodgy results.
    \par Language:
        C
    \param in
        The input frameset
    \param lthr
        The lower threshold
    \param hthr
        The upper threshold
    \param ndit
        The number of DITs in the exposure
    \param jst
        The first extension to be tested
    \param jfn
        The last extension to be tested
    \param mins
        An array of minimum flux values for each extension
    \param maxs
        An array of maximum flux values for each extension
    \param aves
        An array of average flux values for each extension
    \param out
        The output frameset with frames to be kept
    \param status
        An output status. 0 -> all ok, 1 -> only one extension
        is being used and that particular detector is dead, 
        2-> all of the images in the nominated detector are 
        outside of the flux thresholds. The nominated detector
        is either (1) the first live detector in the case where
        we user looking at multiple detectors, or (2) the only
        detector in the case where we're only looking at 1.
    \return
        Nothing
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern void vircam_cull(cpl_frameset *in, float lthr, float hthr,
                        int ndit, int jst, int jfn, cpl_array **mins,
                        cpl_array **maxs, cpl_array **aves, 
                        cpl_frameset **out, int *status) {
    int i,n,nj,j,live,gotit;
    double val,dndit,fmin,fmax,sum;
    cpl_propertylist *p;
    cpl_image *im;
    cpl_frame *fr;

    /* Initialise a few things */

    nj = jfn - jst + 1;
    *status = 0;
    *out = NULL;
    gotit = 0;
    dndit = (double)ndit;
    *mins = cpl_array_new(nj,CPL_TYPE_DOUBLE);
    *maxs = cpl_array_new(nj,CPL_TYPE_DOUBLE);
    *aves = cpl_array_new(nj,CPL_TYPE_DOUBLE);

    /* Loop through each extension to be included */

    n = cpl_frameset_get_size(in);
    for (j = jst; j <= jfn; j++) {

        /* Look at the current extension in the first frame of the
           frameset. If it's flagged as dead, then assume the detector
           is dead for all of them. If this is the only extension we
           look at, then status = 1 is passed back to say all 
           images are ignored because of a dead detector */

        fr = cpl_frameset_get_position(in,0);
        p = cpl_propertylist_load(cpl_frame_get_filename(fr),j);
        vircam_pfits_get_detlive(p,&live);
        cpl_propertylist_delete(p);
        if (! live) {
            cpl_array_set(*mins,j-jst,0.0);
            cpl_array_set(*maxs,j-jst,0.0);
            cpl_array_set(*aves,j-jst,0.0);
            if (nj == 1)
                *status = 1;
            continue;
        }

        /* OK, we have a live detector. If this is the first good detector
           then get an array we sort the output frameset here */

        if (! gotit) {
            *out = cpl_frameset_new();
            gotit = 1;
        }

        /* Loop through all of the images for the current detector and 
           get an estimate of the median level. Keep tabs on the min 
           and max. We normalise to ndit = 1 */

        fmin = 1000000000.0;
        fmax = -1000000000.0;
        sum = 0.0;
        for (i = 0; i < n; i++) {
            fr = cpl_frameset_get_position(in,i);
            im = cpl_image_load(cpl_frame_get_filename(fr),CPL_TYPE_FLOAT,0,j);
            val = cpl_image_get_median(im);
            val /= dndit;
            cpl_image_delete(im);
            fmin = min(fmin,val);
            fmax = max(fmax,val);
            sum += val;
            if (gotit == 1) {
                if (val >= lthr && val <= hthr)
                    cpl_frameset_insert(*out,cpl_frame_duplicate(fr));
            }
        }
        cpl_array_set(*mins,j-jst,fmin);
        cpl_array_set(*maxs,j-jst,fmax);
        cpl_array_set(*aves,j-jst,sum/(double)n);

        /* Ok we've looked at the thresholds now. Flag so that we don't
           do it again. Check the number of frames that made it
           through the threshold and set the status to say if all of
           them failed */

        if (gotit == 1) {
            gotit = 2;
            if (cpl_frameset_get_size(*out) == 0) 
                *status = 2;
        }
    }

    /* Right, get out of here */
  
    return;
}

/*---------------------------------------------------------------------------*/
/**
    \ingroup reductionmodules
    \brief Work out the illumination correction

    \par Name:
        vircam_illum
    \par Purpose:
        Work out the illumination correction
    \par Description:
        Work out the illumination correction from a set of images by 
        dividing the image space into a number of cells. The median of 
        photometric zeropoints for all the stars in a cell is compared with
        the median zeropoint over the whole set of images.
    \par Language:
        C
    \param images
        The input data images
    \param mstds
        The input matched standards catalogues (one for each image)
    \param pl
        The list of extension header propertylists (one for each image 
        catalogue)
    \param nimages
        The number of images in the input lists
    \param filt
        The filter for the observations
    \param phottab
        The photometric calibration table for VIRCAM
    \param nbsize
        The size of the side of a cell in pixels 
    \param illcor
        The illumination correction table for this detector 
    \param illcor_rms
        The RMS of the illumination correction
    \param status
        An input/output status that is the same as the returned values below.
    \retval CASU_OK
        if everything is ok
    \retval CASU_WARN
        if there weren't any photometric standards or all the objects in
        the catalogue failed.
    \retval CASU_FATAL
        if there is a problem with the photometric calibration table
    \par QC headers:
        None
    \par DRS headers:
        None
    \author
        Jim Lewis, CASU
 */
/*---------------------------------------------------------------------------*/

extern int vircam_illum(casu_fits **images, cpl_table **mstds, 
                        cpl_propertylist **pl, int nimages, char *filt,
                        cpl_table *phottab, int nbsize, cpl_table **illcor, 
                        float *illcor_rms,int *status) {
    const char *fctid = "vircam_illum";
    float **stdmagptr,fracx,fracy,**results,cdfudge,saturate,apcor3,exptime;
    float airmass,*catcore3,*xx,*yy,extinct,cf,fluxmag3,refmag,dm3,med,mad;
    float xmin,xmax,ymin,ymax,*resall,medall,lcut,hcut,sig,cut,*good;
    int nx,ny,ifracx,ifracy,nbsizx,nbsizy,nbx,nby,*nres,*nall,i,j,ncat,k;
    int ix,iy,ind,nresall,nallocall,ngood;
    casu_fits *im;
    cpl_propertylist *ehu_im,*ehu_cat;
    cpl_table *stds,*cl;
    photstrct p;

    /* Inherited status */

    *illcor = NULL;
    *illcor_rms = 0.0;
    if (*status != CASU_OK)
        return(*status);

    /* Check for nonsense errors */

    if (nimages <= 0) {
        cpl_msg_error(fctid,"No images included in photometric calibration");
        FATAL_ERROR
    }

    /* Set up the structure that will give us the colour equations for
       this filter later on */

    if (vircam_phot_open(phottab,filt,&p) != CASU_OK)
        FATAL_ERROR

    /* Get a workspace to hold the pointers to the magnitude columns */

    stdmagptr = cpl_malloc(p.ncolumns_coleq*sizeof(float *));

    /* Check to see if nbsize is close to exact divisor */

    nx = (int)cpl_image_get_size_x(casu_fits_get_image(images[0]));
    ny = (int)cpl_image_get_size_y(casu_fits_get_image(images[0]));
    fracx = ((float)nx)/((float)nbsize);
    fracy = ((float)ny)/((float)nbsize);
    ifracx = (int)(fracx + 0.1);
    ifracy = (int)(fracy + 0.1);
    nbsizx = nx/ifracx;
    nbsizy = ny/ifracy;
    nbsize = max(casu_nint(0.9*nbsize),min(nbsize,min(nbsizx,nbsizy)));
    nbsize = min(nx,min(ny,nbsize)); /* trap for small maps */

    /* Divide the map into partitions */

    nbx = nx/nbsize;
    nby = ny/nbsize;

    /* Get some workspace to hold all the zeropoints for all the images
       in the input list. This is an initial allocation and more can be
       made available later if needed */

    results = cpl_malloc(nbx*nby*sizeof(float *));
    good = cpl_malloc(nbx*nby*sizeof(float));
    nres = cpl_calloc(nbx*nby,sizeof(int));
    nall = cpl_malloc(nbx*nby*sizeof(int));
    for (i = 0; i < nbx*nby; i++) {
        results[i] = cpl_malloc(INITALLOC*sizeof(float));
        nall[i] = INITALLOC;
    }
    resall = cpl_malloc(INITALLOC*sizeof(float));
    nresall = 0;
    nallocall = INITALLOC;

    /* Set up the output illumination correction table */

    *illcor = vircam_illcor_newtab(nbx*nby);
    
    /* Loop for the input images and catalogues. Create some shorthand
       variables and work out what the zeropoint fudge factor will be 
       based on the sampling of the current frame and the nominal sampling
       of a VIRCAM image */

    for (i = 0; i < nimages; i++) {
        im = images[i];
        ehu_im = casu_fits_get_ehu(im);
        stds = mstds[i];
        if (stds == NULL)
            continue;
        ehu_cat = pl[i];
        cdfudge = 2.5*log10((double)pixsize(ehu_im)/NOMPIXSIZE);

        /* Now for the input catalogues. Start by getting some useful info
           from the header */

        saturate = cpl_propertylist_get_float(ehu_cat,"ESO QC SATURATION");
        apcor3 = cpl_propertylist_get_float(ehu_cat,"APCOR3");
        (void)vircam_pfits_get_exptime(casu_fits_get_phu(im),&exptime);
        (void)vircam_pfits_get_airmass(casu_fits_get_phu(im),&airmass);
        if (cpl_error_get_code() != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Unable to get header info for %s",
                          casu_fits_get_fullname(im));
            cpl_error_reset();
            continue;
        }

        /* Get objects that are not saturated and aren't too elliptical */

        cpl_table_select_all(stds);  
        (void)cpl_table_and_selected_float(stds,"Ellipticity",CPL_LESS_THAN,
                                           0.5);
        (void)cpl_table_and_selected_float(stds,"Peak_height",CPL_LESS_THAN,
                                           saturate);
        if (cpl_error_get_code() != CPL_ERROR_NONE) {
            cpl_msg_error(fctid,"Unable select data from matched stds tab %s",
                          casu_fits_get_fullname(im));
            cpl_error_reset();
            continue;
        }

        /* Now extract all those that have passed the test. If none are left
           then issue a warning and move on to the next one */

        cl = cpl_table_extract_selected(stds);
        ncat = (int)cpl_table_get_nrow(cl);
        if (ncat == 0) {
            cpl_msg_error(fctid,"No good standards available for %s",
                         casu_fits_get_fullname(im));
            cpl_table_delete(cl);
            cpl_error_reset();
            continue;
        }

        /* Dereference some of the columns */

        catcore3 = cpl_table_get_data_float(cl,"Aper_flux_3");
        xx = cpl_table_get_data_float(cl,"X_coordinate");
        yy = cpl_table_get_data_float(cl,"Y_coordinate");
        for (j = 0; j < p.ncolumns_coleq; j++) 
            stdmagptr[j] = cpl_table_get_data_float(cl,(p.coleq_columns)[j]);

        /* Loop for all the standards */

        extinct = p.atm_extcoef*(airmass - 1.0);
        for (j = 0; j < ncat; j++) {
            
            /* Do core magnitude calculation */

            cf = catcore3[j]/exptime;
            if (cf < 1.0)
                cf = 1.0;
            fluxmag3 = 2.5*log10((double)cf) + apcor3;

            /* Work out a reference magnitude */

            refmag = p.mag_offset;
            for (k = 0; k < p.ncolumns_coleq; k++) 
                refmag += ((p.coleq_coefs)[k]*stdmagptr[k][j]);

            /* Work out zero point */

            dm3 = refmag + fluxmag3 + extinct + cdfudge;

            /* What cell does this belong to? */

            ix = (int)(xx[j]/(float)nbsize);
            iy = (int)(yy[j]/(float)nbsize);
            ind = iy*nbx + ix;
            if (nres[ind] == nall[ind] - 1) {
                results[ind] = cpl_realloc(results[ind],
                                        (nall[ind]+INITALLOC)*sizeof(float));
                nall[ind] += INITALLOC;
            }
            results[ind][nres[ind]] = dm3;
            nres[ind] += 1;

            /* Add this into the global solution too */

            if (nresall == nallocall) {
                resall = cpl_realloc(resall,(nallocall+INITALLOC)*sizeof(float));
                nallocall += INITALLOC;
            }
            resall[nresall++] = dm3;
        }

        /* Get rid of the table subset */

        freetable(cl);
    }

    /* What is the global zero point ? */

    if (nresall != 0) {
        (void)casu_medmad(resall,NULL,(long)nresall,&medall,&mad);
        cut = max(3.0*1.48*mad,0.1);
        lcut = medall - cut;
        hcut = medall + cut;
        (void)casu_medmadcut(resall,NULL,(long)nresall,lcut,hcut,
                             &medall,&sig);
        sig *= 1.48;
    }

    /* Ok, what is the mean zeropoint for each cell. Do all of this stuff
       even if nresall == 0 so that we at least have a table with all
       zeros at the end of this */

    ngood = 0;
    for (i = 0; i < nbx*nby; i++) {
        if (nres[i] > 0) {
            (void)casu_medmad(results[i],NULL,(long)nres[i],&med,&mad);
            cut = max(3.0*1.48*mad,0.3);
            lcut = med - cut;
            hcut = med + cut;
            (void)casu_medmadcut(results[i],NULL,(long)nres[i],lcut,hcut,
                                 &med,&sig);
            sig *= 1.48;
            med -= medall;
            good[ngood++] = med;
        } else {
            med = -99.0;
        }

        /* Where are we in the map? */

        iy = i/nbx;
        ix = i - iy*nbx;
        xmin = ix*nbsize;
        xmax = xmin + nbsize - 1;
        if (ix == nbx)
            xmax = nx;
        ymin = iy*nbsize;
        ymax = ymin + nbsize - 1;
        if (iy == nby)
            ymax = ny;

        /* Write out the results into illumination table */

        cpl_table_set_float(*illcor,"xmin",(cpl_size)i,xmin);
        cpl_table_set_float(*illcor,"xmax",(cpl_size)i,xmax);
        cpl_table_set_float(*illcor,"ymin",(cpl_size)i,ymin);
        cpl_table_set_float(*illcor,"ymax",(cpl_size)i,ymax);
        cpl_table_set_float(*illcor,"illcor",(cpl_size)i,med);
    }

    /* Get the RMS of the map */

    if (ngood > 0)
        casu_medsig(good,NULL,(long)ngood,&med,illcor_rms);
    else
        *illcor_rms = 0.0;
    
    /* Delete some workspace */

    for (i = 0; i < nbx*nby; i++)
        freespace(results[i]);
    freespace(results);
    freespace(nres);
    freespace(nall);
    freespace(stdmagptr);
    freespace(resall);
    freespace(good);
    vircam_phot_close(&p);

    /* Set out the appropriate return status */

    if (nresall != 0)
        GOOD_STATUS
    else 
        WARN_RETURN

}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_phot_open
    \par Purpose:
        Open and parse the photcal table
    \par Description:
        Check the validity of the input photcal table. Select the row that
        represents the input filter. Parse the values in that row and fill
        in the relevant parts of the data structure.
    \par Language:
        C
    \param phottab
        The input input photometric calibration table.
    \param filt
        The input filter name
    \param p
        The photometric data structure
    \retval CASU_OK 
        if all went ok.
    \retval CASU_FATAL 
        if there were any errors in the table
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static int vircam_phot_open(cpl_table *phottab, char *filt, photstrct *p) {
    const char *fctid = "vircam_phot_open";
    int ns,nerr,null,nr;
    cpl_table *subset;
    char **filts;
    const char *req_cols[8] = {"filter_name","atm_extcoef","mag_offset",
                               "coleq_columns","coleq_coefs","gal_extcoef",
                               "default_zp","default_zp_err"};

    /* Initialise a few things */

    p->coleq_coefs = NULL;
    p->coleq_columns = NULL;
    p->ncolumns_coleq = 0;
    p->mag_offset = 0.0;

    /* Check the table and make sure it has all the required columns */

    nerr = 0;
    for (ns = 0; ns < 8; ns++) {
        if (! cpl_table_has_column(phottab,req_cols[ns])) {
            cpl_msg_error(fctid,"Photometry table missing column %s",
                          req_cols[ns]);
            nerr++;
        }
    }
    if (nerr > 0)
        return(CASU_FATAL);

    /* Search the table and find the rows that matches the filter. NB: we
       can't use cpl_table_and_selected_string because it matches by regular
       expression rather than exact matching! So just loop through and choose
       the first one that matches exactly. There shouldn't be more than one! */

    filts = cpl_table_get_data_string(phottab,"filter_name");
    nr = cpl_table_get_nrow(phottab);
    nerr = 1;
    for (ns = 0; ns < nr; ns++) {
        if (strncmp(filts[ns],filt,16) == 0) {
            nerr = 0;
            break;
        }
    }
    if (nerr == 1) {
        cpl_msg_error(fctid,"Unable to match photometry table to filter %s",
                      filt);
        return(CASU_FATAL);
    }
    cpl_table_and_selected_window(phottab,ns,1);

    /* Read the information from the selected row */

    subset = cpl_table_extract_selected(phottab);
    p->filter_name = (char *)cpl_table_get_string(subset,"filter_name",0);
    p->atm_extcoef = cpl_table_get_float(subset,"atm_extcoef",0,&null);
    p->mag_offset = cpl_table_get_float(subset,"mag_offset",0,&null);
    p->gal_extcoef = cpl_table_get_float(subset,"gal_extcoef",0,&null);
    p->default_zp = cpl_table_get_float(subset,"default_zp",0,&null);
    p->default_zp_err = cpl_table_get_float(subset,"default_zp_err",0,&null);
    if (extract_columns(subset,p) != CASU_OK) {
        freetable(subset);
        return(CASU_FATAL);
    }
    if (extract_coleq(subset,p) != CASU_OK) {
        freetable(subset);
        return(CASU_FATAL);
    }
    freetable(subset);

    /* Get out of here */

    return(CASU_OK);
}
    
/*---------------------------------------------------------------------------*/
/**
    \par Name:
        vircam_phot_close
    \par Purpose:
        Free the workspace associated with the photcal table
    \par Description:
        Free the workspace associated with the photcal table
    \par Language:
        C
    \param p
        The photometric data structure
    \return Nothing
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static void vircam_phot_close(photstrct *p) {
    int j;

    for (j = 0; j < p->ncolumns_coleq; j++)
        freespace((p->coleq_columns)[j]);
    freespace(p->coleq_columns);
    freespace(p->coleq_coefs);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        extract_columns
    \par Purpose:
        Read the 'columns' column from the table and parse.
    \par Description:
        Read the 'columns' column from the table and parse.
    \par Language:
        C
    \param tab
        The input photcal table
    \param p
        The photometric data structure
    \retval CASU_OK 
        Always
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static int extract_columns(cpl_table *tab, photstrct *p) {
    int nv,i,j;
    char *v,*w,**z;

    /* Get the relevant value from the table */

    v = cpl_strdup(cpl_table_get_string(tab,"coleq_columns",0));
        
    /* Count the number of commas in the string to see how many 
       columns there are */

    nv = 1;
    j = strlen(v);
    for (i = 0; i < j; i++)
        if (v[i] == ',')
            nv++;
    p->ncolumns_coleq = nv;

    /* Now parse the string into the column names */

    z = cpl_malloc(nv*sizeof(char *));
    char *saveptr = NULL;
    for (i = 0; i < nv; i++) {
        if (i == 0) 
            w = strtok_r(v,",",&saveptr);
        else
            w = strtok_r(NULL,",",&saveptr);
        z[i] = cpl_strdup(w);
    }
    p->coleq_columns = z;
    
    freespace(v);
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        extract_coleq
    \par Purpose:
        Extract the colour equation coefficients
    \par Description:
        Extract the colour equation coefficients
    \par Language:
        C
    \param tab
        The input photcal table
    \param p
        The photometric data structure
    \retval CASU_OK 
        Always
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static int extract_coleq(cpl_table *tab, photstrct *p) {
    int nv,i,j;
    char *v,*w;
    float *z;

    /* Get the relevant value from the table */

    v = cpl_strdup(cpl_table_get_string(tab,"coleq_coefs",0));

    /* Count the number of commas in the string to see how many 
       columns there are */

    nv = 1;
    j = strlen(v);
    for (i = 0; i < j; i++)
        if (v[i] == ',')
            nv++;

    /* Now parse the string into the colour equation values */

    z = cpl_malloc(nv*sizeof(float));
    char *saveptr1;
    for (i = 0; i < nv; i++) {
        if (i == 0) 
            w = strtok_r(v,",", &saveptr1);
        else
            w = strtok_r(NULL,",", &saveptr1);
        z[i] = (float)atof(w);
    }
    p->coleq_coefs = z;
    freespace(v);
    return(CASU_OK);
}

/*---------------------------------------------------------------------------*/
/**
    \par Name:
        pixsize
    \par Purpose:
        Work out a rough pixel size
    \par Description:
        Given the values of the X elements of the CD matrix in an input
        FITS header, work out a rough pixel size in arcseconds.
    \par Language:
        C
    \param plist
        The input propertylist representing an image extension FITS header
    \return The value of the pixel size in arcseconds
    \author
        Jim Lewis, CASU
*/
/*---------------------------------------------------------------------------*/

static double pixsize (cpl_propertylist *plist) {
    cpl_wcs *wcs;
    double *cd,pix;

    wcs = cpl_wcs_new_from_propertylist(plist);
    cd = cpl_matrix_get_data((cpl_matrix *)cpl_wcs_get_cd(wcs));
    pix = 3600.0*sqrt(fabs(cd[0]*cd[3] - cd[1]*cd[2]));
    cpl_wcs_delete(wcs);
    return(pix);
}
        
/**@}*/

/*

$Log: not supported by cvs2svn $
Revision 1.76  2012/01/27 12:25:10  jim
Fixed some casts

Revision 1.75  2012/01/16 12:32:18  jim
A few more changes to fit in with cpl6

Revision 1.74  2012/01/15 17:40:09  jim
Minor modifications to take into accout the changes in cpl API for v6

Revision 1.73  2010/12/09 13:23:56  jim
Minor doc fix

Revision 1.72  2010/09/09 12:11:09  jim
Fixed problems with docs that make doxygen barf

Revision 1.71  2010/06/30 12:42:00  jim
A few fixes to stop compiler compaints

Revision 1.70  2010/06/07 12:42:40  jim
Modifications to get rid of compiler gripes

Revision 1.69  2010/03/22 06:07:24  jim
Fixed bug in vircam_overexp

Revision 1.68  2010/03/22 06:04:14  jim
vircam_overexp now works out the ensemble stats on all frames rather than
just those that are between the threshold levels.

Revision 1.67  2010/01/31 19:30:45  jim
Fixed bug in where global MAD wasn't being multiplied by 1.48 to turn it
into an global RMS in the difference image routine

Revision 1.66  2009/09/21 11:57:27  jim
Modified vircam_overexp to pass back ensemble stats.

Revision 1.65  2009/06/08 08:07:13  jim
Fixed bugs in vircam_polyfit, where clipping loop was being done one too few
times and where the error estimate was being done incorrectly

Revision 1.64  2009/05/20 12:21:52  jim
modified _propertylist_merge to avoid clash of data types

Revision 1.63  2009/02/23 10:45:13  jim
Fixed vircam_backmap so that if there were no points in a bin the it does
something sensible to the value

Revision 1.62  2009/02/20 11:35:11  jim
Fixed bug in vircam_polyfit which stopped the iteration scheme.

Revision 1.61  2008/12/08 06:39:45  jim
Added new routine vircam_check_crval

Revision 1.60  2008/11/25 18:55:36  jim
rewrote vircam_sort to be a bit more reliable

Revision 1.59  2008/11/02 14:35:07  jim
Fixed teensy problem with two error messages

Revision 1.58  2008/10/13 08:16:57  jim
Fixed call to cpl_msg_warning

Revision 1.57  2008/10/01 04:58:30  jim
Added vircam_frameset_fexists

Revision 1.56  2008/07/10 13:04:03  jim
Fixed some comments

Revision 1.55  2008/05/06 12:15:02  jim
Changed vircam_catpars to check that we can read the index file and to send
out error messages if there are problems

Revision 1.54  2008/05/06 08:38:29  jim
Modified call to cpl_propertylist_get_ from float to double in
vircam_gaincor_calc.

Revision 1.53  2007/11/14 14:47:53  jim
vircam_linfit now works only with doubles

Revision 1.52  2007/11/14 10:46:07  jim
Fixed bugs in vircam_polyfit

Revision 1.51  2007/10/25 17:33:31  jim
Modified to add vircam_polyfit and to remove lint warnings

Revision 1.50  2007/10/19 09:25:10  jim
Fixed problems with missing includes

Revision 1.49  2007/10/19 06:55:06  jim
Modifications made to use new method for directing the recipes to the
standard catalogues using the sof

Revision 1.48  2007/10/15 12:50:28  jim
Modified for compatibility with cpl_4.0

Revision 1.47  2007/07/09 13:21:06  jim
Changed argument list to vircam_exten_range to include a cpl_frame

Revision 1.46  2007/05/08 10:41:01  jim
Added vircam_gaincor_calc

Revision 1.45  2007/05/03 11:15:33  jim
Fixed little problem with table wcs

Revision 1.44  2007/05/02 09:14:48  jim
Added vircam_findcol and vircam_rename_property

Revision 1.43  2007/03/01 12:42:42  jim
Modified slightly after code checking

Revision 1.42  2007/02/14 12:53:34  jim
Removed vircam_paf_print_header

Revision 1.41  2007/02/07 10:12:15  jim
Removed vircam_ndit_correct as it is now no longer necessary

Revision 1.40  2007/01/15 12:36:27  jim
Fixed bug in nditcor where factor was being divided by instead of multiplied

Revision 1.39  2006/12/06 12:58:41  jim
Fixed a small bug in vircam_backmap

Revision 1.38  2006/12/01 14:08:33  jim
added vircam_backmap

Revision 1.37  2006/11/28 20:52:17  jim
Added vircam_illcor_newtab

Revision 1.36  2006/11/27 12:00:17  jim
cosmetic changes

Revision 1.35  2006/10/05 09:23:00  jim
Small modifications to a couple of cpl calls to bring them into line with
cpl v3.0

Revision 1.34  2006/08/27 20:28:15  jim
added vircam_is_dummy

Revision 1.33  2006/08/21 09:09:29  jim
Added routines vircam_create_diffimg_stats and vircam_dummy_property

Revision 1.32  2006/07/11 14:53:58  jim
Fixed vircam_ndit_correct so that images with bad status are not corrected

Revision 1.31  2006/07/04 09:19:06  jim
replaced all sprintf statements with snprintf

Revision 1.30  2006/06/30 21:32:09  jim
Fixed bug in vircam_prov so that sprintf is replaced by snprintf

Revision 1.29  2006/06/20 18:59:51  jim
Fixed small bug in vircam_ndit_correct

Revision 1.28  2006/06/09 22:24:47  jim
Added vircam_ndit_correct

Revision 1.27  2006/06/09 11:26:26  jim
Small changes to keep lint happy

Revision 1.26  2006/06/06 13:07:54  jim
Change the stats routine used in vircam_difference_image

Revision 1.25  2006/05/30 13:31:47  jim
Added vircam_timestamp

Revision 1.24  2006/05/24 13:35:49  jim
Added vircam_dummy_image and vircam_dummy_catalogue

Revision 1.23  2006/05/04 15:16:33  jim
Fixed memory bug in vircam_overexp

Revision 1.22  2006/04/20 11:29:41  jim
Added vircam_overexp

Revision 1.21  2006/03/22 11:42:32  jim
Moved some routines into vircam_mask

Revision 1.20  2006/03/15 10:43:42  jim
Fixed a few things

Revision 1.19  2006/03/08 14:32:22  jim
Lots of little modifications

Revision 1.18  2006/03/01 10:31:29  jim
Now uses new vir_fits objects

Revision 1.17  2006/02/18 11:48:55  jim
*** empty log message ***

Revision 1.16  2006/01/23 10:30:50  jim
Mainly documentation mods

Revision 1.15  2005/12/14 22:17:33  jim
Updated docs

Revision 1.14  2005/12/01 16:26:03  jim
Added vircam_dummy_confidence

Revision 1.13  2005/11/25 15:33:22  jim
Some code fixes to keep splint happy

Revision 1.12  2005/11/25 09:56:15  jim
Tidied up some more documentation

Revision 1.11  2005/11/25 09:36:23  jim
Added vircam_linfit

Revision 1.10  2005/11/07 13:15:16  jim
Fixed lots of bugs and added some error checking

Revision 1.9  2005/11/03 15:16:28  jim
Lots of changes mainly to strengthen error reporting

Revision 1.8  2005/11/03 13:28:50  jim
All sorts of changes to tighten up error handling

Revision 1.7  2005/10/14 13:21:04  jim
Added some new routines

Revision 1.6  2005/09/20 15:07:46  jim
Fixed a few bugs and added a few things

Revision 1.5  2005/09/07 12:47:22  jim
renamed a malloc and free to the cpl equivallent

Revision 1.4  2005/08/10 09:55:05  jim
Modified vircam_madfit so that if determinant is zero, then it sends a
zero slope and intercept back

Revision 1.3  2005/08/10 08:42:00  jim
Modified vircam_madfit so that if the initial least squares fit gets a
perfect result, then it just returns with those values rather than to try
and do the median fit

Revision 1.2  2005/08/09 15:30:00  jim
vircam_exten_range had number of chips wrong!

Revision 1.1.1.1  2005/08/05 08:29:09  jim
Initial import


*/

